Package: pysparklyr
Title: Provides a 'PySpark' Back-End for the 'sparklyr' Package
Version: 0.2.1
Authors@R: c(
    person("Edgar", "Ruiz", , "edgar@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: It enables 'sparklyr' to integrate with 'Spark Connect', and
    'Databricks Connect' by providing a wrapper over the 'PySpark'
    'python' library.
License: MIT + file LICENSE
URL: https://github.com/mlverse/pysparklyr
BugReports: https://github.com/mlverse/pysparklyr/issues
Depends: R (>= 4.1)
Imports: arrow, cli, DBI, dbplyr, dplyr, fs, glue, httr2, lifecycle,
        methods, processx, purrr, reticulate (>= 1.44.0), rlang,
        rstudioapi, sparklyr (>= 1.9.4), tidyr, tidyselect, vctrs,
        uuid, withr, connectcreds
Suggests: vcr, crayon, R6, testthat (>= 3.0.0), tibble, rsconnect,
        rsample, workflows, tune, parsnip, dials, tailor, recipes
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-11-12
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-20 17:31:03 UTC; edgar
Author: Edgar Ruiz [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Edgar Ruiz <edgar@posit.co>
Repository: CRAN
Date/Publication: 2026-04-20 20:10:02 UTC
Built: R 4.6.0; ; 2026-04-26 20:47:13 UTC; unix

Package: pysparklyr
Title: Provides a 'PySpark' Back-End for the 'sparklyr' Package
Version: 0.2.1
Authors@R: c(
    person("Edgar", "Ruiz", , "edgar@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: It enables 'sparklyr' to integrate with 'Spark Connect', and
    'Databricks Connect' by providing a wrapper over the 'PySpark'
    'python' library.
License: MIT + file LICENSE
URL: https://github.com/mlverse/pysparklyr
BugReports: https://github.com/mlverse/pysparklyr/issues
Depends: R (>= 4.1)
Imports: arrow, cli, DBI, dbplyr, dplyr, fs, glue, httr2, lifecycle,
        methods, processx, purrr, reticulate (>= 1.44.0), rlang,
        rstudioapi, sparklyr (>= 1.9.4), tidyr, tidyselect, vctrs,
        uuid, withr, connectcreds
Suggests: vcr, crayon, R6, testthat (>= 3.0.0), tibble, rsconnect,
        rsample, workflows, tune, parsnip, dials, tailor, recipes
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-11-12
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-20 17:31:03 UTC; edgar
Author: Edgar Ruiz [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Edgar Ruiz <edgar@posit.co>
Repository: CRAN
Date/Publication: 2026-04-20 20:10:02 UTC
Built: R 4.6.0; ; 2026-04-30 01:15:37 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",tbl_pyspark)
S3method(collect,spark_pyobj)
S3method(collect,tbl_pyspark)
S3method(compute,tbl_pyspark)
S3method(connection_is_open,pyspark_connection)
S3method(ft_binarizer,ml_connect_pipeline)
S3method(ft_binarizer,pyspark_connection)
S3method(ft_binarizer,tbl_pyspark)
S3method(ft_bucketed_random_projection_lsh,ml_connect_pipeline)
S3method(ft_bucketed_random_projection_lsh,pyspark_connection)
S3method(ft_bucketed_random_projection_lsh,tbl_pyspark)
S3method(ft_bucketizer,ml_connect_pipeline)
S3method(ft_bucketizer,pyspark_connection)
S3method(ft_bucketizer,tbl_pyspark)
S3method(ft_count_vectorizer,ml_connect_pipeline)
S3method(ft_count_vectorizer,pyspark_connection)
S3method(ft_count_vectorizer,tbl_pyspark)
S3method(ft_dct,ml_connect_pipeline)
S3method(ft_dct,pyspark_connection)
S3method(ft_dct,tbl_pyspark)
S3method(ft_discrete_cosine_transform,ml_connect_pipeline)
S3method(ft_discrete_cosine_transform,pyspark_connection)
S3method(ft_discrete_cosine_transform,tbl_pyspark)
S3method(ft_elementwise_product,ml_connect_pipeline)
S3method(ft_elementwise_product,pyspark_connection)
S3method(ft_elementwise_product,tbl_pyspark)
S3method(ft_feature_hasher,ml_connect_pipeline)
S3method(ft_feature_hasher,pyspark_connection)
S3method(ft_feature_hasher,tbl_pyspark)
S3method(ft_hashing_tf,ml_connect_pipeline)
S3method(ft_hashing_tf,pyspark_connection)
S3method(ft_hashing_tf,tbl_pyspark)
S3method(ft_idf,ml_connect_pipeline)
S3method(ft_idf,pyspark_connection)
S3method(ft_idf,tbl_pyspark)
S3method(ft_imputer,ml_connect_pipeline)
S3method(ft_imputer,pyspark_connection)
S3method(ft_imputer,tbl_pyspark)
S3method(ft_index_to_string,ml_connect_pipeline)
S3method(ft_index_to_string,pyspark_connection)
S3method(ft_index_to_string,tbl_pyspark)
S3method(ft_interaction,ml_connect_pipeline)
S3method(ft_interaction,pyspark_connection)
S3method(ft_interaction,tbl_pyspark)
S3method(ft_max_abs_scaler,ml_connect_pipeline)
S3method(ft_max_abs_scaler,pyspark_connection)
S3method(ft_max_abs_scaler,tbl_pyspark)
S3method(ft_min_max_scaler,ml_connect_pipeline)
S3method(ft_min_max_scaler,pyspark_connection)
S3method(ft_min_max_scaler,tbl_pyspark)
S3method(ft_minhash_lsh,ml_connect_pipeline)
S3method(ft_minhash_lsh,pyspark_connection)
S3method(ft_minhash_lsh,tbl_pyspark)
S3method(ft_ngram,ml_connect_pipeline)
S3method(ft_ngram,pyspark_connection)
S3method(ft_ngram,tbl_pyspark)
S3method(ft_normalizer,ml_connect_pipeline)
S3method(ft_normalizer,pyspark_connection)
S3method(ft_normalizer,tbl_pyspark)
S3method(ft_one_hot_encoder,ml_connect_pipeline)
S3method(ft_one_hot_encoder,pyspark_connection)
S3method(ft_one_hot_encoder,tbl_pyspark)
S3method(ft_pca,ml_connect_pipeline)
S3method(ft_pca,pyspark_connection)
S3method(ft_pca,tbl_pyspark)
S3method(ft_polynomial_expansion,ml_connect_pipeline)
S3method(ft_polynomial_expansion,pyspark_connection)
S3method(ft_polynomial_expansion,tbl_pyspark)
S3method(ft_quantile_discretizer,ml_connect_pipeline)
S3method(ft_quantile_discretizer,pyspark_connection)
S3method(ft_quantile_discretizer,tbl_pyspark)
S3method(ft_r_formula,ml_connect_pipeline)
S3method(ft_r_formula,pyspark_connection)
S3method(ft_r_formula,tbl_pyspark)
S3method(ft_regex_tokenizer,ml_connect_pipeline)
S3method(ft_regex_tokenizer,pyspark_connection)
S3method(ft_regex_tokenizer,tbl_pyspark)
S3method(ft_robust_scaler,ml_connect_pipeline)
S3method(ft_robust_scaler,pyspark_connection)
S3method(ft_robust_scaler,tbl_pyspark)
S3method(ft_sql_transformer,ml_connect_pipeline)
S3method(ft_sql_transformer,pyspark_connection)
S3method(ft_sql_transformer,tbl_pyspark)
S3method(ft_standard_scaler,ml_connect_pipeline)
S3method(ft_standard_scaler,pyspark_connection)
S3method(ft_standard_scaler,tbl_pyspark)
S3method(ft_stop_words_remover,ml_connect_pipeline)
S3method(ft_stop_words_remover,pyspark_connection)
S3method(ft_stop_words_remover,tbl_pyspark)
S3method(ft_string_indexer,ml_connect_pipeline)
S3method(ft_string_indexer,pyspark_connection)
S3method(ft_string_indexer,tbl_pyspark)
S3method(ft_tokenizer,ml_connect_pipeline)
S3method(ft_tokenizer,pyspark_connection)
S3method(ft_tokenizer,tbl_pyspark)
S3method(ft_vector_assembler,ml_connect_pipeline)
S3method(ft_vector_assembler,pyspark_connection)
S3method(ft_vector_assembler,tbl_pyspark)
S3method(ft_vector_indexer,ml_connect_pipeline)
S3method(ft_vector_indexer,pyspark_connection)
S3method(ft_vector_indexer,tbl_pyspark)
S3method(ft_vector_slicer,ml_connect_pipeline)
S3method(ft_vector_slicer,pyspark_connection)
S3method(ft_vector_slicer,tbl_pyspark)
S3method(ft_word2vec,ml_connect_pipeline)
S3method(ft_word2vec,pyspark_connection)
S3method(ft_word2vec,tbl_pyspark)
S3method(head,tbl_pyspark)
S3method(hive_context,pyspark_connection)
S3method(invoke,ml_connect_estimator)
S3method(invoke,ml_connect_model)
S3method(invoke,ml_connect_transformer)
S3method(invoke,pyspark_connection)
S3method(invoke,python.builtin.object)
S3method(invoke,spark_pyobj)
S3method(invoke,tbl_pyspark)
S3method(invoke_static,pyspark_connection)
S3method(ml_aft_survival_regression,ml_connect_pipeline)
S3method(ml_aft_survival_regression,pyspark_connection)
S3method(ml_aft_survival_regression,tbl_pyspark)
S3method(ml_binary_classification_evaluator,pyspark_connection)
S3method(ml_binary_classification_evaluator,tbl_pyspark)
S3method(ml_bisecting_kmeans,ml_connect_pipeline)
S3method(ml_bisecting_kmeans,pyspark_connection)
S3method(ml_bisecting_kmeans,tbl_pyspark)
S3method(ml_clustering_evaluator,pyspark_connection)
S3method(ml_clustering_evaluator,tbl_pyspark)
S3method(ml_cross_validator,pyspark_connection)
S3method(ml_decision_tree_classifier,ml_connect_pipeline)
S3method(ml_decision_tree_classifier,pyspark_connection)
S3method(ml_decision_tree_classifier,tbl_pyspark)
S3method(ml_decision_tree_regressor,ml_connect_pipeline)
S3method(ml_decision_tree_regressor,pyspark_connection)
S3method(ml_decision_tree_regressor,tbl_pyspark)
S3method(ml_fit,ml_connect_cross_validator)
S3method(ml_fit,ml_connect_pipeline)
S3method(ml_gbt_classifier,ml_connect_pipeline)
S3method(ml_gbt_classifier,pyspark_connection)
S3method(ml_gbt_classifier,tbl_pyspark)
S3method(ml_gbt_regressor,ml_connect_pipeline)
S3method(ml_gbt_regressor,pyspark_connection)
S3method(ml_gbt_regressor,tbl_pyspark)
S3method(ml_generalized_linear_regression,ml_connect_pipeline)
S3method(ml_generalized_linear_regression,pyspark_connection)
S3method(ml_generalized_linear_regression,tbl_pyspark)
S3method(ml_isotonic_regression,ml_connect_pipeline)
S3method(ml_isotonic_regression,pyspark_connection)
S3method(ml_isotonic_regression,tbl_pyspark)
S3method(ml_kmeans,ml_connect_pipeline)
S3method(ml_kmeans,pyspark_connection)
S3method(ml_kmeans,tbl_pyspark)
S3method(ml_linear_regression,ml_connect_pipeline)
S3method(ml_linear_regression,pyspark_connection)
S3method(ml_linear_regression,tbl_pyspark)
S3method(ml_logistic_regression,ml_connect_pipeline)
S3method(ml_logistic_regression,pyspark_connection)
S3method(ml_logistic_regression,tbl_pyspark)
S3method(ml_multiclass_classification_evaluator,pyspark_connection)
S3method(ml_multiclass_classification_evaluator,tbl_pyspark)
S3method(ml_pipeline,pyspark_connection)
S3method(ml_predict,ml_connect_model)
S3method(ml_random_forest_classifier,ml_connect_pipeline)
S3method(ml_random_forest_classifier,pyspark_connection)
S3method(ml_random_forest_classifier,tbl_pyspark)
S3method(ml_random_forest_regressor,ml_connect_pipeline)
S3method(ml_random_forest_regressor,pyspark_connection)
S3method(ml_random_forest_regressor,tbl_pyspark)
S3method(ml_regression_evaluator,pyspark_connection)
S3method(ml_regression_evaluator,tbl_pyspark)
S3method(ml_save,ml_connect_pipeline_stage)
S3method(ml_transform,ml_connect_pipeline_model)
S3method(pivot_longer,tbl_pyspark)
S3method(print,ml_connect_cross_validator_model)
S3method(print,ml_connect_estimator)
S3method(print,ml_connect_model)
S3method(print,ml_output_params)
S3method(print,spark_pyobj)
S3method(python_obj_get,default)
S3method(python_obj_get,ml_connect_model)
S3method(python_obj_get,ml_connect_pipeline)
S3method(python_obj_get,ml_connect_pipeline_model)
S3method(python_obj_get,python.builtin.object)
S3method(python_obj_get,spark_pyobj)
S3method(python_obj_get,tbl_pyspark)
S3method(same_src,pyspark_connection)
S3method(sample_frac,tbl_pyspark)
S3method(sample_n,tbl_pyspark)
S3method(sdf_copy_to,pyspark_connection)
S3method(sdf_read_column,spark_pyjobj)
S3method(sdf_register,spark_pyobj)
S3method(sdf_schema,tbl_pyspark)
S3method(spark_apply,tbl_pyspark)
S3method(spark_connect_method,spark_method_databricks_connect)
S3method(spark_connect_method,spark_method_snowpark_connect)
S3method(spark_connect_method,spark_method_spark_connect)
S3method(spark_connection,pyspark_connection)
S3method(spark_connection,spark_pyobj)
S3method(spark_dataframe,pyspark_connection)
S3method(spark_dataframe,spark_pyobj)
S3method(spark_dataframe,tbl_pyspark)
S3method(spark_ide_columns,pyspark_connection)
S3method(spark_ide_objects,pyspark_connection)
S3method(spark_ide_preview,pyspark_connection)
S3method(spark_integ_test_skip,connect_databricks)
S3method(spark_integ_test_skip,connect_spark)
S3method(spark_jobj,ml_connect_model)
S3method(spark_log,pyspark_connection)
S3method(spark_read_csv,pyspark_connection)
S3method(spark_read_json,pyspark_connection)
S3method(spark_read_orc,pyspark_connection)
S3method(spark_read_parquet,pyspark_connection)
S3method(spark_read_text,pyspark_connection)
S3method(spark_session,pyspark_connection)
S3method(spark_version,pyspark_connection)
S3method(spark_version,spark_pyobj)
S3method(spark_web,pyspark_connection)
S3method(spark_write_csv,tbl_pyspark)
S3method(spark_write_delta,tbl_pyspark)
S3method(spark_write_json,tbl_pyspark)
S3method(spark_write_orc,tbl_pyspark)
S3method(spark_write_parquet,tbl_pyspark)
S3method(spark_write_table,tbl_pyspark)
S3method(spark_write_text,tbl_pyspark)
S3method(tbl,pyspark_connection)
S3method(tbl_ptype,tbl_pyspark)
S3method(tidyselect_data_has_predicates,tbl_pyspark)
S3method(tune_grid_spark,pyspark_connection)
export(connection_databricks_shinyapp)
export(deploy_databricks)
export(install_databricks)
export(install_pyspark)
export(installed_components)
export(ml_prepare_dataset)
export(pyspark_config)
export(requirements_write)
export(spark_connect_service_start)
export(spark_connect_service_stop)
import(DBI)
import(cli)
import(dbplyr)
import(fs)
import(glue)
import(httr2)
import(reticulate)
importFrom(connectcreds,connect_viewer_token)
importFrom(connectcreds,has_viewer_token)
importFrom(dplyr,as_tibble)
importFrom(dplyr,collect)
importFrom(dplyr,compute)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,group_vars)
importFrom(dplyr,mutate)
importFrom(dplyr,same_src)
importFrom(dplyr,sample_frac)
importFrom(dplyr,sample_n)
importFrom(dplyr,select)
importFrom(dplyr,slice_sample)
importFrom(dplyr,tbl)
importFrom(dplyr,tbl_ptype)
importFrom(dplyr,tibble)
importFrom(glue,glue)
importFrom(lifecycle,deprecated)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setOldClass)
importFrom(processx,process)
importFrom(purrr,as_mapper)
importFrom(purrr,discard)
importFrom(purrr,flatten)
importFrom(purrr,imap)
importFrom(purrr,imap_chr)
importFrom(purrr,iwalk)
importFrom(purrr,list_rbind)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_chr)
importFrom(purrr,map_lgl)
importFrom(purrr,map_vec)
importFrom(purrr,pmap_chr)
importFrom(purrr,reduce)
importFrom(purrr,transpose)
importFrom(purrr,walk)
importFrom(rlang,`!!!`)
importFrom(rlang,`!!`)
importFrom(rlang,`%||%`)
importFrom(rlang,abort)
importFrom(rlang,arg_match)
importFrom(rlang,as_utf8_character)
importFrom(rlang,enquo)
importFrom(rlang,exec)
importFrom(rlang,is_character)
importFrom(rlang,is_missing)
importFrom(rlang,is_string)
importFrom(rlang,parse_exprs)
importFrom(rlang,quo_is_null)
importFrom(rlang,set_names)
importFrom(rlang,sym)
importFrom(rlang,warn)
importFrom(rstudioapi,getSourceEditorContext)
importFrom(rstudioapi,jobRunScript)
importFrom(rstudioapi,showQuestion)
importFrom(sparklyr,connection_is_open)
importFrom(sparklyr,ft_binarizer)
importFrom(sparklyr,ft_bucketed_random_projection_lsh)
importFrom(sparklyr,ft_bucketizer)
importFrom(sparklyr,ft_count_vectorizer)
importFrom(sparklyr,ft_dct)
importFrom(sparklyr,ft_discrete_cosine_transform)
importFrom(sparklyr,ft_elementwise_product)
importFrom(sparklyr,ft_feature_hasher)
importFrom(sparklyr,ft_hashing_tf)
importFrom(sparklyr,ft_idf)
importFrom(sparklyr,ft_imputer)
importFrom(sparklyr,ft_index_to_string)
importFrom(sparklyr,ft_interaction)
importFrom(sparklyr,ft_max_abs_scaler)
importFrom(sparklyr,ft_min_max_scaler)
importFrom(sparklyr,ft_minhash_lsh)
importFrom(sparklyr,ft_ngram)
importFrom(sparklyr,ft_normalizer)
importFrom(sparklyr,ft_one_hot_encoder)
importFrom(sparklyr,ft_pca)
importFrom(sparklyr,ft_polynomial_expansion)
importFrom(sparklyr,ft_quantile_discretizer)
importFrom(sparklyr,ft_r_formula)
importFrom(sparklyr,ft_regex_tokenizer)
importFrom(sparklyr,ft_robust_scaler)
importFrom(sparklyr,ft_sql_transformer)
importFrom(sparklyr,ft_standard_scaler)
importFrom(sparklyr,ft_stop_words_remover)
importFrom(sparklyr,ft_string_indexer)
importFrom(sparklyr,ft_tokenizer)
importFrom(sparklyr,ft_vector_assembler)
importFrom(sparklyr,ft_vector_indexer)
importFrom(sparklyr,ft_vector_slicer)
importFrom(sparklyr,ft_word2vec)
importFrom(sparklyr,hive_context)
importFrom(sparklyr,invoke)
importFrom(sparklyr,invoke_new)
importFrom(sparklyr,invoke_static)
importFrom(sparklyr,ml_aft_survival_regression)
importFrom(sparklyr,ml_binary_classification_evaluator)
importFrom(sparklyr,ml_bisecting_kmeans)
importFrom(sparklyr,ml_clustering_evaluator)
importFrom(sparklyr,ml_cross_validator)
importFrom(sparklyr,ml_decision_tree_classifier)
importFrom(sparklyr,ml_decision_tree_regressor)
importFrom(sparklyr,ml_fit)
importFrom(sparklyr,ml_gbt_classifier)
importFrom(sparklyr,ml_gbt_regressor)
importFrom(sparklyr,ml_generalized_linear_regression)
importFrom(sparklyr,ml_isotonic_regression)
importFrom(sparklyr,ml_kmeans)
importFrom(sparklyr,ml_linear_regression)
importFrom(sparklyr,ml_load)
importFrom(sparklyr,ml_logistic_regression)
importFrom(sparklyr,ml_multiclass_classification_evaluator)
importFrom(sparklyr,ml_pipeline)
importFrom(sparklyr,ml_predict)
importFrom(sparklyr,ml_random_forest_classifier)
importFrom(sparklyr,ml_random_forest_regressor)
importFrom(sparklyr,ml_regression_evaluator)
importFrom(sparklyr,ml_save)
importFrom(sparklyr,ml_transform)
importFrom(sparklyr,random_string)
importFrom(sparklyr,sdf_copy_to)
importFrom(sparklyr,sdf_read_column)
importFrom(sparklyr,sdf_register)
importFrom(sparklyr,sdf_schema)
importFrom(sparklyr,spark_apply)
importFrom(sparklyr,spark_connect_method)
importFrom(sparklyr,spark_connection)
importFrom(sparklyr,spark_dataframe)
importFrom(sparklyr,spark_ide_columns)
importFrom(sparklyr,spark_ide_connection_updated)
importFrom(sparklyr,spark_ide_objects)
importFrom(sparklyr,spark_ide_preview)
importFrom(sparklyr,spark_install_find)
importFrom(sparklyr,spark_integ_test_skip)
importFrom(sparklyr,spark_jobj)
importFrom(sparklyr,spark_log)
importFrom(sparklyr,spark_read_csv)
importFrom(sparklyr,spark_read_json)
importFrom(sparklyr,spark_read_orc)
importFrom(sparklyr,spark_read_parquet)
importFrom(sparklyr,spark_read_text)
importFrom(sparklyr,spark_session)
importFrom(sparklyr,spark_table_name)
importFrom(sparklyr,spark_version)
importFrom(sparklyr,spark_web)
importFrom(sparklyr,spark_write_csv)
importFrom(sparklyr,spark_write_delta)
importFrom(sparklyr,spark_write_json)
importFrom(sparklyr,spark_write_orc)
importFrom(sparklyr,spark_write_parquet)
importFrom(sparklyr,spark_write_table)
importFrom(sparklyr,spark_write_text)
importFrom(sparklyr,tune_grid_spark)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(tidyr,pivot_longer)
importFrom(tidyselect,matches)
importFrom(tidyselect,tidyselect_data_has_predicates)
importFrom(utils,capture.output)
importFrom(utils,compareVersion)
importFrom(utils,head)
importFrom(utils,installed.packages)
importFrom(utils,menu)
importFrom(utils,type.convert)
importFrom(uuid,UUIDgenerate)
importFrom(vctrs,vec_as_names)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# pysparklyr

<!-- badges: start -->

[![R-CMD-check](https://github.com/mlverse/pysparklyr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mlverse/pysparklyr/actions/workflows/R-CMD-check.yaml)
[![Spark-Connect](https://github.com/mlverse/pysparklyr/actions/workflows/spark-tests.yaml/badge.svg)](https://github.com/mlverse/pysparklyr/actions/workflows/spark-tests.yaml)
[![codecov](https://codecov.io/gh/mlverse/pysparklyr/graph/badge.svg?token=O1N9qPabpF)](https://app.codecov.io/gh/mlverse/pysparklyr)
[![CRAN
status](https://www.r-pkg.org/badges/version/pysparklyr)](https://CRAN.R-project.org/package=pysparklyr)
<!-- badges: end -->

Integrates `sparklyr` with PySpark and Databricks. The main reason of
this package is because the new Spark and Databricks Connect connection
method does not work with standard `sparklyr` integration.

## Installing

To install the version in CRAN use:

``` r
install.packages("pysparklyr")
```

To get the development version from GitHub use:

``` r
remotes::install_github("mlverse/pysparklyr")
```

## Using

To learn how to use, please visit the Spark / Databricks Connect
article, available in the official `sparklyr` website: [Spark Connect,
and Databricks Connect
v2](https://spark.posit.co/deployment/databricks-connect.html)

function(df = mtcars) {
  library(arrow);
  fn <- function(...) 1;
  fn_run <- fn(df);
  gp_field <- 'am';
  col_names <- c('am', 'x');
  gp <- df[1, gp_field];
  if(is.vector(fn_run)) {
    ret <- data.frame(x = fn_run);
  } else {
    ret <- as.data.frame(fn_run);
  };
  ret[gp_field] <- gp;
  cols <- colnames(ret);
  gp_cols <- cols == gp_field;
  new_cols <- c(cols[gp_cols], cols[!gp_cols]);
  ret <- ret[, new_cols];
  if(!is.null(col_names)) colnames(ret) <- col_names;
  ret
}

function(df = mtcars, context = 1) {
  library(arrow);
  fn <- function(...) 1;
  fn_run <- fn(df, context);
  gp_field <- 'am';
  col_names <- c('am', 'x');
  gp <- df[1, gp_field];
  if(is.vector(fn_run)) {
    ret <- data.frame(x = fn_run);
  } else {
    ret <- as.data.frame(fn_run);
  };
  ret[gp_field] <- gp;
  cols <- colnames(ret);
  gp_cols <- cols == gp_field;
  new_cols <- c(cols[gp_cols], cols[!gp_cols]);
  ret <- ret[, new_cols];
  if(!is.null(col_names)) colnames(ret) <- col_names;
  ret
}

library(rlang)
library(fs)

rsApiUpdateDialog <- function(code) {
  if (exists(".rs.api.updateDialog")) {
    updateDialog <- get(".rs.api.updateDialog")
    updateDialog(code = code)
  }
}

is_windows <- function() {
  identical(.Platform$OS.type, "windows")
}

# Copied from reticulate
is_virtualenv <- function(dir) {
  # check for expected files for virtualenv / venv
  subdir <- if (is_windows()) "Scripts" else "bin"
  files <- c(
    file.path(subdir, "activate_this.py"),
    file.path(subdir, "pyvenv.cfg"),
    "pyvenv.cfg"
  )
  paths <- file.path(dir, files)
  any(file.exists(paths))
}


#' @import rstudioapi
connection_spark_ui <- function() {
  env_var_name <- "DATABRICKS_SELECTED_CLUSTER_ID"
  if (Sys.getenv(env_var_name) != "") {
    cluster_label <- Sys.getenv(env_var_name)
    Sys.unsetenv(env_var_name)
  } else {
    cluster_label <- ""
  }

  elementSpacing <- if (is_windows()) 2 else 7

  tags$div(
    tags$style(
      type = "text/css",
      paste0(
        "table {
          font-family: 'Verdana', 'Helvetica';
          font-size: 12px;
        }",
        ".shiny-input-text {
          font-family: 'Verdana', 'Helvetica';
          font-size: 12px;
          width: 100%;
        }"
      )
    ),
    tags$table(
      tags$tr(
        tags$td("Cluster ID:"),
        tags$td(
          textInput(
            inputId = "cluster_id",
            label = "",
            value = cluster_label,
            width = "200px"
          ),
          textOutput("get_version"),
          style = paste("height: 20px")
        )
      ),
      tags$tr(
        tags$td(style = paste("height: 5px"))
      ),
      tags$tr(
        tags$td("Python Env:"),
        tags$td(uiOutput("get_env"))
      ),
      tags$tr(
        tags$td(style = paste("height: 5px"))
      ),
      tags$tr(
        tags$td(textOutput("dbr_label")),
        tags$td(uiOutput("dbr_ui"))
      ),
      tags$tr(
        tags$td(textOutput("host_label")),
        tags$td(uiOutput("host_ui"))
      ),
      tags$tr(
        tags$td(style = paste("height: 0px")),
        tags$td(textOutput("matches_host"))
      ),
      tags$tr(
        tags$td(style = paste("height: 5px"))
      ),
      tags$tr(
        tags$td(textOutput("auth_label")),
        tags$td(textOutput("auth_ui"))
      )
    )
  )
}

connection_spark_server <- function(input, output, session) {
  dbr_version <- reactiveVal("")
  token <- pysparklyr:::databricks_token()
  host <- pysparklyr:::databricks_host(fail = FALSE)
  output$auth_label <- reactive({
    ret <- ""
    if (!is.null(names(token))) {
      ret <- "Password:"
    }
    ret
  })

  output$auth_ui <- reactive({
    t_source <- names(token)
    if (is.null(t_source)) {
      ret <- ""
    } else {
      if (t_source == "environment") {
        ret <- "✓ Found - Using 'DATABRICKS_TOKEN'"
      }
      if (t_source == "oauth") {
        ret <- "✓ Found - Managed by Posit Workbench OAuth"
      }
      if (t_source == "") {
        ret <- "✘ Not Found - Add it to your 'DATABRICKS_TOKEN' env variable"
      }
    }
    ret
  })

  output$host_label <- reactive({
    ret <- ""
    if (host != "") {
      ret <- "Host URL:"
    }
    ret
  })

  output$host_ui <- renderUI({
    if (host == "") {
      tags$p("")
    } else {
      textInput(
        inputId = "host_url",
        label = "",
        value = host,
        width = "400px"
      )
    }
  })

  output$dbr_label <- reactive({
    ret <- ""
    if (token == "") {
      ret <- "DBR Version:"
    }
    ret
  })

  output$dbr_ui <- renderUI({
    if (token != "") {
      ret <- tags$p("")
    } else {
      ret <- textInput(
        inputId = "dbr_ver",
        label = "",
        value = "",
        width = "50px"
      )
    }
    ret
  })

  output$matches_host <- reactive({
    ret <- ""
    if (!is.null(input$host)) {
      if (input$host != host) {
        ret <- "✓ Using supplied custom Host URL in code"
      }
      if (host == "") {
        ret <- ""
      }
      ret
    }
  })

  output$get_version <- reactive({
    ret <- ""
    host_url <- input$host_url %||% ""
    if (input$cluster_id != "" && host_url != "") {
      version <- try(pysparklyr:::databricks_dbr_version(
        cluster_id = input$cluster_id,
        host = input$host_url,
        token = token
      ))
      if (!inherits(version, "try-error")) {
        dbr_version(version)
        ret <- paste0("✓ Found - Cluster's DBR is ", version)
      } else {
        dbr_version("")
        if (token != "") {
          ret <- "✘ Could not verify cluster version"
        }
      }
    }
    ret
  })

  output$get_env <- renderUI({
    env <- c(
      "Create a temporary UV environment" = "default",
      "Let {reticulate} find an environment" = "reticulate"
    )
    sel_env <- "default"
    env_folders <- c(".venv", ".virtualenv")
    active_project <- try(rstudioapi::getActiveProject(), silent = TRUE)
    if (!inherits(active_project, "try-error")) {
      project_folders <- path(active_project, env_folders)
    } else {
      project_folders <- NULL
    }
    home_folders <- path_abs(path("~", env_folders))
    home_folders <- home_folders[file_exists(home_folders)]
    env_folders <- c(env_folders, home_folders, project_folders)
    for (folder in env_folders) {
      if (is_virtualenv(folder)) {
        sel_env <- folder
        env <- c(env, folder)
      }
    }
    version <- input$dbr_ver %||% dbr_version()
    try_version <- try(pysparklyr:::version_prep(version), silent = TRUE)
    err_version <- inherits(try_version, "try-error")
    if (version != "" && !err_version) {
      if (!inherits(version, "try-error")) {
        verified <- pysparklyr:::use_envname(
          version = version,
          backend = "databricks",
          main_library = "databricks-connect",
          messages = FALSE,
          match_first = FALSE,
          ignore_reticulate_python = TRUE
        )
        verified_name <- names(verified)
        if (verified_name == "exact") {
          sel_env <- verified
          env <- c(env, as.character(verified))
        }
      }
    }
    selectInput("py_env", "", choices = env, selected = sel_env)
  })

  code_create <- function(cluster_id, host_url, dbr, envname = NULL) {
    host_label <- NULL
    dbr_label <- NULL
    envname_label <- NULL

    if (is.null(host_url)) {
      host_url <- ""
    }
    if (host != "" && host != host_url) {
      host <- paste0("    master = \"", host_url, "\",")
    }
    if (!is.null(dbr)) {
      if (dbr != "") {
        dbr_label <- paste0("    version = \"", dbr, "\",")
      }
    }
    if (!is.null(envname)) {
      if (envname == "reticulate") {
        envname <- ""
      }
      if (envname != "default") {
        envname_label <- paste0("    envname = \"", envname, "\",")
      }
    }
    code_lines <- c(
      "library(sparklyr)",
      "sc <- spark_connect(",
      paste0("    cluster_id = \"", cluster_id, "\","),
      dbr_label,
      host_label,
      envname_label,
      "    method = \"databricks_connect\"",
      ")"
    )
    code_lines <- code_lines[!is.null(code_lines)]
    ret <- ""
    if (cluster_id != "") {
      ret <- paste0(code_lines, collapse = "\n")
    }
    ret
  }

  code_reactive <- reactive({
    code_create(
      cluster_id = input$cluster_id,
      host_url = input$host_url,
      dbr = input$dbr_ver %||% dbr_version(),
      envname = input$py_env
    )
  })

  observe({
    rsApiUpdateDialog(code_reactive())
  })
}

shinyApp(connection_spark_ui, connection_spark_server)

function(df = mtcars) {
  library(arrow);
  fn <- function(...) 1;
  col_names <- c('am', 'x');
  fn_run <- fn(df);
  if(is.vector(fn_run)) {
    ret <- data.frame(x = fn_run);
  } else {
    ret <- as.data.frame(fn_run);
  };
  if(!is.null(col_names)) colnames(ret) <- col_names;
  ret
}

#' @export
spark_ide_objects.pyspark_connection <- function(
  con,
  catalog = NULL,
  schema = NULL,
  name = NULL,
  type = NULL
) {
  catalog_python(
    con = con,
    catalog = catalog,
    schema = schema,
    name = name,
    type = type
  )
}

#' @export
spark_ide_columns.pyspark_connection <- function(
  con,
  table = NULL,
  view = NULL,
  catalog = NULL,
  schema = NULL
) {
  tbl_df <- rs_get_table(con, catalog, schema, table)

  tbl_sample <- collect(head(tbl_df))

  tbl_info <- map_chr(tbl_sample, \(.x) paste0(rs_type(.x), " ", rs_vals(.x)))

  data.frame(
    name = names(tbl_info),
    type = tbl_info
  )
}

#' @export
spark_ide_preview.pyspark_connection <- function(
  con,
  rowLimit,
  table = NULL,
  view = NULL,
  catalog = NULL,
  schema = NULL
) {
  tbl_df <- rs_get_table(con, catalog, schema, table)
  collect(head(tbl_df, rowLimit))
}

catalog_python <- function(
  con,
  catalog = NULL,
  schema = NULL,
  name = NULL,
  type = NULL
) {
  df_catalogs <- data.frame()
  df_tables <- data.frame()

  limit <- as.numeric(
    Sys.getenv("SPARKLYR_CONNECTION_OBJECT_LIMIT", unset = NA)
  )
  sc_catalog <- python_conn(con)$catalog
  if (is.null(catalog) && is.null(schema)) {
    sql_catalogs <- con[["misc"]][["sql_catalogs"]] %||% "show catalogs"
    catalogs <- dbGetQuery(con, sql_catalogs)
    if (nrow(catalogs) > 0) {
      catalog_name <- ifelse(
        "catalog" %in% colnames(catalogs),
        "catalog",
        "name"
      )
      out <- data.frame(name = catalogs[[catalog_name]], type = "catalog")
    }
    if (!is.na(limit)) {
      out <- head(out, limit)
    }
  } else {
    if (is.null(schema)) {
      out <- rs_get_databases(con, limit, catalog)
    } else {
      if (is.null(catalog)) {
        sql_schema <- con[["misc"]][["sql_tables_schema"]] %||%
          "show tables in `{schema}`"
      } else {
        sql_schema <- con[["misc"]][["sql_tables_catalog_schema"]] %||%
          "show tables in `{catalog}`.`{schema}`"
      }
      tables <- dbGetQuery(con, glue(sql_schema))
      tables_name <- ifelse(
        "tableName" %in% colnames(tables),
        "tableName",
        "name"
      )
      out <- df_tables
      if (nrow(tables) > 0) {
        if ("isTemporary" %in% colnames(tables)) {
          tables <- tables[!tables$isTemporary, ]
        }
        if (nrow(tables) > 0) {
          out <- data.frame(
            name = tables[[tables_name]],
            schema = schema,
            type = "table"
          )
          out$catalog <- catalog
        }
      }
      if (!is.na(limit)) {
        out <- head(out, limit)
      }
    }
  }
  out
}

rs_get_databases <- function(con, limit = NA, catalog = NULL) {
  out <- data.frame()
  if (!is.null(catalog)) {
    sql_databases <- con[["misc"]][["sql_schemas_catalog"]] %||%
      "show databases in `{catalog}`"
  } else {
    sql_databases <- con[["misc"]][["sql_schemas"]] %||% "show databases"
  }
  databases <- dbGetQuery(con, glue(sql_databases))
  if (nrow(databases) > 0) {
    db_names <- databases$databaseName %||%
      databases$namespace %||%
      databases$name
    out <- data.frame(name = db_names, type = "schema")
    if (!is.na(limit)) {
      out <- head(out, limit)
    }
  }
  out
}

rs_get_table <- function(con, catalog = NULL, schema = NULL, table = NULL) {
  cq <- ifelse(is.null(con$quote), "`", con$quote)
  from <- c(catalog, schema, table)
  from <- paste0(cq, from, cq, collapse = ".")
  tbl(con, I(from))
}

rs_type <- function(x) {
  class <- class(x)[[1]]
  if (class == "integer") {
    class <- "int"
  }
  if (class == "numeric") {
    class <- "num"
  }
  if (class == "POSIXct") {
    class <- "dttm"
  }
  if (class == "character") {
    class <- "chr"
  }
  class
}

rs_vals <- function(x) {
  ln <- 30
  x <- paste0(x, collapse = " ")
  if (nchar(x) > ln) {
    x <- substr(x, 1, (ln - 3))
    x <- paste0(x, "...")
  }
  x
}

test_that("Object retrieval function work", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()

  expect_s3_class(
    spark_ide_preview.pyspark_connection(sc, table = "mtcars", rowLimit = 10),
    "data.frame"
  )

  expect_s3_class(spark_ide_objects(sc), "data.frame")

  expect_snapshot(
    spark_ide_columns(sc, table = "mtcars")
  )

  expect_snapshot(
    spark_ide_preview(sc, table = "mtcars", rowLimit = 10)
  )

  expect_s3_class(
    catalog_python(sc, catalog = "spark_catalog"),
    "data.frame"
  )

  expect_s3_class(
    catalog_python(sc, catalog = "spark_catalog", schema = "default"),
    "data.frame"
  )

  expect_s3_class(
    rs_get_databases(sc, limit = 10, catalog = "spark_catalog"),
    "data.frame"
  )
})

#' @export
head.tbl_pyspark <- function(x, n = 6L, ...) {
  if (!is.null(python_sdf(x))) {
    sdf <- tbl_pyspark_sdf(x)
    sdf_limit <- sdf$limit(as.integer(n))
    x <- python_obj_tbl_set(x, sdf_limit)
    NextMethod()
  } else {
    NextMethod()
  }
}

#' @export
sample_n.tbl_pyspark <- function(
  tbl,
  size,
  replace = FALSE,
  weight = NULL,
  .env = NULL,
  ...
) {
  slice_sample(
    .data = tbl,
    n = size,
    replace = replace,
    weight_by = !!enquo(weight)
  )
}

#' @export
sample_frac.tbl_pyspark <- function(
  tbl,
  size = 1,
  replace = FALSE,
  weight = NULL,
  .env = NULL,
  ...
) {
  weight <- enquo(weight)
  if (!quo_is_null(weight)) {
    abort("`weight` is not supported in this Spark connection")
  }
  df <- tbl_pyspark_sdf(tbl)
  sampled <- df$sample(fraction = size, withReplacement = TRUE)
  tbl_pyspark_temp(sampled, tbl)
}


#' @export
compute.tbl_pyspark <- function(x, name = NULL, ...) {
  cache_query(x, name = name, storage_level = "MEMORY_AND_DISK")
}

cache_query <- function(table, name = NULL, storage_level = "MEMORY_AND_DISK") {
  # https://spark.apache.org/docs/latest/sql-ref-syntax-aux-cache-cache-table.html
  if (is.null(name)) {
    name <- random_string()
  }
  x <- remote_query(table)
  query <- glue(
    "CACHE TABLE {name} OPTIONS ('storageLevel' '{storage_level}') {x}"
  )
  sc <- spark_connection(table)
  invoke(sc, "sql", query)
  spark_ide_connection_updated(sc, name)
  tbl(sc, name)
}

#' @export
collect.tbl_pyspark <- function(x, ...) {
  x |>
    tbl_pyspark_sdf() |>
    to_pandas_cleaned()
}

#' @export
spark_dataframe.tbl_pyspark <- function(x, ...) {
  conn <- x[[1]]
  query <- x[[2]]
  qry <- query |>
    sql_render(conn) |>
    query_cleanup(conn)
  invoke(conn, "sql", qry)
}

#' @export
sdf_copy_to.pyspark_connection <- function(
  sc,
  x,
  name = spark_table_name(substitute(x)),
  memory = TRUE,
  repartition = 0,
  overwrite = FALSE,
  struct_columns,
  ...
) {
  context <- python_conn(sc)
  col_names <- colnames(x)
  col_names <- gsub("\\.", "_", col_names)
  colnames(x) <- col_names
  schema <- NULL
  if (is_snowflake(sc)) {
    if (memory) {
      cli_abort(
        "Snowflake's Snowpark does not support `memory = TRUE` please use `memory = FALSE`"
      )
    }
    x <- as.list(x) |> transpose()
    schema <- col_names
  }
  if (!is_snowflake(sc) && memory) {
    if (context$catalog$tableExists(name)) {
      if (overwrite) {
        context$catalog$dropTempView(name)
      } else {
        cli_abort(
          "Temp table {name} already exists, use `overwrite = TRUE` to replace"
        )
      }
    }
  }
  df_copy <- context$createDataFrame(r_to_py(x), schema = schema)
  repartition <- as.integer(repartition)
  if (repartition > 0) {
    df_copy$createTempView(name)
    df_copy$repartition(repartition)
    if (memory && !sc$serverless) {
      storage_level <- import("pyspark.storagelevel")
      df_copy$persist(storage_level$StorageLevel$MEMORY_AND_DISK)
    }
    out <- tbl(src = sc, from = name)
  } else {
    out <- df_copy |>
      tbl_pyspark_temp(sc)
    if (memory && !sc$serverless) {
      out <- cache_query(table = out, name = name)
    }
  }

  spark_ide_connection_updated(sc, name)

  out
}

#' @export
tbl.pyspark_connection <- function(src, from, ...) {
  if (inherits(from, "AsIs")) {
    sql_from <- from
  } else {
    sql_from <- as.sql(from, con = src$con)
  }
  con <- python_conn(src)
  pyspark_obj <- con$table(sql_from)
  vars <- as.character(pyspark_obj$columns)
  src <- python_obj_con_set(src, pyspark_obj)
  out <- tbl_sql(
    subclass = "pyspark",
    src = src,
    from = from,
    vars = vars,
    ...
  )
  out_class <- class(out)
  new_class <- c(out_class[1], "tbl_spark", out_class[2:length(out_class)])
  class(out) <- new_class
  out
}

#' @export
tbl_ptype.tbl_pyspark <- function(.data) {
  abort("Predicates are not supported with this back-end")
}

#' @export
tidyselect_data_has_predicates.tbl_pyspark <- function(x) {
  FALSE
}

#' @export
same_src.pyspark_connection <- function(x, y) {
  identical(x$master, y$master) &&
    identical(x$method, y$method) &&
    identical(x$state, y$state)
}

tbl_pyspark_sdf <- function(x) {
  out <- python_sdf(x)
  if (is.null(out)) {
    con <- python_conn(x[[1]])
    qry <- x |>
      remote_query() |>
      query_cleanup(con)
    out <- con$sql(qry)
  }
  out
}

tbl_temp_name <- function() glue("{temp_prefix()}{random_string()}")

#' @export
sdf_register.spark_pyobj <- function(x, name = NULL) {
  # Attempting to cache a data frame with 0 columns returns an error.
  # So it returns nothing when this is the case (#110)
  x_obj <- python_obj_get(x)
  if (inherits(x_obj, "pyspark.sql.connect.dataframe.DataFrame")) {
    if (length(x_obj$columns) == 0) {
      return(invisible())
    }
  }
  sc <- spark_connection(x)
  tbl_pyspark_temp(
    x = x_obj,
    conn = sc,
    tmp_name = name
  )
}

python_sdf <- function(x) {
  pyobj <- python_obj_get(x)
  class_pyobj <- class(pyobj)
  name <- remote_name(x)
  out <- NULL
  if (!is.null(name) && any(grepl("dataframe", class_pyobj))) {
    out <- pyobj
  }
  out
}

python_obj_get <- function(x) {
  UseMethod("python_obj_get")
}

#' @export
python_obj_get.default <- function(x) {
  if (inherits(x, "character")) {
    return(x)
  }
  py_x <- try(get_spark_pyobj(x), silent = TRUE)
  if (!inherits(py_x, "try-error")) {
    return(py_x[["pyspark_obj"]])
  } else {
    sc <- spark_connection(x)
    if (inherits(sc$session, "python.builtin.object")) {
      return(sc$session)
    }
  }
}

#' @export
python_obj_get.python.builtin.object <- function(x) {
  x
}

#' @export
python_obj_get.spark_pyobj <- function(x) {
  x[["pyspark_obj"]]
}

#' @export
python_obj_get.ml_connect_model <- function(x) {
  x[["pipeline"]][["pyspark_obj"]]
}

#' @export
python_obj_get.ml_connect_pipeline <- function(x) {
  x[[".jobj"]][["pyspark_obj"]]
}

#' @export
python_obj_get.ml_connect_pipeline_model <- function(x) {
  x[[".jobj"]]
}

#' @export
python_obj_get.tbl_pyspark <- function(x) {
  x[["src"]][["session"]]
}


python_obj_con_set <- function(sc, obj) {
  sc$session <- obj
  sc
}

python_obj_tbl_set <- function(tbl, obj) {
  conn <- spark_connection(tbl)
  sc <- python_obj_con_set(conn, obj)
  tbl[[1]] <- sc
  tbl
}

tbl_pyspark_temp <- function(x, conn, tmp_name = NULL) {
  sc <- spark_connection(conn)
  if (is.null(tmp_name)) {
    tmp_name <- tbl_temp_name()
  }
  py_x <- python_obj_get(x)
  py_x$createOrReplaceTempView(tmp_name)
  if (!is_snowflake(sc)) {
    tmp_name <- sc |>
      dbQuoteIdentifier(tmp_name) |>
      as.character()
  }
  tbl(sc, I(tmp_name))
}

setOldClass(c("tbl_pyspark", "tbl_spark"))

#' @export
`[.tbl_pyspark` <- function(x, i) {
  # this is defined to match the interface to sparklyr::`[.tbl_spark`. But it really
  # should be more flexible, taking row specs, multiple args, etc. matching
  # semantics of R dataframes and take advantage of
  # reticulate::`[.python.builtin.object` for constructing slices, etc.
  if (is.null(i)) {
    # special case, since pyspark has no "emptyDataFrame" method to invoke
    sc <- spark_connection(x)

    pyspark.sql.types <- reticulate::import("pyspark.sql.types")
    ss <- x$src$state$spark_context # SparkSession obj
    edf <- ss$createDataFrame(list(), pyspark.sql.types$StructType(list()))

    tmp_name <- tbl_temp_name()
    edf$createOrReplaceTempView(tmp_name)
    return(tbl(sc, tmp_name))
  }
  NextMethod()
}

query_cleanup <- function(x, con) {
  if (is_snowflake(con)) {
    x <- gsub("`", "", x)
  }
  x
}

is_snowflake <- function(sc) {
  inherits(sc, "connect_snowflake") ||
    inherits(sc, "snowflake.snowpark.session.Session")
}

test_that("Pandas to R type conversion works", {
  sc <- use_test_spark_connect()

  sql_text <- "
    SELECT
    CAST(1 AS BYTE) AS byte_col,
    CAST(2 AS SHORT) AS short_col,
    CAST(3 AS INT) AS int_col,
    CAST(4 AS LONG) AS long_col,
    CAST(1.1 AS FLOAT) AS float_col,
    CAST(2.2 AS DOUBLE) AS double_col,
    CAST(123.45 AS DECIMAL(10,2)) AS decimal_col,
    TRUE AS boolean_col,
    CAST(NULL AS BOOLEAN) AS boolean_col_null,
    'hello' AS string_col,
    CAST('abc' AS BINARY) AS binary_col,
    DATE '2025-01-01' AS date_col,
    CAST(NULL AS DATE) AS date_col_null,
    TIMESTAMP '2025-01-01 12:34:56' AS timestamp_col,
    CAST(NULL AS TIMESTAMP) AS timestamp_col_null
  "

  sql_text_null <- "
    SELECT
    CAST(NULL AS BYTE),
    CAST(NULL AS SHORT),
    CAST(NULL AS INT),
    CAST(NULL AS LONG),
    CAST(NULL AS FLOAT),
    CAST(NULL AS DOUBLE),
    CAST(NULL AS DECIMAL(10,2)),
    CAST(NULL AS BOOLEAN),
    CAST(NULL AS BOOLEAN),
    CAST(NULL AS STRING),
    CAST(NULL AS BINARY),
    CAST(NULL AS DATE),
    CAST(NULL AS DATE),
    CAST(NULL AS TIMESTAMP),
    CAST(NULL AS TIMESTAMP)
  "

  tbl_result <- sparklyr::sdf_sql(
    sc = sc,
    sql = paste(
      sql_text,
      sql_text_null,
      sql_text,
      sql_text_null,
      sep = "UNION ALL"
    )
  ) |>
    python_obj_get() |>
    to_pandas_cleaned()

  expect_vector(tbl_result$byte_col, integer())
  expect_vector(tbl_result$short_col, integer())
  expect_vector(tbl_result$int_col, integer())
  expect_vector(tbl_result$long_col, numeric())
  expect_vector(tbl_result$float_col, numeric())
  expect_vector(tbl_result$double_col, numeric())
  expect_vector(tbl_result$decimal_col, numeric())
  expect_vector(tbl_result$boolean_col, logical())
  expect_vector(tbl_result$boolean_col_null, logical())
  expect_vector(tbl_result$string_col, character())
  expect_vector(tbl_result$binary_col, list())
  expect_vector(tbl_result$date_col, Sys.Date())
  expect_vector(tbl_result$date_col_null, Sys.Date())
  expect_vector(tbl_result$timestamp_col, Sys.time())
  expect_vector(tbl_result$timestamp_col_null, Sys.time())
})

# ------------------------- Prediction method ----------------------------------
#' @export
ml_predict.ml_connect_model <- function(x, dataset, ...) {
  transform_impl(x, dataset, prep = TRUE, remove = TRUE)
}

ml_get_last_item <- function(x) {
  classes <- x |>
    strsplit("\\.") |>
    unlist()

  classes[length(classes)]
}

transform_impl <- function(
  x,
  dataset,
  prep = TRUE,
  remove = FALSE,
  conn = NULL,
  as_df = TRUE
) {
  ml_installed()
  if (prep) {
    ml_df <- ml_prep_dataset(
      x = dataset,
      label = x$label,
      features = x$features,
      lf = "all"
    )
  } else {
    ml_df <- python_obj_get(dataset)
  }

  py_object <- python_obj_get(x)

  ret <- x |>
    get_spark_pyobj() |>
    invoke("transform", ml_df)

  if (remove) {
    if (inherits(x, "ml_connect_model")) {
      if (invoke(x, "hasParam", "labelCol")) {
        label_col <- invoke(x, "getLabelCol")
        ret <- invoke(ret, "drop", label_col)
      }
      features_col <- invoke(x, "getFeaturesCol")
      ret <- invoke(ret, "drop", features_col)
    } else {
      stages <- invoke(py_object, "stages")
      for (i in stages) {
        if (invoke(i, "hasParam", "inputCol")) {
          input_col <- invoke(i, "getInputCol")
          ret <- invoke(ret, "drop", input_col)
        } else {
          label_col <- invoke(i, "getLabelCol")
          features_col <- invoke(i, "getFeaturesCol")
          ret <- invoke(ret, "drop", label_col)
          ret <- invoke(ret, "drop", features_col)
        }
      }
    }
  }

  if (is.null(conn)) {
    conn <- spark_connection(dataset)
  }
  if (as_df) {
    ret <- tbl_pyspark_temp(
      x = ret,
      conn = conn
    )
  }
  ret
}

# --------------------- Print method for the models ----------------------------

#' @export
print.ml_connect_model <- function(x, ...) {
  model_name <- unlist(strsplit(py_repr(python_obj_get(x)), ":"))[[1]]
  cli_h2("MLib model: {model_name}")
  print_coefficients(x)
  print_summary(x)
}

print_parameters <- function(x) {
  p <- get_params(x)
  p_descriptions <- p |>
    map_chr(\(.x) .x$description) |>
    as.character() |>
    strsplit("\\.") |>
    map(\(.x) .x[[1]])
  p_names <- map_chr(p, \(.x) .x$name)
  p_values <- p |>
    map(\(.x) {
      cl <- class(.x$value)
      .x$value
    }) |>
    as.character()
  p_values[nchar(p_values) >= 20] <- paste0(
    substr(p_values[nchar(p_values) >= 20], 1, 17),
    "..."
  )
  out <- paste0(
    cli::col_blue(cli::symbol$square_small_filled),
    " ",
    cli::ansi_align(
      text = cli::style_hyperlink(paste0(p_names, ":"), p_descriptions),
      width = max(cli::ansi_nchar(p_names)) + 2
    ),
    cli::ansi_align(
      text = cli::col_silver(p_values),
      width = max(cli::ansi_nchar(p_values))
    )
  )
  two_col_print(out)
}

print_coefficients <- function(x) {
  py_x <- python_obj_get(x)
  coefs <- try(py_x[["coefficients"]], silent = TRUE)
  if (inherits(coefs, "try-error")) {
    return(invisible())
  }
  cli_h3("Coefficients:")
  coefs <- round(coefs["values"], 3)
  features <- x$features
  if (py_x$getFitIntercept()) {
    features <- c("Intercept", features)
    coefs <- c(round(py_x["intercept"], 3), coefs)
  }
  coefs <- as.character(coefs)
  coefs[coefs > 0] <- paste0(" ", coefs[coefs > 0])
  features <- paste0(features, ": ")

  out <- paste0(
    cli::col_blue(cli::symbol$square_small_filled),
    " ",
    cli::ansi_align(
      text = cli::col_silver(features),
      width = max(cli::ansi_nchar(features)) + 2
    ),
    cli::ansi_align(
      text = cli::col_silver(coefs),
      width = max(cli::ansi_nchar(coefs))
    )
  )
  two_col_print(out)
  cat("\n")
}

print_summary <- function(x) {
  py_x <- python_obj_get(x)
  has_summary <- try(py_x$hasSummary, silent = TRUE)
  if (inherits(has_summary, "try-error")) {
    has_summary <- FALSE
  }
  if (!has_summary) {
    print_parameters(x)
    return(invisible())
  }
  cli_h3("Summary:")
  py_x <- python_obj_get(x)
  summary_names <- names(py_x$summary)
  summary_values <- map(summary_names, function(x) py_x$summary[[x]])
  summary_valid <- summary_values |>
    map_lgl(function(x) inherits(x, "numeric") | inherits(x, "character"))
  summary_names <- summary_names[summary_valid]
  summary_values <- summary_values[summary_valid]
  summary_values <- summary_values |>
    map(function(x) {
      if (is.numeric(x)) {
        x <- round(x, 3)
      }
      x <- paste(x, collapse = ", ")
      if (nchar(x) > 30) {
        x <- glue("{substr(x, 1, 27)}...")
      }
      x
    })
  summary_names <- glue("{summary_names}: ")
  out <- paste0(
    cli::col_blue(cli::symbol$square_small_filled),
    " ",
    cli::ansi_align(
      text = cli::col_silver(summary_names),
      width = max(cli::ansi_nchar(summary_names)) + 1
    ),
    cli::ansi_align(
      text = cli::col_silver(summary_values),
      width = max(cli::ansi_nchar(summary_values))
    ),
    "\n"
  )
  walk(out, cat)
}

two_col_print <- function(x) {
  p_length <- length(x)
  is_odd <- p_length / 2 != floor(p_length / 2)
  if (is_odd) {
    x <- c(x, "")
    p_length <- p_length + 1
  }

  # p_sel <- rep(c(TRUE, FALSE), p_length)
  p_sel <- rep(c(TRUE, FALSE), 1, each = (p_length / 2))

  p_left <- x[p_sel]
  p_left <- p_left[!is.na(p_left)]
  p_right <- x[!p_sel]
  p_right <- p_right[!is.na(p_right)]

  p_two <- paste0(p_left, "   ", p_right)

  cat(paste0(p_two, collapse = "\n"))
}


#' @export
spark_jobj.ml_connect_model <- function(x, ...) {
  x$pipeline
}

sparklyr_reporter <- function() {
  MultiReporter$new(
    reporters = list(
      SummaryReporter$new(),
      PerformanceReporter$new()
    )
  )
}

PerformanceReporter <- R6::R6Class(
  "PerformanceReporter",
  inherit = Reporter,
  public = list(
    results = list(
      context = character(0),
      time = numeric(0)
    ),
    last_context = NA_character_,
    last_test = NA_character_,
    last_time = Sys.time(),
    last_test_time = 0,
    n_ok = 0,
    n_skip = 0,
    n_warn = 0,
    n_fail = 0,
    failures = c(),
    line = function(...) cat(paste0("\n", ...), file = self$out),
    initialize = function(file = getOption("testthat.output_file", stdout())) {
      if (is.character(file)) {
        file <- normalizePath(file, mustWork = FALSE)
      }
      self$out <- file
      if (is.character(self$out) && file.exists(self$out)) {
        # If writing to a file, overwrite it if it exists
        file.remove(self$out)
      }

      # Capture at init so not affected by test settings
      self$width <- cli::console_width()
      self$unicode <- cli::is_utf8_output()
      self$crayon <- crayon::has_color()

      testthat_msg <- Sys.getenv("TESTTHAT_MSG")
      if (testthat_msg != "") self$line(testthat_msg)
    },
    start_context = function(context) {
      self$last_context <- context
      self$last_time <- Sys.time()
    },
    add_result = function(context, test, result) {
      elapsed_time <- as.numeric(Sys.time()) - as.numeric(self$last_time)

      is_error <- inherits(result, "expectation_failure") ||
        inherits(result, "expectation_error")

      if (is_error) {
        self$n_fail <- self$n_fail + 1
        self$failures <- c(
          self$failures,
          paste0(test, " (Context: ", context, ")")
        )
      } else if (inherits(result, "expectation_skip")) {
        self$n_skip <- self$n_skip + 1
      } else if (inherits(result, "expectation_warning")) {
        self$n_warn <- self$n_warn + 1
      } else {
        self$n_ok <- self$n_ok + 1
      }

      if (identical(self$last_test, test)) {
        elapsed_time <- self$last_test_time + elapsed_time
        self$results$time[length(self$results$time)] <- elapsed_time
        self$last_test_time <- elapsed_time
      } else {
        self$results$context[
          length(self$results$context) + 1
        ] <- self$last_context
        self$results$time[length(self$results$time) + 1] <- elapsed_time
        self$last_test_time <- elapsed_time
      }

      self$last_test <- test
      self$last_time <- Sys.time()
    },
    end_reporter = function() {
      cat("\n")
      data <- data.frame(
        context = self$results$context,
        time = self$results$time
      )

      summary <- data |>
        dplyr::group_by(context) |>
        dplyr::summarise(time = sum(time)) |>
        dplyr::mutate(
          time = format(time, width = "9", digits = "3", scientific = F)
        )

      total <- data |>
        dplyr::summarise(time = sum(time)) |>
        dplyr::mutate(time = format(time, digits = "3", scientific = F)) |>
        dplyr::pull()

      cat("\n")
      cat("--- Performance Summary  ----\n\n")
      print(as.data.frame(summary), row.names = FALSE)

      cat(paste0("\nTotal: ", total, "s\n"))

      cat("\n")
      cat("------- Tests Summary -------\n\n")
      self$cat_line("OK:       ", format(self$n_ok, width = 5))
      self$cat_line("Failed:   ", format(self$n_fail, width = 5))
      self$cat_line("Warnings: ", format(self$n_warn, width = 5))
      self$cat_line("Skipped:  ", format(self$n_skip, width = 5))
      if (length(self$failures) > 0) {
        self$cat_line(
          "Failures:  ",
          do.call(paste, as.list(c(self$failures, sep = "\n")))
        )
      }
      cat("\n")
      if (self$n_fail > 0) stop("There were failures")
    }
  )
)

#' @export
ml_pipeline.pyspark_connection <- function(x, ..., uid = NULL) {
  ml_installed()
  if (spark_version(x) > "4.0") {
    ml <- import("pyspark.ml")
  } else {
    ml <- import("pyspark.ml.connect")
  }
  pipeline <- ml$Pipeline()
  pipeline |>
    as_spark_pyobj(x) |>
    as_pipeline()
}

#' @export
ml_fit.ml_connect_pipeline <- function(x, dataset, ...) {
  fitted <- ml_fit_impl(x, dataset)
  sc <- spark_connection(x)
  stages <- map(invoke(fitted, "stages"), ml_print_params, sc)
  as_pipeline_model(fitted, stages)
}

ml_fit_impl <- function(x, dataset) {
  ml_installed()
  py_dataset <- python_obj_get(dataset)
  py_x <- get_spark_pyobj(x)

  fitted <- try(
    invoke(py_x, "fit", py_dataset),
    silent = TRUE
  )

  if (inherits(fitted, "try-error")) {
    py_error <- reticulate::py_last_error()
    rlang::abort(
      paste(connection_label(x), "error:"),
      body = fitted
    )
  }

  fitted
}


as_pipeline_model <- function(jobj, stages) {
  structure(
    list(
      uid = invoke(jobj, "uid"),
      param_map = list(),
      stages = stages,
      .jobj = jobj
    ),
    class = c(
      "ml_connect_pipeline_model",
      "ml_pipeline_model",
      "ml_transformer",
      "ml_connect_transformer",
      "ml_connect_pipeline_stage",
      "ml_pipeline_stage"
    )
  )
}

#' @export
ml_transform.ml_connect_pipeline_model <- function(x, dataset, ...) {
  transform_impl(x, dataset, prep = FALSE, remove = TRUE)
}

ml_print_params <- function(x, sc = NULL) {
  py_x <- python_obj_get(x)
  if (is.null(sc)) {
    x_params <- x$param_map
  } else {
    x_params <- x |>
      as_spark_pyobj(sc) |>
      ml_get_params()
  }
  x_params <- x_params |>
    map_chr(paste, collapse = ", ") |>
    imap_chr(function(x, y) glue("{y}: {x}")) |>
    paste0(collapse = "\n")
  name_label <- paste0("<", capture.output(py_x), ">")
  class_1 <- ml_get_last_item(class(py_x)[[1]])
  class_2 <- ml_get_last_item(class(py_x)[[3]])
  ret <- paste0(x_params, collapse = "\n")
  ret <- paste0(
    class_1,
    " (",
    class_2,
    ")",
    "\n",
    name_label,
    "\n",
    "(Parameters)\n",
    ret
  )
  class(ret) <- "ml_output_params"
  ret
}

#' @export
print.ml_output_params <- function(x, ...) {
  cat(x)
}

ml_connect_add_stage <- function(x, stage) {
  stages <- c(x$param_map, list(stage))
  py_stages <- NULL
  for (i in stages) {
    py_stages <- c(py_stages, python_obj_get(i))
  }
  pipeline <- x |>
    get_spark_pyobj() |>
    invoke("setStages", py_stages)
  outputs <- c(x$stages, list(ml_print_params(stage)))
  as_pipeline(pipeline, stages, outputs, TRUE)
}

as_pipeline <- function(
  jobj,
  stages = list(),
  outputs = list(),
  get_uid = FALSE
) {
  if (get_uid) {
    uid <- invoke(jobj, "uid")
  } else {
    uid <- "[Not initialized]"
  }
  structure(
    list(
      uid = uid,
      param_map = stages,
      .jobj = jobj,
      stages = outputs
    ),
    class = c(
      "ml_connect_pipeline",
      "ml_pipeline",
      "ml_connect_estimator",
      "ml_estimator",
      "ml_connect_pipeline_stage",
      "ml_pipeline_stage"
    )
  )
}

#' @export
ml_save.ml_connect_pipeline_stage <- function(x, path, overwrite = FALSE, ...) {
  ml_installed()
  path <- path_abs(path)

  py_obj <- get_spark_pyobj(x)

  version <- py_obj |>
    spark_connection() |>
    spark_version()

  if (version >= "4.0") {
    model <- python_obj_get(py_obj)
    model$save(path)
  } else {
    x |>
      spark_jobj() |>
      invoke(
        method = "saveToLocal",
        path = path,
        overwrite = overwrite
      )
  }
  return(invisible())
}

# TODO: export ml_load() in sparklyr as S3 method
ml_connect_load <- function(sc, path) {
  ml_installed()
  path <- path_abs(path)
  conn <- python_obj_get(sc)
  if (spark_version(sc) >= "4.0") {
    connect_pipeline <- import("pyspark.ml")
    pipeline <- connect_pipeline$Pipeline$load(path)
  } else {
    connect_pipeline <- import("pyspark.ml.connect.pipeline")
    pipeline <- connect_pipeline$Pipeline$loadFromLocal(path)
  }

  class_current <- ml_get_last_item(class(pipeline)[[1]])
  if (class_current == "Pipeline") {
    outputs <- map(pipeline$getStages(), ml_print_params, sc)
    ret <- as_pipeline(pipeline, outputs, get_uid = TRUE)
  }
  if (class_current == "PipelineModel") {
    outputs <- map(pipeline$stages, ml_print_params, sc)
    ret <- as_pipeline_model(pipeline, outputs)
  }
  ret
}

use_test_pull <- function(x, table = FALSE) {
  x <- dplyr::pull(x)
  if (table) {
    x <- table(x)
  }

  # Handle Spark ML vectors that come through as structured data frames
  # (Spark 4.1+ with Pandas 3.0+ converts ML vectors to data frames with type, size, indices, values columns)
  if (is.data.frame(x) && all(c("type", "size", "indices", "values") %in% names(x))) {
    # Extract the values column which contains the actual vector data
    x <- data.frame(
      x = map_chr(x$values, function(vec) {
        if (is.null(vec) || length(vec) == 0) {
          ""
        } else {
          paste(as.vector(vec), collapse = ", ")
        }
      })
    )
    return(x)
  }

  if (inherits(x[[1]], "array")) {
    x <- as.double(map(x, as.vector))
  }
  if (inherits(x[[1]], "pyspark.ml.linalg.SparseVector")) {
    x <- data.frame(
      x = map_chr(x, function(x) paste(as.vector(x$values), collapse = ", "))
    )
  }
  if (inherits(x[[1]], "pyspark.ml.linalg.DenseVector")) {
    x <- data.frame(
      x = map_chr(x, function(x) paste(as.vector(x$array), collapse = ", "))
    )
  }
  # Handle vectors that have been converted to numeric during pandas conversion
  # (Spark 4.1+ with Pandas 3.0+ converts ML vectors to numeric vectors)
  if (is.list(x) && length(x) > 0 && is.numeric(x[[1]]) && !is.data.frame(x)) {
    x <- data.frame(
      x = map_chr(x, function(vec) paste(as.vector(vec), collapse = ", "))
    )
  }
  x
}

use_test_iris_va <- function() {
  use_test_table_iris() |>
    ft_vector_assembler(
      input_cols = c("Sepal_Length", "Sepal_Width", "Petal_Length"),
      output_col = "vec_x"
    )
}

use_test_mtcars_va <- function() {
  use_test_table_mtcars() |>
    ft_vector_assembler(
      input_cols = c("mpg", "wt", "cyl"),
      output_col = "vec_x"
    )
}


# ------------------------------ Tune Grid -------------------------------------

use_tune_grid <- function() {
  if (is.null(.test_env$tune_grid)) {
    out <- list()
    set.seed(1010110)
    object <- parsnip::decision_tree(tree_depth = tune()) |>
      parsnip::set_mode("classification") |>
      parsnip::set_engine("rpart")
    preprocessor <- recipes::recipe(Species ~ ., data = iris) |>
      recipes::step_normalize(recipes::all_predictors()) |>
      recipes::step_pca(
        recipes::all_numeric_predictors(),
        num_comp = tune::tune()
      )
    resamples <- rsample::bootstraps(iris, 5)
    wf_grid <- dials::grid_regular(
      dials::num_comp(c(1L, 15L)),
      dials::tree_depth()
    )
    cntrl <- tune::control_grid(verbose = TRUE, save_pred = TRUE)
    results <- tune::tune_grid(
      object = object,
      preprocessor = preprocessor,
      resamples = resamples,
      grid = wf_grid
    )
    post <- tailor::tailor() |>
      tailor::adjust_probability_calibration()
    out$object <- object
    out$preprocessor <- preprocessor
    out$resamples <- resamples
    out$grid <- wf_grid
    out$control <- cntrl
    out$post <- post
    out$results <- results
    .test_env$tune_grid <- out
  }
  .test_env$tune_grid
}

expect_equal_preds <- function(object, expected) {
  act <- quasi_label(enquo(object))
  act2 <- quasi_label(enquo(expected))
  res <- compare_pred_class(object, expected)
  res_n <- nrow(res)
  if (res_n > 0) {
    fail(
      glue::glue(
        "{act$lab} predictions do not match {act2$lab}. There were {res_n} differences."
      )
    )
  } else {
    pass()
  }
  invisible(act$val)
}

compare_pred_class <- function(x, y) {
  res <- NULL
  pred_size <- length(x$.predictions)
  for (i in seq_len(pred_size)) {
    x_i <- x$.predictions[[i]]
    y_i <- y$.predictions[[i]]
    x_preds <- dplyr::select(x_i, .config, .row, x_class = .pred_class)
    y_preds <- dplyr::select(y_i, .config, .row, y_class = .pred_class)
    res <- dplyr::left_join(x_preds, y_preds, by = c(".config", ".row")) |>
      dplyr::filter(x_class != y_class) |>
      dplyr::mutate(split = i) |>
      dplyr::bind_rows(res)
  }
  res
}

#' Creates the 'label' and 'features' columns
#'
#' @details
#' At this time, 'Spark ML Connect', does not include a Vector Assembler
#' transformer. The main thing that this function does, is create a 'Pyspark'
#' array column. Pipelines require a 'label' and 'features' columns. Even though
#' it is is single column in the dataset, the 'features' column will contain all
#' of the predictors insde an array. This function also creates a new 'label'
#' column that copies the outcome variable. This makes it a lot easier to remove
#' the 'label', and 'outcome' columns.
#'
#' @returns A `tbl_pyspark`, with either the original columns from `x`, plus the
#' 'label' and 'features' column, or, the 'label' and 'features' columns only.
#'
#' @param x A `tbl_pyspark` object
#' @param formula Used when \code{x} is a \code{tbl_spark}. R formula.
#' @param label The name of the label column.
#' @param features The name(s) of the feature columns as a character vector.
#' @param features_col Features column name, as a length-one character vector.
#' @param label_col Label column name, as a length-one character vector.
#' @param keep_original Boolean flag that indicates if the output will contain,
#' or not, the original columns from `x`. Defaults to `TRUE`.
#' @param ... Added for backwards compatibility. Not in use today.
#' @export
ml_prepare_dataset <- function(
  x,
  formula = NULL,
  label = NULL,
  features = NULL,
  label_col = "label",
  features_col = "features",
  keep_original = TRUE,
  ...
) {
  if (keep_original) {
    lf <- "all"
  } else {
    lf <- "only"
  }

  prep <- ml_prep_dataset(
    x = x,
    formula = formula,
    features = features,
    label = label,
    features_col = features_col,
    label_col = label_col,
    lf = lf
  )

  tbl_pyspark_temp(prep, spark_connection(x))
}

ml_prep_dataset <- function(
  x,
  formula = NULL,
  label = NULL,
  features = NULL,
  label_col = "label",
  features_col = "features",
  lf = c("only", "all"),
  additional = NULL,
  fitting = FALSE
) {
  lf <- match.arg(lf)
  pyspark <- import("pyspark")
  if (!is.null(formula)) {
    f <- ml_formula(formula, x)
    features <- f$features
    label <- f$label
  } else {
    if (is.null(features) && is.null(label)) {
      return(x)
    }
  }
  ret <- python_obj_get(x)
  if (spark_version(spark_connection(x)) >= "4.0.0") {
    va <- pyspark$ml$feature$VectorAssembler()
    vector_assembler <- va$copy()
    vector_assembler$setOutputCol(features_col)
    if (fitting && length(features) == 1) {
      features <- list(features)
    }
    vector_assembler$setInputCols(features)
    ret <- vector_assembler$transform(ret)
  } else {
    features_array <- pyspark$sql$functions$array(features)
    ret <- ret$withColumn(features_col, features_array)
  }
  if (!is.null(label) && !is.null(label_col)) {
    ret <- ret$withColumn(label_col, ret[label])
  }
  select_cols <- c(features_col, additional)
  if (lf == "only") {
    if (!is.null(label_col)) {
      select_cols <- c(label_col, select_cols)
      if (!is.null(label)) {
        attr(ret, "label") <- label
      }
    }
    ret <- ret$select(select_cols)
    attr(ret, "features") <- features
  }
  ret
}

#' Starts and stops Spark Connect locally
#' @param version Spark version to use (3.4 or above)
#' @param scala_version Acceptable Scala version of packages to be loaded
#' @param include_args Flag that indicates whether to add the additional arguments
#' to the command that starts the service. At this time, only the 'packages'
#' argument is submitted.
#' @param python_version Python version to use if a temporary Python environment
#' will be created
#' @param python Path to the Python executable to use while running (optional)
#' @param additional_args Vector of additional arguments to use when starting
#' the service
#' @param ... Optional arguments; currently unused
#' @returns It returns messages to the console with the status of starting, and
#' stopping the local Spark Connect service.
#' @export
spark_connect_service_start <- function(
  version = "4.0",
  scala_version = "2.13",
  python_version = NULL,
  python = NULL,
  include_args = TRUE,
  additional_args = NULL,
  ...
) {
  get_version <- spark_install_find(version = version)
  cmd <- path(get_version$sparkVersionDir, "sbin", "start-connect-server.sh")
  args <- c(
    "--packages",
    glue(
      "org.apache.spark:spark-connect_{scala_version}:{get_version$sparkVersion}"
    ),
    additional_args
  )
  if (!include_args) {
    args <- ""
  }
  cli_div(theme = cli_colors())
  cli_text("{.header Starting {.emph Spark Connect} locally ...}")
  java_version <- try(
    system2("java", "-version", stdout = TRUE, stderr = TRUE),
    silent = TRUE
  )
  if (!inherits(java_version, "try-error")) {
    out_java <- setNames(java_version, rep("", times = length(java_version)))
    cli_bullets(out_java)
  }
  if (is.null(python)) {
    envname <- use_envname(
      backend = "pyspark",
      main_library = "pyspark",
      version = version,
      python_version = python_version,
      messages = TRUE,
      ask_if_not_installed = FALSE
    )
    py_require("rpy2")
    invisible(import_check("pyspark", envname))
    python <- py_exe()
  }
  withr::with_envvar(
    new = c(
      "PYSPARK_PYTHON" = python,
      "PYTHON_VERSION_MISMATCH" = python,
      "PYSPARK_DRIVER_PYTHON" = python
    ),
    {
      prs <- process$new(
        command = cmd,
        args = args,
        stdout = "|",
        stderr = "|",
        stdin = "|"
      )
    }
  )

  # Wait briefly to see if there's immediate output or errors
  # Don't use read_all_output() as it blocks until process closes stdout
  # Spark Connect is a long-running service, so we only check initial output
  prs$poll_io(timeout = 2000)  # Wait max 2 seconds

  # Read only what's available, don't wait for EOF
  if (prs$is_alive()) {
    output <- prs$read_output_lines()
    if (length(output) > 0) {
      cli_bullets(c(" " = "{.info {paste(output, collapse = '\n')}}"))
    }
    error <- prs$read_error_lines()
    if (length(error) > 0 && any(nzchar(error))) {
      cli_alert_warning(paste(error, collapse = "\n"))
    }
  } else {
    # Process exited immediately, likely an error
    if (prs$get_exit_status() != 0) {
      error <- prs$read_all_error()
      cli_abort(c("Failed to start Spark Connect service", "x" = error))
    }
  }

  # Store the process for potential cleanup
  pysparklyr_env$spark_connect_process <- prs

  cli_end()
  invisible()
}

#' @rdname spark_connect_service_start
#' @export
spark_connect_service_stop <- function(version = "4.0", ...) {
  get_version <- spark_install_find(version = version)
  cmd <- path(get_version$sparkVersionDir, "sbin", "stop-connect-server.sh")
  cli_div(theme = cli_colors())
  cli_h3("{.header Stopping {.emph Spark Connect}}")
  prs <- process$new(
    command = cmd,
    stdout = "|",
    stderr = "|",
    stdin = "|"
  )

  # Wait for shutdown command to complete
  prs$wait(timeout = 5000)  # Wait max 5 seconds

  cli_bullets(c(" " = "{.info - Shutdown command sent}"))

  # Clean up stored process reference
  if (!is.null(pysparklyr_env$spark_connect_process)) {
    pysparklyr_env$spark_connect_process <- NULL
  }

  cli_end()
  invisible()
}

# This file is part of the standard setup for testthat.
# It is recommended that you do not modify it.
#
# Where should you do additional test configuration?
# Learn more about the roles of various files in:
# * https://r-pkgs.org/testing-design.html#sec-tests-files-overview
# * https://testthat.r-lib.org/articles/special-files.html

# Sys.setenv("CODE_COVERAGE" = "true")
# Sys.setenv("SPARK_VERSION" = "4.1.1"); Sys.setenv("SCALA_VERSION" = "2.13"); Sys.setenv("PYTHON_VERSION" = "3.13")
# Sys.setenv("SPARK_VERSION" = "3.5.7"); Sys.setenv("SCALA_VERSION" = "2.12"); Sys.setenv("PYTHON_VERSION" = "3.10")
if (identical(Sys.getenv("CODE_COVERAGE"), "true")) {
  library(testthat)
  library(pysparklyr)

  test_check("pysparklyr")
}

#' @export
invoke.pyspark_connection <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = python_conn(jobj),
    method = method,
    ... = ...
  )
}

#' @export
invoke.ml_connect_model <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ... = ...
  )
}


#' @export
invoke.ml_connect_transformer <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ... = ...
  )
}

#' @export
invoke.ml_connect_estimator <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ... = ...
  )
}

#' @export
invoke.tbl_pyspark <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ... = ...
  )
}

#' @export
invoke.python.builtin.object <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ... = ...
  )
}

#' @export
invoke.spark_pyobj <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ...
  )
}

invoke_simple <- function(jobj, method, ...) {
  invoke_conn(
    jobj = jobj,
    context = jobj,
    method = method,
    ...
  )
}

invoke_conn <- function(jobj, context, method, ...) {
  py_method <- python_obj_get(method)
  py_context <- python_obj_get(context)

  x <- py_get_attr(py_context, py_method)
  out <- NULL
  if (inherits(x, "python.builtin.method")) {
    run_x <- py_call(x, ...)

    if (
      inherits(run_x, "numpy.number") |
        inherits(run_x, "python.builtin.str") |
        inherits(run_x, "python.builtin.bool") |
        inherits(run_x, "python.builtin.int")
    ) {
      out <- py_to_r(run_x)
    }

    if (is.null(out)) {
      conn <- spark_connection(jobj)
      if (is.null(conn)) {
        stop("invoke cannot have a NULL connection")
      }
      out <- as_spark_pyobj(run_x, conn)
    }
  }

  if (is.null(out)) {
    out <- py_to_r(x)
  }

  out
}

#' @export
invoke_static.pyspark_connection <- function(sc, class, method, ...) {
  list()
}

spark_session_add_file <- function(x, sc, file_name = NULL) {
  # Tracks in a environment variable which file names have already been
  # uploaded to avoid doing it again. This is why the hashed values of
  # the object in `x` is preferred, which ensure proper tracking of the
  # actual objects that have been uploaded
  connection_id <- sc[["connection_id"]]
  artifacts <- pysparklyr_env[["artifacts"]][[connection_id]]
  if (is.null(file_name)) {
    file_name <- rlang::hash(x)
  }
  files <- artifacts[["files"]]
  if (!file_name %in% artifacts[["files"]]) {
    temp_file <- path_temp(file_name, ext = "rds")
    saveRDS(x, temp_file)
    sc_obj <- spark_session(sc)
    sc_obj$addArtifact(temp_file, file = TRUE)
    file_delete(temp_file)
    pysparklyr_env[["artifacts"]][[connection_id]]$files <- c(file_name, files)
  }
  invisible()
}

#' A Shiny app that can be used to construct a \code{spark_connect} statement
#'
#' @export
#' @keywords internal
#' @returns A Shiny app
connection_databricks_shinyapp <- function() {
  if (!"shiny" %in% installed.packages()) {
    install_shiny <- showQuestion(
      "Shiny Required",
      "The 'shiny' package is not installed, install?",
      ok = "Install"
    )
    if (identical(install_shiny, TRUE)) {
      install_command <- get_wrapper("install.packages")
      install_command("shiny")
    }

    if (!"shiny" %in% installed.packages()) {
      cli_abort(
        "The `shiny` package is not installed, please install and retry."
      )
    }
  }

  local_dir <- "inst/rstudio/shinycon"
  if (file.exists(local_dir)) {
    app_dir <- local_dir
  } else {
    app_dir <- system.file("rstudio/shinycon", package = "pysparklyr")
  }
  shinyAppDir <- get_wrapper("shinyAppDir", envir = asNamespace("shiny"))
  shinyAppDir(app_dir)
}

get_wrapper <- function(x, pos = -1, envir = as.environment(pos)) {
  get(
    x = x,
    pos = pos,
    envir = envir
  )
}

test_that("Snippet caller works", {
  local_mocked_bindings(
    showQuestion = function(...) FALSE,
    get_wrapper = function(...) {
      ret <- function(x) x
      return(ret)
    }
  )
  x <- connection_databricks_shinyapp()
  expect_equal(path_file(x), "shinycon")
})

test_that("Snippet caller works", {
  local_mocked_bindings(
    showQuestion = function(...) TRUE,
    get_wrapper = function(...) {
      ret <- function(x) x
      return(ret)
    },
    installed.packages = function(...) c()
  )
  expect_error(
    connection_databricks_shinyapp(),
    "The `shiny` package is not installed, please install and retry"
  )
})

test_that("Get wrapper works", {
  expect_true(inherits(get_wrapper("mean"), "function"))
})

# This is to enable mocked tests, and the fact that rsconnect is
# not being directly imported by this package
shhhh <- function(x) {
  suppressPackageStartupMessages(requireNamespace(x, quietly = TRUE))
}
is_installed <- function(pkg) {
  res <- try(shhhh(pkg), silent = TRUE)
  res
}
rsconnect_accounts <- if (is_installed("rsconnect")) rsconnect::accounts
rsconnect_deployments <- if (is_installed("rsconnect")) rsconnect::deployments
rsconnect_deployApp <- if (is_installed("rsconnect")) rsconnect::deployApp

#' Deploys Databricks backed content to publishing server
#'
#' @description
#'  This is a convenience function that is meant to make it easier for
#' you to publish your Databricks backed content to a publishing server. It is
#' meant to be primarily used with Posit Connect.
#'
#' @param appDir A directory containing an application (e.g. a Shiny app or plumber API)
#' Defaults to NULL. If left NULL, and if called within RStudio, it will attempt
#' to use the folder of the currently opened document within the IDE. If there are
#' no opened documents, or not working in the RStudio IDE, then it will use
#' `getwd()` as the default value.
#' @param python Full path to a python binary for use by `reticulate.` It defaults to NULL.
#' If left NULL, this function will attempt to find a viable local Python
#' environment to replicate using the following hierarchy:
#' 1. `version` - Cluster's DBR version
#' 2. `cluster_id` - Query the cluster to obtain its DBR version
#' 3. If one is loaded in the current R session, it will verify that the Python
#' environment is suited to be used as the one to use
#' @param account The name of the account to use to publish
#' @param server The name of the target server to publish
#' @param lint Lint the project before initiating the project? Default to FALSE.
#' It has been causing issues for this type of content.
#' @param version The Databricks Runtime (DBR) version. Use if `python` is NULL.
#' @param cluster_id The Databricks cluster ID. Use if `python`, and `version` are
#' NULL
#' @param host The Databricks host URL. Defaults to NULL. If left NULL, it will
#' use the environment variable `DATABRICKS_HOST`
#' @param token The Databricks authentication token. Defaults to NULL. If left NULL, it will
#' use the environment variable `DATABRICKS_TOKEN`
#' @param forceGeneratePythonEnvironment If an existing requirements.txt file is found,
#' it will be overwritten when this argument is TRUE.
#' @param confirm Should the user be prompted to confirm that the correct
#' information is being used for deployment? Defaults to `interactive()`
#' @param ... Additional named arguments passed to `rsconnect::deployApp()` function
#' @returns No value is returned to R. Only output to the console.
#' @export
deploy_databricks <- function(
  appDir = NULL,
  python = NULL,
  account = NULL,
  server = NULL,
  lint = FALSE,
  forceGeneratePythonEnvironment = TRUE,
  version = NULL,
  cluster_id = NULL,
  host = NULL,
  token = NULL,
  confirm = interactive(),
  ...
) {
  if (!is_installed("rsconnect")) {
    cli_abort(
      paste(
        "The package {.pkg rsconnect} is needed for publication but it is not",
        "installed. Use {.run install.packages(\"rsconnect\")} to install and",
        "then try again."
      )
    )
  }

  cli_div(theme = cli_colors())

  cluster_id <- cluster_id %||% Sys.getenv("DATABRICKS_CLUSTER_ID")

  if (is.null(version) && !is.null(cluster_id)) {
    version <- databricks_dbr_version(
      cluster_id = cluster_id,
      host = databricks_host(host),
      token = databricks_token(token)
    )
  }
  env_vars <- NULL
  env_var_message <- NULL
  var_error <- NULL

  # Host URL
  if (!is.null(host)) {
    env_vars <- c("CONNECT_DATABRICKS_HOST" = host)
  } else {
    env_host <- Sys.getenv("DATABRICKS_HOST", unset = "")
    if (env_host != "") {
      env_vars <- c(env_vars, "DATABRICKS_HOST" = env_host)
      host <- env_host
    }
  }
  if (!is.null(host)) {
    env_var_message <- c("i" = paste0("{.header Host URL:} ", host))
  } else {
    var_error <- c(
      " " = paste0(
        "{.header - No host URL was provided or found. Please either set the}",
        " {.emph 'DATABRICKS_HOST'} {.header environment variable,}",
        " {.header or pass the }{.code host} {.header argument.}"
      )
    )
  }

  # Token
  if (!is.null(token)) {
    env_vars <- c(env_vars, "CONNECT_DATABRICKS_TOKEN" = token)
  } else {
    env_token <- Sys.getenv("DATABRICKS_TOKEN", unset = "")
    if (env_token != "") {
      env_vars <- c(env_vars, "DATABRICKS_TOKEN" = env_token)
      token <- env_token
    }
  }
  if (!is.null(token)) {
    env_var_message <- c(
      env_var_message,
      " " = "{.header Token:} '<REDACTED>'"
    )
  }

  if (!is.null(var_error)) {
    cli_abort(c("Cluster setup errors:", var_error), call = NULL)
  }

  withr::with_envvar(
    new = env_vars,
    {
      deploy(
        appDir = appDir,
        lint = lint,
        python = python,
        version = version,
        backend = "databricks",
        main_library = "databricks-connect",
        envVars = names(env_vars),
        env_var_message = env_var_message,
        account = account,
        server = server,
        confirm = confirm,
        ...
      )
    }
  )
}

deploy <- function(
  appDir = NULL,
  account = NULL,
  server = NULL,
  lint = FALSE,
  envVars = NULL,
  python = NULL,
  version = NULL,
  backend = NULL,
  main_library = NULL,
  env_var_message = NULL,
  confirm,
  ...
) {
  if (is.null(backend)) {
    abort("'backend' is empty, please provide one")
  }
  rs_accounts <- rsconnect_accounts()
  accts_msg <- NULL
  if (nrow(rs_accounts) == 0) {
    abort("There are no server accounts setup")
  } else {
    if (is.null(account)) {
      account <- rs_accounts$name[1]
    }
    if (is.null(server)) {
      server <- rs_accounts$server[1]
    }
    if (nrow(rs_accounts > 1)) {
      accts_msg <- "Change 'Posit server'"
    }
  }
  cli_div(theme = cli_colors())
  cli_h1("Starting deployment")
  editor_doc <- NULL
  if (is.null(appDir)) {
    if (check_interactive() && check_rstudio()) {
      editor_doc <- getSourceEditorContext()
      if (!is.null(editor_doc)) {
        appDir <- dirname(editor_doc$path)
      }
    }
  }
  if (is.null(appDir)) {
    appDir <- getwd()
  }
  appDir <- path(appDir)

  cli_alert_info("{.header Source directory: {.emph {appDir}}}")
  python <- deploy_find_environment(
    python = python,
    version = version,
    backend = backend,
    main_library = main_library
  )
  cli_inform(c(
    "i" = "{.header Posit server:} {.emph {server}}",
    " " = "{.header Account name:} {.emph {account}}"
  ))
  if (!is.null(env_var_message)) {
    cli_bullets(env_var_message)
  }
  cli_inform("")
  if (check_interactive() && confirm) {
    cli_inform("{.header Does everything look correct?}")
    choice <- menu(choices = c("Yes", "No", accts_msg))
    if (choice == 2) {
      return(invisible())
    }
    if (choice == 3) {
      chr_accounts <- rs_accounts |>
        transpose() |>
        map_chr(\(.x) glue("Server: {.x$server} | Account: {.x$name}"))
      choice <- menu(title = "Select publishing target:", chr_accounts)
      server <- rs_accounts$server[choice]
      account <- rs_accounts$name[choice]
    }

    req_file <- path(appDir, "requirements.txt")
    prev_deployments <- rsconnect_deployments(appDir)
    if (
      !file_exists(req_file) &&
        nrow(prev_deployments) == 0 &&
        check_interactive()
    ) {
      cli_inform(c(
        "{.header Would you like to create the 'requirements.txt' file?}",
        "{.class Why consider? This will allow you to skip using `version` or `cluster_id`}"
      ))
      choice <- menu(choices = c("Yes", "No"))
      if (choice == 1) {
        requirements_write(
          destfile = req_file,
          python = python
        )
      }
    }
  }

  rsconnect_deployApp(
    appDir = appDir,
    python = python,
    envVars = envVars,
    server = server,
    account = account,
    lint = FALSE,
    ...
  )
}

deploy_find_environment <- function(
  version = NULL,
  python = NULL,
  backend = NULL,
  main_library = NULL
) {
  ret <- NULL
  failed <- NULL
  env_name <- ""
  exe_py <- py_exe()
  if (is.null(python)) {
    if (!is.null(version)) {
      env_name <- use_envname(
        version = version,
        backend = backend,
        main_library = main_library
      )
      if (names(env_name) == "exact") {
        check_conda <- try(conda_python(env_name), silent = TRUE)
        check_virtualenv <- try(virtualenv_python(env_name), silent = TRUE)
        if (!inherits(check_conda, "try-error")) {
          ret <- check_conda
        }
        if (!inherits(check_virtualenv, "try-error")) ret <- check_virtualenv
      }
      if (is.null(ret)) failed <- env_name
    } else {
      if (grepl("r-sparklyr-", exe_py)) {
        ret <- exe_py
      } else {
        failed <- "Please pass a 'version' or a 'cluster_id'"
      }
    }
  } else {
    validate_python <- file_exists(python)
    if (validate_python) {
      ret <- python
    } else {
      failed <- python
    }
  }
  if (is.null(ret)) {
    if (is.null(exe_py)) {
      cli_abort("No Python environment could be found")
    } else {
      ret <- exe_py
    }
  } else {
    ret <- path_expand(ret)
  }
  cli_bullets(c("i" = "{.header Python:} {ret}"))
  ret
}

utils::globalVariables(
  c("rsconnect_accounts", "rsconnect_deployApp", "rsconnect_deployments")
)

test_that("Find environments works", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = "testhost",
      "DATABRICKS_TOKEN" = "testtoken"
    ),
    {
      local_mocked_bindings(
        py_exe = function(...) {
          return(NULL)
        },
        virtualenv_starter = function(...) {
          return(NULL)
        }
      )
      env_path <- test_databricks_stump_env()
      expect_equal(
        deploy_find_environment(python = env_path),
        env_path
      )
      expect_error(
        deploy_find_environment()
      )
    }
  )
})

test_that("Tests deploy_databricks() happy path cases", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = "test",
      "DATABRICKS_TOKEN" = "test"
    ),
    {
      local_mocked_bindings(
        deploy = function(...) {
          return(list(...))
        },
        databricks_dbr_version = function(...) {
          return("17.3")
        }
      )
      expect_snapshot(deploy_databricks())
      expect_snapshot(deploy_databricks(host = "another1", token = "token"))
    }
  )
})

test_that("Tests deploy_databricks() error cases", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = NA,
      "DATABRICKS_TOKEN" = NA
    ),
    {
      local_mocked_bindings(
        deploy = function(...) {
          return(list(...))
        },
        databricks_dbr_version = function(...) {
          return("17.3")
        }
      )
      expect_snapshot(deploy_databricks(), error = TRUE)
    }
  )
})

accounts_df <- function() {
  data.frame(
    name = c("my_account", "my_account2", "my_account3"),
    server = c("my_server", "my_server2", "my_server3")
  )
}

test_databricks_deploy_output <- function() {
  list(
    appDir = path(getwd()),
    python = test_databricks_stump_env(),
    envVars = c("DATABRICKS_HOST", "DATABRICKS_TOKEN"),
    server = "my_server",
    account = "my_account",
    lint = FALSE
  )
}

test_that("Basic use, passing the cluster's ID", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = "test",
      "DATABRICKS_TOKEN" = "test"
    ),
    {
      local_mocked_bindings(
        rsconnect_deployApp = function(...) list(...),
        rsconnect_accounts = function(...) accounts_df(),
        databricks_dbr_version = function(...) {
          return("17.3")
        },
        py_exe = function(...) test_databricks_stump_env()
      )
      expect_equal(
        deploy_databricks(cluster_id = test_databricks_cluster_id()),
        test_databricks_deploy_output()
      )
    }
  )
})

test_databricks_deploy_file <- function() {
  deploy_folder <- path(tempdir(), random_table_name("dep"))
  dir_create(deploy_folder)
  deploy_file <- path(deploy_folder, "test.qmd")
  writeLines("x", con = deploy_file)
  deploy_file
}

test_that("Simulates interactive session, selects Yes (1) for both prompts", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = "testhost",
      "DATABRICKS_TOKEN" = "testtoken"
    ),
    {
      deploy_file <- test_databricks_deploy_file()
      local_mocked_bindings(
        check_interactive = function(...) TRUE,
        check_rstudio = function(...) TRUE,
        menu = function(...) {
          return(1)
        },
        rsconnect_deployApp = function(...) list(...),
        rsconnect_accounts = function(...) accounts_df(),
        getSourceEditorContext = function(...) {
          x <- list()
          x$path <- as_fs_path(deploy_file)
          x
        },
        py_exe = function(...) test_databricks_stump_env()
      )
      out <- test_databricks_deploy_output()
      out$appDir <- as_fs_path(path_dir(deploy_file))

      expect_equal(
        deploy_databricks(
          version = test_databricks_cluster_version(),
          server = "my_server",
          account = "my_account",
          confirm = TRUE
        ),
        out
      )
    }
  )
})

skip_if_not_databricks()


test_that("Basic use, passing DBR version works", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_test_env()),
    {
      local_mocked_bindings(
        deployApp = function(...) list(...),
        accounts = function(...) accounts_df()
      )
      # Initializes environment
      invisible(test_databricks_stump_env())

      expect_equal(
        deploy_databricks(version = test_databricks_cluster_version()),
        test_databricks_deploy_output()
      )
    }
  )
})


test_that("Use the Cluster ID from environment variable", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_test_env()),
    {
      local_mocked_bindings(
        deployApp = function(...) list(...),
        accounts = function(...) accounts_df()
      )
      expect_equal(
        deploy_databricks(),
        test_databricks_deploy_output()
      )
    }
  )
})

test_that("Additional arguments are passed on to deployApp()", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_test_env()),
    {
      local_mocked_bindings(
        deployApp = function(...) list(...),
        accounts = function(...) accounts_df()
      )

      out <- test_databricks_deploy_output()
      out$appTitle <- "My Cool Title"

      expect_equal(
        deploy_databricks(
          version = test_databricks_cluster_version(),
          appTitle = "My Cool Title"
        ),
        out
      )
    }
  )
})
test_that("Custom environment variables are passed when `host` and `token` are used", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "CONNECT_DATABRICKS_HOST" = "",
      "CONNECT_DATABRICKS_TOKEN" = ""
    ),
    {
      local_mocked_bindings(
        deployApp = function(...) list(...),
        accounts = function(...) accounts_df()
      )

      out <- test_databricks_deploy_output()
      out$envVars <- c("CONNECT_DATABRICKS_HOST", "CONNECT_DATABRICKS_TOKEN")

      expect_equal(
        deploy_databricks(
          version = test_databricks_cluster_version(),
          host = "new_host",
          token = "new_token"
        ),
        out
      )
    }
  )
})

test_that("Fails if no host/token is found", {
  withr::with_envvar(
    new = c(
      "DATABRICKS_HOST" = NA,
      "DATABRICKS_TOKEN" = NA
    ),
    {
      expect_error(
        deploy_databricks(version = test_databricks_cluster_version()),
        "Cluster setup errors"
      )
    }
  )
})


test_that("Simulates interactive session, selects 3 for both prompts", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_test_env()),
    {
      deploy_file <- test_databricks_deploy_file()
      local_mocked_bindings(
        check_interactive = function(...) TRUE,
        check_rstudio = function(...) TRUE,
        menu = function(...) {
          return(3)
        },
        deployApp = function(...) list(...),
        accounts = function(...) accounts_df(),
        getSourceEditorContext = function(...) {
          x <- list()
          x$path <- as_fs_path(deploy_file)
          x
        }
      )
      out <- test_databricks_deploy_output()
      out$appDir <- as_fs_path(path_dir(deploy_file))
      out$server <- "my_server3"
      out$account <- "my_account3"

      expect_equal(
        deploy_databricks(
          version = test_databricks_cluster_version(),
          confirm = TRUE
        ),
        out
      )
    }
  )
})


test_that("Misc deploy tests", {
  expect_error(deploy(), "'backend'")

  local_mocked_bindings(
    accounts = function(...) data.frame()
  )

  expect_error(
    deploy(backend = "databricks"),
    "There are no server accounts setup"
  )
})

skip_if_not_databricks <- function() {
  if (!is.null(.test_env$databricks)) {
    skip_if(.test_env$databricks)
  }
  skip_if_no_db_host()
  skip_if_no_db_token()
  skip_if_no_db_cluster_id()
}

skip_if_no_db_cluster_id <- function() {
  cluster_id <- Sys.getenv("DATABRICKS_CLUSTER_ID", unset = NA)
  skip_if(
    is.na(cluster_id),
    "No Databricks Cluster ID available"
  )
  skip_if_no_db_host()
  skip_if_no_db_token()
  bad_dbr <- inherits(
    try(
      databricks_dbr_info(
        cluster_id = cluster_id,
        host = databricks_host(),
        token = databricks_token()
      ),
      silent = TRUE
    ),
    "try-error"
  )
  .test_env$databricks <- bad_dbr
  skip_if(bad_dbr, "Cluster could not be contacted")
}

skip_if_no_db_host <- function() {
  skip_if(
    inherits(try(databricks_host(), silent = TRUE), "try-error"),
    "No Databricks Host available"
  )
}

skip_if_no_db_token <- function() {
  skip_if_no_db_host()
  skip_if(
    inherits(try(databricks_token(), silent = TRUE), "try-error"),
    "No Databricks Token available"
  )
}

#' @export
ml_cross_validator.pyspark_connection <- function(
  x,
  estimator = NULL,
  estimator_param_maps = NULL,
  evaluator = NULL,
  num_folds = 3,
  collect_sub_models = FALSE,
  parallelism = 1,
  seed = NULL,
  uid = NULL,
  ...
) {
  if (!inherits(estimator, "ml_connect_pipeline")) {
    abort("Only ML Pipelines are supported at this time")
  }
  tuning <- import("pyspark.ml.tuning")
  pipeline <- python_obj_get(estimator)
  stages <- pipeline$getStages()
  stages_name <- stages |>
    map_chr(py_repr) |>
    tolower()
  grid <- tuning$ParamGridBuilder()
  map_names <- estimator_param_maps |>
    names() |>
    map_chr(snake_to_camel) |>
    tolower()
  for (map_name in map_names) {
    stage_match <- startsWith(stages_name, map_name)
    matches <- sum(stage_match)
    if (matches == 1) {
      curr_stage <- stages[stage_match][[1]]
      curr_map <- estimator_param_maps[map_name == map_names][[1]]
      list_map <- imap(curr_map, function(x, y) list(name = y, values = x))
      for (param in list_map) {
        grid$addGrid(
          param = curr_stage[[snake_to_camel(param$name)]],
          values = param$values
        )
      }
    } else if (matches == 0) {
      abort(glue("Could not match `{map_name}` to a function in the pipeline"))
    }
  }
  built_grid <- grid |>
    as_spark_pyobj(x) |>
    invoke("build")
  cv_estimator <- ml_process_fn(
    args = list(
      x = x,
      estimator = python_obj_get(estimator),
      estimator_param_maps = python_obj_get(built_grid),
      evaluator = python_obj_get(evaluator),
      num_folds = num_folds,
      parallelism = parallelism,
      seed = seed,
      collect_sub_models = collect_sub_models
    ),
    fn = "CrossValidator",
    ml_type = "tuning",
    has_fit = TRUE
  )
  metric_name <- try(invoke(evaluator, "getMetricName"), silent = TRUE)
  if (inherits(metric_name, "try-error")) {
    metric_name <- NULL
  }
  cv_estimator$metric_name <- metric_name
  class(cv_estimator) <- c(
    "ml_connect_cross_validator",
    "ml_cross_validator",
    class(cv_estimator)
  )
  cv_estimator
}

#' @export
print.ml_connect_cross_validator_model <- function(x, ...) {
  print(py_repr(python_obj_get(x)))
}

#' @export
ml_fit.ml_connect_cross_validator <- function(x, dataset, ...) {
  fitted <- ml_fit_impl(x, dataset)
  metric_name <- x$metric_name
  x <- python_obj_get(fitted)
  metrics <- x$avgMetrics |>
    map2(
      x$getEstimatorParamMaps(),
      function(x, y) c(set_names(x, metric_name), y)
    ) |>
    map(dplyr::as_tibble) |>
    list_rbind()
  metric_names <- metrics |>
    colnames() |>
    strsplit("__") |>
    map(unlist) |>
    map_chr(function(x) x[length(x)])
  colnames(metrics) <- metric_names
  structure(
    list(
      uid = invoke(fitted, "uid"),
      param_map = list(),
      .jobj = fitted,
      avg_metrics_df = metrics
    ),
    class = c(
      "ml_connect_cross_validator_model",
      "ml_cross_validator_model"
    )
  )
}

#' @importFrom sparklyr tune_grid_spark
#' @export
tune_grid_spark.pyspark_connection <- function(
  sc,
  object,
  preprocessor = NULL,
  resamples,
  ...,
  param_info = NULL,
  grid = 10,
  metrics = NULL,
  eval_time = NULL,
  control = tune::control_grid(),
  num_tasks = NULL
) {
  # Makes sure tidymodels packages are installed
  if (
    !is_installed("tune") |
      !is_installed("workflows") |
      !is_installed("rsample")
  ) {
    cli_abort(
      paste(
        "There are missing Tidymodels packages.",
        "Please run {.run install.packages(\"tidymodels\")} and",
        "then try again."
      )
    )
  }

  rpy2_installed()
  call <- rlang::caller_env()

  prepped <- prep_static(
    object = object,
    preprocessor = preprocessor,
    resamples = resamples,
    param_info = param_info,
    grid = grid,
    metrics = metrics,
    eval_time = eval_time,
    control = control,
    call = call
  )

  static <- prepped$static

  verbose <- control[["verbose"]]

  if (all(resamples$id == "train/test split")) {
    resamples$.seeds <- map(resamples$id, \(x) integer(0))
  } else {
    # Make and set the worker/process seeds if workers get resamples
    resamples$.seeds <- tune::get_parallel_seeds(nrow(resamples))
  }
  # These are not in tune_grid_loop() but it prepares the variables for the next
  # section
  vec_resamples <- resamples |>
    vctrs::vec_split(by = 1:nrow(resamples)) |>
    (\(x) x$val)()
  pasted_pkgs <- paste0("'", prepped$needed_pkgs, "'", collapse = ", ")

  # --------------- Prepares and uploads R objects to Spark --------------------
  # Creating unique file names to avoid re-uploading if possible
  hash_static <- rlang::hash(static)
  hash_resamples <- rlang::hash(vec_resamples)

  # Uploads the files to the Spark temp folder, this function skips the upload
  # if the hashed file name has already been uploaded during the current session
  if (verbose) {
    cli_progress_step(
      "Uploading model, pre-processor, and other info to the Spark session"
    )
  }
  spark_session_add_file(static, sc, hash_static)
  if (verbose) {
    cli_progress_step("Uploading the re-samples to the Spark session")
  }
  spark_session_add_file(vec_resamples, sc, hash_resamples)

  # -------------------------- Creates the UDF ---------------------------------
  # Uses the `loop_call` function as the base of the UDF that will be sent to
  # the Spark session. It works by modifying the text of the function, specifically
  # the file names it reads to load the different R object components

  # Inject function loading code for Spark 4.1.1+ compatibility
  # This will be inserted at the top of loop_call and will load tune functions
  function_capture_code <- "library(tidymodels)"

  grid_code <- loop_call |>
    deparse() |>
    paste0(collapse = "\n") |>
    str_replace("\"rsample\"", pasted_pkgs) |>
    str_replace("debug <- TRUE", "debug <- FALSE") |>
    str_replace("xy <- 1", function_capture_code) |>
    str_replace("static.rds", path(hash_static, ext = "rds")) |>
    str_replace("resamples.rds", path(hash_resamples, ext = "rds"))

  # -------------------- Creates and uploads the grid  -------------------------
  res_id_df <- purrr::map_df(
    seq_len(length(resamples$id)),
    \(x) data.frame(index = x, id = resamples$id[[x]])
  )
  if (static$control[["parallel_over"]] == "everything") {
    full_grid <- grid |>
      dplyr::cross_join(res_id_df) |>
      dplyr::arrange("index")
  } else {
    full_grid <- res_id_df
  }

  full_grid <- full_grid |>
    dplyr::select(-"id") |>
    dplyr::mutate(row_num = dplyr::row_number())

  # The grid is copied to Spark, it will be re-partitioned if `num_tasks`
  # is set. If not set, Spark will decide how many partitions the data will have,
  # that impacts how many discrete jobs there will be set for this run
  if (verbose) {
    cli_progress_step("Copying the grid to the Spark session")
  }
  sc_obj <- spark_session(sc)
  tbl_grid <- sc_obj$createDataFrame(full_grid)
  if (!is.null(num_tasks)) {
    tbl_grid <- tbl_grid$repartition(as.integer(num_tasks))
  }

  # ---------------- Create schema of output from Spark ------------------------
  # The pandas mapping function requires all of the output column names
  # and types to be specified. Types have to be converted too
  cols <- imap_chr(
    static$configs[, colnames(static$configs) != ".config"],
    \(x, y) {
      if (inherits(x, "character")) {
        p_class <- "string"
      } else if (inherits(x, "numeric")) {
        p_class <- "float"
      } else {
        p_class <- class(x)
      }
      paste0(y, " ", p_class, ",")
    }
  ) |>
    paste0(collapse = " ") |>
    paste(
      "metric string, estimator string, estimate double,",
      "config string, index integer"
    )

  # -------------------- Execute the model tuning in Spark ---------------------
  # The grid is passed to mapInPandas() which will run the resulting code in the
  # Spark session in as many parallel jobs as tbl_grid is partitioned by
  if (verbose) {
    cli_progress_step("Executing the model tuning in Spark")
  }
  tuned_results <- tbl_grid |>
    sa_in_pandas(
      .f = grid_code,
      .schema = cols,
      .as_sdf = FALSE,
    ) |>
    collect()

  # -------------------- Starts building the output object ---------------------
  # Finalizes metrics tables by adding the 'id' label, and `.config`, and
  # restoring the 'dot' prefix to the metric fields (Spark does not like
  # names with dots)
  res <- tuned_results[
    tuned_results$metric != "preds_path" &
      tuned_results$metric != "preds_cols",
  ] |>
    dplyr::rename(
      .metric = "metric",
      .estimator = "estimator",
      .estimate = "estimate",
      .config = "config"
    ) |>
    dplyr::left_join(res_id_df, by = "index") |>
    dplyr::select(-"index") |>
    dplyr::arrange("id")

  # Converts metrics to list separated by id's
  res_names <- colnames(res)
  metrics_names <- res_names[res_names != "id"]
  metrics_map <- map(
    unique(res$id),
    \(x) res[res$id == x, metrics_names]
  )

  # Creates dummy 'notes' table
  notes <- list(tibble(
    location = character(0),
    type = character(0),
    note = character(0),
    trace = list()
  ))

  # Joins the resamples, metrics and notes tables, and adds needed attributes
  # to make it a viable tune_results object.
  out <- resamples |>
    as_tibble() |>
    mutate(
      .metrics = metrics_map,
      .notes = notes
    ) |>
    vctrs::new_data_frame() |> # Removes rsample object's attributes
    dplyr::select(-".seeds")

  # -------------------------- Appends predictions -----------------------------
  if (isTRUE(control[["save_pred"]])) {
    # Gets the column schema spec from a row with a metric name of `pred_cols`.
    # It converts the plain text into a list object
    preds_cols_list <- tuned_results[tuned_results$metric == "preds_cols", ] |>
      dplyr::pull("estimator") |>
      strsplit("\\|") |>
      unlist() |>
      map(strsplit, ":") |>
      map(unlist)

    pred_col_names <- preds_cols_list |>
      map_chr(\(x) x[[1]])

    # Converts the R class to Spark type
    preds_cols <- preds_cols_list |>
      map(\(x) {
        new_type <- x[[2]]
        if (new_type %in% c("character", "factor")) {
          new_type <- "string"
        }
        if (new_type == "numeric") {
          new_type <- "float"
        }
        new_name <- gsub("\\.", "_", x[[1]])
        paste(new_name, new_type)
      }) |>
      reduce(c) |>
      paste(collapse = ", ") |>
      paste(", index integer")

    # From the results, it gets the rows that have the path to the RDS files
    # containing the predictions
    pred_paths <- tuned_results[tuned_results$metric == "preds_path", ] |>
      dplyr::select("index", "estimator") |>
      dplyr::rename(path = "estimator")

    # Runs a Spark job that reads the predictions RDS files from Spark and
    # returns them in a single large table

    fn <- loop_predictions |>
      deparse() |>
      paste0(collapse = "\n")

    pred_results <- pred_paths |>
      sc_obj$createDataFrame() |>
      sa_in_pandas(
        .f = fn,
        .schema = preds_cols,
        .as_sdf = FALSE,
      ) |>
      collect()

    colnames(pred_results) <- c(pred_col_names, "index")

    preds <- pred_results |>
      dplyr::left_join(res_id_df, by = "index") |>
      dplyr::select(-"index") |>
      dplyr::arrange("id")

    preds_map <- map(
      unique(preds$id),
      \(x) preds[preds$id == x, pred_col_names]
    )

    # Appends `.predictions` to the output object
    out <- out |>
      mutate(.predictions = preds_map)
  }

  # Sets the `workflow` attribute if requested by `control`
  wf <- NULL
  wf <- if (isTRUE(control[["save_workflow"]])) {
    wf <- prepped$wf
  }

  # ------------------------- Finalizes output object --------------------------
  tibble::new_tibble(
    x = out,
    nrow = nrow(out),
    parameters = static$param_info,
    metrics = static$metrics,
    eval_time = eval_time,
    eval_time_target = NULL,
    outcomes = tune::outcome_names(prepped$wf),
    rset_info = tune::pull_rset_attributes(resamples),
    workflow = wf,
    class = c(class(out), "tune_results")
  )
}

# `x` only contains a table with the grid containing every single combination
loop_call <- function(x) {
  # ----------------- Ensures all needed packages are available ----------------
  needed_pkgs <- c("rsample")
  missing_pkgs <- NULL
  for (pkg in needed_pkgs) {
    if (!rlang::is_installed(pkg)) {
      missing_pkgs <- c(missing_pkgs, pkg)
    }
  }
  if (!is.null(missing_pkgs)) {
    missing_pkgs <- paste0("`", missing_pkgs, "`", collapse = ", ")
    stop("Packages ", missing_pkgs, " are missing")
  }
  xy <- 1

  # ------------------- Reads files with needed R objects ----------------------
  # Loads the needed R objects from disk
  debug <- TRUE
  static_fname <- "static.rds"
  resample_fname <- "resamples.rds"
  if (isFALSE(debug)) {
    pyspark <- reticulate::import("pyspark")
    static_file <- pyspark$SparkFiles$get(static_fname)
    resample_file <- pyspark$SparkFiles$get(resample_fname)
  } else {
    temp_path <- Sys.getenv("TEMP_SPARK_GRID", unset = "~")
    static_file <- file.path(temp_path, static_fname)
    resample_file <- file.path(temp_path, resample_fname)
  }
  static <- readRDS(static_file)
  resamples <- readRDS(resample_file)

  # ------------ Iterates through all the combinations in `x` ------------------
  # Spark will more likely send more than one row (combination) in `x`. It
  # will depend on how the grid data frame was partitioned inside Spark.
  out <- NULL
  for (i in seq_len(nrow(x))) {
    curr_x <- x[i, ]
    curr_row <- curr_x$row_num
    curr_grid <- curr_x[, colnames(curr_x) != "row_num"]
    if (static$control$parallel_over == "everything") {
      curr_grid <- curr_x[, colnames(curr_x) != "index"]
    } else {
      curr_grid <- static$grid
    }
    index <- curr_x$index
    curr_resample <- resamples[[index]]

    data_splits <- tune::.get_data_subsets(
      static$wflow,
      curr_resample$splits[[1]],
      static$split_args
    )
    static$data <- data_splits

    # loop_over_all_stages() requires the grid to be a tibble
    curr_grid <- tibble::as_tibble(curr_grid)
    assign(".Random.seed", c(1L, 2L, 3L), envir = .GlobalEnv)
    res <- tune::.loop_over_all_stages(curr_resample, curr_grid, static)
    # -------------------- Extracts metrics from results -----------------------
    # Mapping function accepts only tables as output, so only the metrics are
    # being sent back instead of the entire results object
    metrics_df <- Reduce(rbind, res$.metrics)
    metrics_df$index <- index
    # ------------------------ Saves predictions -------------------------------
    # If saves_pred is TRUE, the .predictions table is saved in the same
    # directory the initial context and resample objects. There will be one
    # file per permutation. The metrics table will be modified to include the
    # path to the predictions file, and also the column schema of all of the
    # tables
    if (isTRUE(static[["control"]][["save_pred"]])) {
      static_name <- substr(static_fname, 1, nchar(static_fname) - 4)
      resample_name <- substr(resample_fname, 1, nchar(resample_fname) - 4)
      preds_fname <- paste0(resample_name, "-", static_name, "-", index, ".rds")
      if (isTRUE(debug)) {
        base_path <- temp_path
      } else {
        base_path <- pyspark$SparkFiles$getRootDirectory()
      }
      preds <- res$.predictions[[1]]
      preds_file <- file.path(base_path, preds_fname)
      saveRDS(preds, preds_file)
      preds_df <- metrics_df[1, ]
      preds_df$`.estimator` <- preds_file
      preds_df$`.metric` <- "preds_path"
      if (curr_row == 1) {
        cols <- preds |>
          map_chr(class) |>
          imap(\(x, y) paste0(y, ":", x)) |>
          reduce(c) |>
          paste0(collapse = "|")
        preds_cols <- preds_df
        preds_cols$`.estimator` <- cols[[1]]
        preds_cols$`.metric` <- "preds_cols"
        preds_df <- rbind(preds_cols, preds_df)
      }
      metrics_df <- rbind(preds_df, metrics_df)
    }
    out <- rbind(out, metrics_df)
  }
  out
}

loop_predictions <- function(x) {
  out <- NULL
  for (i in seq_len(nrow(x))) {
    cr <- x[i, ]
    for (j in 1:5) {
      if (file.exists(cr$path)) {
        curr <- readRDS(cr$path)
        curr$index <- cr$index
        out <- rbind(out, curr)
        break()
      } else {
        Sys.sleep(0.1)
      }
    }
  }
  out
}

prep_static <- function(
  object,
  preprocessor,
  resamples,
  param_info = NULL,
  grid = grid,
  metrics = NULL,
  eval_time = NULL,
  control = NULL,
  call = NULL
) {
  if (is.null(preprocessor)) {
    wf <- object
  } else {
    wf <- workflows::workflow() |>
      workflows::add_model(object) |>
      workflows::add_recipe(preprocessor)
  }

  # ------------------------- Creates `static` object --------------------------
  # This part mostly recreates `tune_grid_loop()` to properly create the
  # `resamples` and `static` objects in order to pass it to the
  # loop_over_all_stages() function that is called inside Spark
  # https://github.com/tidymodels/tune/blob/main/R/tune_grid_loop.R

  wf_metrics <- tune::check_metrics_arg(metrics, wf, call = call)
  param_info <- tune::check_parameters(
    wflow = wf,
    data = resamples$splits[[1]]$data,
    grid_names = names(grid)
  )
  grid <- tune::.check_grid(
    grid = grid,
    workflow = wf,
    pset = param_info
  )

  # Checking for argument values in the `control` that are not supported
  # by this backend
  control_err <- NULL
  if (!is.null(control[["extract"]])) {
    control_err <- "This backend only supports `extract` set to NULL"
  }
  if (!is.null(control[["backend_options"]])) {
    control_err <- c(
      control_err,
      "This backend only supports `backend_options` set to NULL"
    )
  }
  if (!is.null(control_err)) {
    abort(c(
      "The following incompatability errors in `control` were found:",
      control_err
    ))
  }
  control <- tune::.update_parallel_over(control, resamples, grid)
  eval_time <- tune::check_eval_time_arg(eval_time, wf_metrics, call = call)
  needed_pkgs <- c(
    "rsample",
    "workflows",
    "hardhat",
    "tune",
    "reticulate",
    "parsnip",
    "tailor",
    "yardstick",
    "tidymodels",
    workflows::required_pkgs(wf),
    control$pkgs
  ) |>
    unique()
  out <- list()
  out$static <- list(
    wflow = wf,
    param_info = param_info,
    configs = tune::.get_config_key(grid, wf),
    post_estimation = workflows::.workflow_postprocessor_requires_fit(wf),
    metrics = wf_metrics,
    metric_info = tibble::as_tibble(wf_metrics),
    pred_types = tune::.determine_pred_types(wf, wf_metrics),
    eval_time = eval_time,
    split_args = rsample::.get_split_args(resamples),
    control = control,
    pkgs = needed_pkgs,
    strategy = "sequential",
    data = list(fit = NULL, pred = NULL, cal = NULL),
    grid = grid
  )
  out$wf <- wf
  out$needed_pkgs <- needed_pkgs
  out
}

skip_spark_min_version(4.0)

test_that("parallel_over resamples works", {
  x <- use_tune_grid()
  sc <- use_test_spark_connect()
  spark_results <- tune_grid_spark(
    sc = sc,
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = x$control
  )
  local_results <- tune::tune_grid(
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = x$control
  )
  expect_equal(colnames(local_results), colnames(spark_results))
  expect_equal(
    names(attributes(local_results)),
    names(attributes(spark_results))
  )
  expect_equal(local_results[, 1:4], spark_results[, 1:4])
  expect_equal_preds(local_results, spark_results)
})

test_that("parallel_over everything works", {
  x <- use_tune_grid()
  sc <- use_test_spark_connect()
  cntrl <- tune::control_grid(parallel_over = "everything", save_pred = TRUE)
  spark_results <- tune_grid_spark(
    sc = sc,
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl
  )
  local_results <- tune::tune_grid(
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl
  )
  expect_equal(colnames(local_results), colnames(spark_results))
  expect_equal(
    names(attributes(local_results)),
    names(attributes(spark_results))
  )
  expect_equal(local_results[, 1:4], spark_results[, 1:4])
  expect_equal_preds(local_results, spark_results)
})

test_that("save_workflow works", {
  x <- use_tune_grid()
  sc <- use_test_spark_connect()
  cntrl <- tune::control_grid(save_workflow = TRUE)
  spark_results <- tune_grid_spark(
    sc = sc,
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl
  )
  local_results <- tune::tune_grid(
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl
  )
  expect_equal(colnames(local_results), colnames(spark_results))
  expect_equal(
    names(attributes(local_results)),
    names(attributes(spark_results))
  )
  expect_equal(local_results[, 1:4], spark_results[, 1:4])
})

test_that("Accepts workflow object", {
  x <- use_tune_grid()
  sc <- use_test_spark_connect()

  wf <- workflows::workflow(x$preprocessor, x$object)

  spark_results <- tune_grid_spark(
    sc = sc,
    object = wf,
    resamples = x$resamples,
    grid = x$grid
  )
  local_results <- tune::tune_grid(
    object = wf,
    resamples = x$resamples,
    grid = x$grid
  )
  expect_equal(colnames(local_results), colnames(spark_results))
  expect_equal(
    names(attributes(local_results)),
    names(attributes(spark_results))
  )
  expect_equal(local_results[, 1:4], spark_results[, 1:4])
})

test_that("loop_predictions works", {
  expect_null(
    loop_predictions(data.frame(path = "test", index = 1))
  )

  expect_equal(
    loop_predictions(
      data.frame(path = test_path("_data/ovarian.rds"), index = 1)
    ),
    test_path("_data/ovarian.rds") |>
      readRDS() |>
      dplyr::mutate(index = 1)
  )
})

test_that("loop_call works", {
  x <- use_tune_grid()
  grid <- x$grid
  resamples <- x$resamples
  prepped <- prep_static(
    object = x$object,
    preprocessor = x$preprocessor,
    resamples = resamples,
    grid = x$grid,
    control = x$control,
    call = rlang::caller_env()
  )
  if (all(resamples$id == "train/test split")) {
    resamples$.seeds <- map(resamples$id, \(x) integer(0))
  } else {
    resamples$.seeds <- tune::get_parallel_seeds(nrow(resamples))
  }
  vec_resamples <- resamples |>
    vctrs::vec_split(by = 1:nrow(resamples)) |>
    _$val
  temp_dir <- fs::path_temp()
  saveRDS(prepped$static, file.path(temp_dir, "static.rds"))
  saveRDS(vec_resamples, file.path(temp_dir, "resamples.rds"))
  withr::with_envvar(
    new = c("TEMP_SPARK_GRID" = temp_dir),
    {
      library(tune)
      res <- loop_call(data.frame(index = 1, row_num = 1))
    }
  )
  expect_s3_class(res, "data.frame")
})

test_that("Post-processing works", {
  x <- use_tune_grid()
  sc <- use_test_spark_connect()
  wf <- workflows::workflow(x$preprocessor, x$object, x$post)
  cntrl <- tune::control_grid(save_pred = TRUE)
  results_wf <- tune::tune_grid(
    object = wf,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl
  )
  results_spark <- tune_grid_spark(
    object = wf,
    resamples = x$resamples,
    grid = x$grid,
    control = cntrl,
    sc = sc
  )
  expect_equal_preds(results_spark, results_wf)
})

#' @export
spark_integ_test_skip.connect_databricks <- function(sc, test_name) {
  spark_integ_generic(test_name)
}

#' @export
spark_integ_test_skip.connect_spark <- function(sc, test_name) {
  spark_integ_generic(test_name) |>
    supports("format-csv", test_name) |>
    supports("format-parquet", test_name) |>
    supports("format-orc", test_name) |>
    supports("format-json", test_name) |>
    supports("format-text", test_name)
}


spark_integ_generic <- function(test_name) {
  out <- TRUE

  out |>
    supports("dplyr", test_name) |>
    supports("dplyr-do", test_name, TRUE) |>
    supports("dplyr-hof", test_name, TRUE) |>
    supports("dplyr-cumprod", test_name, TRUE) |>
    supports("DBI", test_name) |>
    supports("pivot-longer", test_name) |>
    supports("pivot-longer-names-repair", test_name, TRUE) |>
    supports("pivot-longer-values-transform", test_name, TRUE) |>
    supports("ml-", test_name, TRUE) |>
    supports("sdf-distinct", test_name)
}

supports <- function(x, test, test_name, default_out = FALSE) {
  if (grepl(test, test_name)) {
    x <- default_out
  }
  x
}

#' @export
spark_version.pyspark_connection <- function(sc) {
  python_conn(sc)$version
}

#' @export
connection_is_open.pyspark_connection <- function(sc) {
  # TODO: replace with actual connection state code
  TRUE
}

#' @export
spark_connection.pyspark_connection <- function(x, ...) {
  x
}

#' @export
hive_context.pyspark_connection <- function(sc) {
  sc
}

#' @export
spark_session.pyspark_connection <- function(sc) {
  sc$session
}

#' @export
spark_dataframe.pyspark_connection <- function(x, ...) {
  x
}

#' @export
spark_web.pyspark_connection <- function(sc, ...) {
  # TODO: Implement later when SparkContext is implemented
  ""
}

#' @export
spark_log.pyspark_connection <- function(sc, n = 100, filter = NULL, ...) {
  # TODO: Implement later when SparkContext is implemented
  ""
}

reticulate_python_check <- function(
  ignore = FALSE,
  unset = FALSE,
  message = TRUE
) {
  if (ignore) {
    return("")
  }

  out <- ""

  env_var <- Sys.getenv("RETICULATE_PYTHON", unset = NA)

  if (current_product_connect() && !is.na(env_var)) {
    out <- env_var
  }

  if (!is.na(env_var) && out == "") {
    if (unset) {
      Sys.unsetenv("RETICULATE_PYTHON")
      if (message) {
        cli_div(theme = cli_colors())
        cli_alert_warning(paste0(
          "{.header Your }{.emph 'RETICULATE_PYTHON'}",
          " {.header environment variable was unset.}"
        ))
        cli_bullets(c(
          " " = "{.class To recover, restart R after you closing your Spark session}"
        ))
        cli_end()
      }
    } else {
      if (message) {
        cli_alert_warning(paste(
          paste0(
            "Your {.emph 'RETICULATE_PYTHON'} environment is set, ",
            "which may cause connectivity issues"
          ),
          paste0(
            "Use {.code Sys.unsetenv(\"RETICULATE_PYTHON\")} to clear for the",
            "duration of your current R session."
          ),
          sep = "\n"
        ))
      }
    }
  }
  out
}

check_arg_supported <- function(x, msg = NULL) {
  arg <- enquo(x)
  if (!is.null(x)) {
    if (is.null(msg)) {
      arg_name <- deparse(x)
      msg <- glue(
        paste(
          "The '{rlang::as_label(arg)}' argument is not currently",
          "supported for this back-end"
        )
      )
    }
    cli_abort(msg)
  }
  invisible()
}

cli_colors <- function(envir = parent.frame()) {
  list(
    span.header = list(color = "silver"),
    span.class = list(color = "darkgray"),
    span.info = list(),
    span.spark = list(color = "darkgray")
  )
}


current_product_connect <- function() {
  out <- FALSE
  if (Sys.getenv("RSTUDIO_PRODUCT") == "CONNECT") {
    out <- TRUE
  }
  out
}

py_check_installed <- function(
  envname = NULL,
  libraries = "",
  msg = ""
) {
  installed_libraries <- py_list_packages(envname = envname)$package
  find_libs <- map_lgl(libraries, \(.x) .x %in% installed_libraries)
  if (!all(find_libs)) {
    missing_lib <- libraries[!find_libs]
    if (is_uv_env()) {
      reticulate::py_require(missing_lib)
      return(invisible())
    }
    cli_div(theme = cli_colors())
    if (check_interactive()) {
      cli_alert_warning(msg)
      cli_bullets(c(
        " " = "{.header Could not find: {missing_lib}}",
        " " = "Do you wish to install? {.class (This will be a one time operation)}"
      ))
      choice <- menu(choices = c("Yes", "Cancel"))
      if (choice == 1) {
        py_install(missing_lib)
      }
      if (choice == 2) {
        stop_quietly()
      }
    } else {
      cli_abort(msg)
    }
    cli_end()
  }
  invisible()
}

is_uv_env <- function() {
  is_uv <- FALSE
  base_exe <- path_dir(py_exe())
  if (path_file(base_exe) == "bin") {
    py_base <- path_dir(base_exe)
    cfg_file <- path(py_base, "pyvenv.cfg")
    if (file_exists(cfg_file)) {
      cfg_contents <- readLines(cfg_file)
      is_uv <- any(grepl("uv = ", cfg_contents))
    }
  }
  is_uv
}


stop_quietly <- function() {
  opt <- options(show.error.messages = FALSE)
  on.exit(options(opt))
  stop()
}

use_arrow <- function() {
  arrow::binary()
}

list_diff <- function(x, y) {
  x_names <- names(x)
  y_names <- names(y)

  # Find elements that are either:
  c(
    x[setdiff(x_names, y_names)], # 1. New names in x
    x[!mapply(identical, x[y_names], y)] # 2. Same name but different value
  )
}

pkg_path <- function(x) {
  if (dir_exists(glue("inst/{x}"))) {
    out <- path_expand(glue("inst/{x}"))
  } else {
    out <- system.file(x, package = "pysparklyr")
    if (out == "") {
      cli_abort("Path '{x}' not found", call = NULL)
    }
  }
  out
}

str_replace <- function(string, pattern, replacement) {
  sub(pattern, replacement, string)
}

str_replace_all <- function(string, pattern, replacement) {
  gsub(pattern, replacement, string)
}

test_that("Reticulate Python check works", {
  withr::with_envvar(
    new = c("RETICULATE_PYTHON" = "path/for/python"),
    {
      expect_equal(
        reticulate_python_check(ignore = TRUE),
        ""
      )
      expect_message(
        reticulate_python_check(unset = FALSE, message = TRUE),
        "Your 'RETICULATE_PYTHON' environment is set, which may cause"
      )
    }
  )
})

test_that("When in Posit Connect no unset happens", {
  local_mocked_bindings(current_product_connect = function(...) TRUE)
  withr::with_envvar(
    new = c("RETICULATE_PYTHON" = "path/for/python"),
    {
      expect_equal(
        reticulate_python_check(),
        "path/for/python"
      )
    }
  )
})

test_that("Check arguments works", {
  x <- ""
  expect_error(
    check_arg_supported(x),
    "The 'x' argument is not currently supported for this back-end"
  )
})

#' @export
spark_connect_method.spark_method_databricks_connect <- function(
  x,
  method,
  master,
  spark_home,
  config = pyspark_config(),
  app_name,
  version = NULL,
  hadoop_version,
  extensions,
  scala_version,
  ...
) {
  args <- list(...)
  cluster_id <- args$cluster_id
  serverless <- args$serverless %||% FALSE
  profile <- args$profile %||% NULL
  token <- args$token
  envname <- args$envname
  silent <- args$silent %||% FALSE

  method <- method[[1]]

  master <- databricks_host(master, fail = FALSE)
  token <- databricks_token(token, fail = FALSE)

  # if serverless ignore specified cluster ids
  if (serverless) {
    cluster_id <- NULL
  } else {
    cluster_id <- cluster_id %||% Sys.getenv("DATABRICKS_CLUSTER_ID")
  }

  cluster_info <- NULL
  if (cluster_id != "" && !serverless && is.null(version) && !is.null(token)) {
    cluster_info <- databricks_dbr_info(
      cluster_id = cluster_id,
      host = master,
      token = token
    )
    version <- databricks_extract_version(cluster_info)
  }

  # Track if version was provided by user
  version_provided <- !is.null(version)

  # load python env
  envname <- use_envname(
    backend = "databricks",
    main_library = "databricks.connect",
    version = version,
    envname = envname,
    messages = !silent,
    match_first = FALSE,
    ask_if_not_installed = FALSE,
    python_version = args$python_version
  )

  if (is.null(envname)) {
    return(invisible)
  }

  # load python libs
  dbc <- import_check("databricks.connect", envname, silent)
  db_sdk <- import_check("databricks.sdk", envname, silent = TRUE)

  if (is.null(token) || token == "") {
    if (!is.null(profile)) {
      sdk_config <- db_sdk$core$Config(profile = profile)
    } else {
      sdk_config <- db_sdk$core$Config()
    }

    token <- sdk_config$token %||% ""
  }

  # create workspace client
  sdk_client <- databricks_sdk_client(
    sdk = db_sdk,
    host = master,
    serverless = serverless,
    cluster_id = cluster_id,
    token = token,
    profile = profile
  )

  # if serverless override cluster_id and set to `NULL`
  if (!serverless) {
    if (cluster_id != "" && is.null(cluster_info)) {
      cluster_info <- databricks_dbr_info(
        cluster_id = cluster_id,
        client = sdk_client,
        silent = silent
      )
    }
  }

  # Check for version mismatch and warn user
  if (!is.null(cluster_info) && !version_provided) {
    cluster_version <- databricks_extract_version(cluster_info)
    if (
      !is.null(version) && cluster_version != "" && cluster_version != version
    ) {
      if (!silent) {
        cli_div(theme = cli_colors())
        cli_alert_warning(
          paste0(
            "Using databricks.connect version {.emph {version}}, which differs from ",
            "Databricks' DBR version {.emph {cluster_version}}. If you experience instability, ",
            "consider using {.code version = \"{cluster_version}\"} to ensure a matching ",
            "version is used during the R session."
          )
        )
        cli_end()
      }
    }
  }

  if (!is.null(cluster_info)) {
    msg <- "{.header Connecting to} {.emph '{cluster_info$cluster_name}'}"
    msg_done <- "{.header Connected to:} {.emph '{cluster_info$cluster_name}'}"
    master_label <- glue("{cluster_info$cluster_name} ({cluster_id})")
  } else if (!serverless) {
    msg <- "{.header Connecting to} {.emph '{cluster_id}'}"
    msg_done <- "{.header Connected to:} '{.emph '{cluster_id}'}'"
    master_label <- glue("Databricks Connect - Cluster: {cluster_id}")
  } else if (serverless) {
    msg <- "{.header Connecting to} {.emph serverless}"
    msg_done <- "{.header Connected to:} '{.emph serverless}'"
    master_label <- glue("Databricks Connect - Cluster: serverless")
  }

  if (!silent) {
    cli_div(theme = cli_colors())
    cli_progress_step(msg, msg_done)
  }

  # build databricks session connection
  user_agent <- build_user_agent()
  conn <- dbc$DatabricksSession$builder$sdkConfig(sdk_client$config)$userAgent(
    user_agent
  )

  if (!silent) {
    cli_progress_done()
    cli_end()
  }

  initialize_connection(
    conn = conn,
    master_label = master_label,
    con_class = "connect_databricks",
    cluster_id = cluster_id,
    serverless = serverless,
    method = method,
    config = config
  )
}

setOldClass(
  c("connect_databricks", "pyspark_connection", "spark_connection")
)

# ----------------------------- Binarizer --------------------------------------
ft_binarizer_impl <- function(
  x,
  input_col,
  output_col,
  threshold = 0,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "Binarizer",
    FALSE
  )
}
#' @export
ft_binarizer.ml_connect_pipeline <- ft_binarizer_impl
#' @export
ft_binarizer.pyspark_connection <- ft_binarizer_impl
#' @export
ft_binarizer.tbl_pyspark <- ft_binarizer_impl

# ------------------------------- Bucketizer -----------------------------------
ft_bucketizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  splits = NULL,
  input_cols = NULL,
  output_cols = NULL,
  splits_array = NULL,
  handle_invalid = "error",
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "Bucketizer",
    FALSE
  )
}
#' @export
ft_bucketizer.ml_connect_pipeline <- ft_bucketizer_impl
#' @export
ft_bucketizer.pyspark_connection <- ft_bucketizer_impl
#' @export
ft_bucketizer.tbl_pyspark <- ft_bucketizer_impl

# ---------------------- Bucket Random Projection LSH --------------------------
ft_bucketed_random_projection_lsh_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  bucket_length = NULL,
  num_hash_tables = 1,
  seed = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "BucketedRandomProjectionLSH",
    TRUE
  )
}
#' @export
ft_bucketed_random_projection_lsh.ml_connect_pipeline <- ft_bucketed_random_projection_lsh_impl
#' @export
ft_bucketed_random_projection_lsh.pyspark_connection <- ft_bucketed_random_projection_lsh_impl
#' @export
ft_bucketed_random_projection_lsh.tbl_pyspark <- ft_bucketed_random_projection_lsh_impl

# ----------------------------- Count vectorizer -------------------------------
ft_count_vectorizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  binary = FALSE,
  min_df = NULL,
  min_tf = NULL,
  vocab_size = 2^18,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "CountVectorizer",
    TRUE
  )
}
#' @export
ft_count_vectorizer.ml_connect_pipeline <- ft_count_vectorizer_impl
#' @export
ft_count_vectorizer.pyspark_connection <- ft_count_vectorizer_impl
#' @export
ft_count_vectorizer.tbl_pyspark <- ft_count_vectorizer_impl

# ---------------------------------- DCT  --------------------------------------
ft_dct_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  inverse = FALSE,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "DCT", FALSE)
}
#' @export
ft_dct.ml_connect_pipeline <- ft_dct_impl
#' @export
ft_dct.pyspark_connection <- ft_dct_impl
#' @export
ft_dct.tbl_pyspark <- ft_dct_impl
#' @export
ft_discrete_cosine_transform.ml_connect_pipeline <- ft_dct_impl
#' @export
ft_discrete_cosine_transform.pyspark_connection <- ft_dct_impl
#' @export
ft_discrete_cosine_transform.tbl_pyspark <- ft_dct_impl

# -------------------------- Elementwise Product  ------------------------------
ft_elementwise_product_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  scaling_vec = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "ElementwiseProduct",
    FALSE
  )
}
#' @export
ft_elementwise_product.ml_connect_pipeline <- ft_elementwise_product_impl
#' @export
ft_elementwise_product.pyspark_connection <- ft_elementwise_product_impl
#' @export
ft_elementwise_product.tbl_pyspark <- ft_elementwise_product_impl


# ---------------------------- Feature Hasher  ---------------------------------
ft_feature_hasher_impl <- function(
  x,
  input_cols = NULL,
  output_col = NULL,
  num_features = 2^18,
  categorical_cols = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "FeatureHasher",
    FALSE
  )
}
#' @export
ft_feature_hasher.ml_connect_pipeline <- ft_feature_hasher_impl
#' @export
ft_feature_hasher.pyspark_connection <- ft_feature_hasher_impl
#' @export
ft_feature_hasher.tbl_pyspark <- ft_feature_hasher_impl

# ----------------------- Hashing term frequencies  ----------------------------

ft_hashing_tf_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  binary = FALSE,
  num_features = 2^18,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "HashingTF",
    FALSE
  )
}
#' @export
ft_hashing_tf.ml_connect_pipeline <- ft_hashing_tf_impl
#' @export
ft_hashing_tf.pyspark_connection <- ft_hashing_tf_impl
#' @export
ft_hashing_tf.tbl_pyspark <- ft_hashing_tf_impl

# -------------------- Inverse document frequency  -----------------------------
ft_idf_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  min_doc_freq = 0,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "IDF", TRUE)
}
#' @export
ft_idf.ml_connect_pipeline <- ft_idf_impl
#' @export
ft_idf.pyspark_connection <- ft_idf_impl
#' @export
ft_idf.tbl_pyspark <- ft_idf_impl

# -------------------------------- Imputer  ------------------------------------
ft_imputer_impl <- function(
  x,
  input_cols = NULL,
  output_cols = NULL,
  missing_value = NULL,
  strategy = "mean",
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "Imputer", TRUE)
}
#' @export
ft_imputer.ml_connect_pipeline <- ft_imputer_impl
#' @export
ft_imputer.pyspark_connection <- ft_imputer_impl
#' @export
ft_imputer.tbl_pyspark <- ft_imputer_impl

# ------------------------- Index to string  -----------------------------------
ft_index_to_string_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  labels = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "IndexToString",
    FALSE
  )
}
#' @export
ft_index_to_string.ml_connect_pipeline <- ft_index_to_string_impl
#' @export
ft_index_to_string.pyspark_connection <- ft_index_to_string_impl
#' @export
ft_index_to_string.tbl_pyspark <- ft_index_to_string_impl

# ----------------------------- Interaction  -----------------------------------
ft_interaction_impl <- function(
  x,
  input_cols = NULL,
  output_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "Interaction",
    FALSE
  )
}
#' @export
ft_interaction.ml_connect_pipeline <- ft_interaction_impl
#' @export
ft_interaction.pyspark_connection <- ft_interaction_impl
#' @export
ft_interaction.tbl_pyspark <- ft_interaction_impl

# --------------------------- Max Abs Scaler  ----------------------------------
ft_max_abs_scaler_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "MaxAbsScaler",
    TRUE
  )
}
#' @export
ft_max_abs_scaler.ml_connect_pipeline <- ft_max_abs_scaler_impl
#' @export
ft_max_abs_scaler.pyspark_connection <- ft_max_abs_scaler_impl
#' @export
ft_max_abs_scaler.tbl_pyspark <- ft_max_abs_scaler_impl

# --------------------------- Min Max Scaler  ----------------------------------
ft_min_max_scaler_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  min = 0,
  max = 1,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "MinMaxScaler",
    TRUE
  )
}
#' @export
ft_min_max_scaler.ml_connect_pipeline <- ft_min_max_scaler_impl
#' @export
ft_min_max_scaler.pyspark_connection <- ft_min_max_scaler_impl
#' @export
ft_min_max_scaler.tbl_pyspark <- ft_min_max_scaler_impl

# --------------------------- Min Hash LSH  ----------------------------------
ft_minhash_lsh_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  num_hash_tables = 1L,
  seed = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "MinHashLSH",
    TRUE
  )
}
#' @export
ft_minhash_lsh.ml_connect_pipeline <- ft_minhash_lsh_impl
#' @export
ft_minhash_lsh.pyspark_connection <- ft_minhash_lsh_impl
#' @export
ft_minhash_lsh.tbl_pyspark <- ft_minhash_lsh_impl

# -------------------------------- N-gram  -------------------------------------
ft_ngram_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  n = 2,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "NGram", FALSE)
}
#' @export
ft_ngram.ml_connect_pipeline <- ft_ngram_impl
#' @export
ft_ngram.pyspark_connection <- ft_ngram_impl
#' @export
ft_ngram.tbl_pyspark <- ft_ngram_impl

# ------------------------------ Normalizer  -----------------------------------
ft_normalizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  p = 2,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "Normalizer",
    FALSE
  )
}
#' @export
ft_normalizer.ml_connect_pipeline <- ft_normalizer_impl
#' @export
ft_normalizer.pyspark_connection <- ft_normalizer_impl
#' @export
ft_normalizer.tbl_pyspark <- ft_normalizer_impl

# --------------------------- One hot encoder  ---------------------------------
ft_one_hot_encoder_impl <- function(
  x,
  input_cols = NULL,
  output_cols = NULL,
  handle_invalid = NULL,
  drop_last = TRUE,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "OneHotEncoder",
    TRUE
  )
}
#' @export
ft_one_hot_encoder.ml_connect_pipeline <- ft_one_hot_encoder_impl
#' @export
ft_one_hot_encoder.pyspark_connection <- ft_one_hot_encoder_impl
#' @export
ft_one_hot_encoder.tbl_pyspark <- ft_one_hot_encoder_impl

# --------------------------------- PCA  ---------------------------------------
ft_pca_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  k = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "PCA", TRUE)
}
#' @export
ft_pca.ml_connect_pipeline <- ft_pca_impl
#' @export
ft_pca.pyspark_connection <- ft_pca_impl
#' @export
ft_pca.tbl_pyspark <- ft_pca_impl

# ------------------------- Polynomial Expansion  ------------------------------
ft_polynomial_expansion_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  degree = 2,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "PolynomialExpansion",
    FALSE
  )
}
#' @export
ft_polynomial_expansion.ml_connect_pipeline <- ft_polynomial_expansion_impl
#' @export
ft_polynomial_expansion.pyspark_connection <- ft_polynomial_expansion_impl
#' @export
ft_polynomial_expansion.tbl_pyspark <- ft_polynomial_expansion_impl

# ----------------------- Quantile Discretizer  ---------------------------------
ft_quantile_discretizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  num_buckets = 2,
  input_cols = NULL,
  output_cols = NULL,
  num_buckets_array = NULL,
  handle_invalid = "error",
  relative_error = 0.001,
  uid = NULL,
  weight_column = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "QuantileDiscretizer",
    TRUE
  )
}
#' @export
ft_quantile_discretizer.ml_connect_pipeline <- ft_quantile_discretizer_impl
#' @export
ft_quantile_discretizer.pyspark_connection <- ft_quantile_discretizer_impl
#' @export
ft_quantile_discretizer.tbl_pyspark <- ft_quantile_discretizer_impl


# ----------------------------- RFormula ---------------------------------------
ft_r_formula_impl <- function(
  x,
  formula = NULL,
  features_col = "features",
  label_col = "label",
  force_index_label = FALSE,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "RFormula", TRUE)
}
#' @export
ft_r_formula.ml_connect_pipeline <- ft_r_formula_impl
#' @export
ft_r_formula.pyspark_connection <- ft_r_formula_impl
#' @export
ft_r_formula.tbl_pyspark <- ft_r_formula_impl

# --------------------------- Regex Tokenizer  ---------------------------------
ft_regex_tokenizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  gaps = TRUE,
  min_token_length = 1,
  pattern = "\\s+",
  to_lower_case = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "RegexTokenizer",
    FALSE
  )
}
#' @export
ft_regex_tokenizer.ml_connect_pipeline <- ft_regex_tokenizer_impl
#' @export
ft_regex_tokenizer.pyspark_connection <- ft_regex_tokenizer_impl
#' @export
ft_regex_tokenizer.tbl_pyspark <- ft_regex_tokenizer_impl

# --------------------------- Robust Scaler  -----------------------------------
ft_robust_scaler_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  lower = 0.25,
  upper = 0.75,
  with_centering = TRUE,
  with_scaling = TRUE,
  relative_error = 0.001,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "RobustScaler",
    TRUE
  )
}
#' @export
ft_robust_scaler.ml_connect_pipeline <- ft_robust_scaler_impl
#' @export
ft_robust_scaler.pyspark_connection <- ft_robust_scaler_impl
#' @export
ft_robust_scaler.tbl_pyspark <- ft_robust_scaler_impl

# --------------------------------- SQL  ---------------------------------------
ft_sql_transformer_impl <- function(x, statement = NULL, uid = NULL, ...) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "SQLTransformer",
    FALSE
  )
}
#' @export
ft_sql_transformer.ml_connect_pipeline <- ft_sql_transformer_impl
#' @export
ft_sql_transformer.pyspark_connection <- ft_sql_transformer_impl
#' @export
ft_sql_transformer.tbl_pyspark <- ft_sql_transformer_impl

# -------------------------- Standard Scaler  ----------------------------------
ft_standard_scaler_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  with_mean = NULL,
  with_std = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "StandardScaler",
    TRUE
  )
}
#' @export
ft_standard_scaler.ml_connect_pipeline <- ft_standard_scaler_impl
#' @export
ft_standard_scaler.pyspark_connection <- ft_standard_scaler_impl
#' @export
ft_standard_scaler.tbl_pyspark <- ft_standard_scaler_impl

# -------------------------- Stop words remover --------------------------------

ft_stop_words_remover_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  case_sensitive = FALSE,
  stop_words = NULL,
  uid = NULL,
  ...
) {
  # TODO: Add way to set stop_words same way as regular sparklyr calls
  # not needed before release
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "StopWordsRemover",
    FALSE
  )
}
#' @export
ft_stop_words_remover.ml_connect_pipeline <- ft_stop_words_remover_impl
#' @export
ft_stop_words_remover.pyspark_connection <- ft_stop_words_remover_impl
#' @export
ft_stop_words_remover.tbl_pyspark <- ft_stop_words_remover_impl

# ---------------------------- String Indexer  ---------------------------------

ft_string_indexer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  handle_invalid = "error",
  string_order_type = "frequencyDesc",
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "StringIndexer",
    TRUE
  )
}
#' @export
ft_string_indexer.ml_connect_pipeline <- ft_string_indexer_impl
#' @export
ft_string_indexer.pyspark_connection <- ft_string_indexer_impl
#' @export
ft_string_indexer.tbl_pyspark <- ft_string_indexer_impl

# ----------------------------- Tokenizer --------------------------------------
ft_tokenizer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "Tokenizer",
    FALSE
  )
}
#' @export
ft_tokenizer.ml_connect_pipeline <- ft_tokenizer_impl
#' @export
ft_tokenizer.pyspark_connection <- ft_tokenizer_impl
#' @export
ft_tokenizer.tbl_pyspark <- ft_tokenizer_impl

# -------------------------- Vector Assembler ----------------------------------
ft_vector_assembler_impl <- function(
  x,
  input_cols = NULL,
  output_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "VectorAssembler",
    FALSE
  )
}
#' @export
ft_vector_assembler.ml_connect_pipeline <- ft_vector_assembler_impl
#' @export
ft_vector_assembler.pyspark_connection <- ft_vector_assembler_impl
#' @export
ft_vector_assembler.tbl_pyspark <- ft_vector_assembler_impl

# ---------------------------- Vector Indexer ----------------------------------
ft_vector_indexer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  handle_invalid = "error",
  max_categories = 20,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "VectorIndexer",
    TRUE
  )
}
#' @export
ft_vector_indexer.ml_connect_pipeline <- ft_vector_indexer_impl
#' @export
ft_vector_indexer.pyspark_connection <- ft_vector_indexer_impl
#' @export
ft_vector_indexer.tbl_pyspark <- ft_vector_indexer_impl

# ---------------------------- Vector Slicer ----------------------------------
ft_vector_slicer_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  indices = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(
    c(as.list(environment()), list(...)),
    "VectorSlicer",
    FALSE
  )
}
#' @export
ft_vector_slicer.ml_connect_pipeline <- ft_vector_slicer_impl
#' @export
ft_vector_slicer.pyspark_connection <- ft_vector_slicer_impl
#' @export
ft_vector_slicer.tbl_pyspark <- ft_vector_slicer_impl

# ------------------------------- Word2Vec -------------------------------------
ft_word2vec_impl <- function(
  x,
  input_col = NULL,
  output_col = NULL,
  vector_size = 100,
  min_count = 5,
  max_sentence_length = 1000,
  num_partitions = 1,
  step_size = 0.025,
  max_iter = 1,
  seed = NULL,
  uid = NULL,
  ...
) {
  ml_process_transformer(c(as.list(environment()), list(...)), "Word2Vec", TRUE)
}
#' @export
ft_word2vec.ml_connect_pipeline <- ft_word2vec_impl
#' @export
ft_word2vec.pyspark_connection <- ft_word2vec_impl
#' @export
ft_word2vec.tbl_pyspark <- ft_word2vec_impl

#' Installs PySpark and Python dependencies
#' @param version Version of 'pyspark' to install. Defaults to `NULL`. If `NULL`,
#'   it will check against PyPi to get the current library version.
#' @param envname The name of the Python Environment to use to install the
#'   Python libraries. Defaults to `NULL.` If `NULL`, a name will automatically
#'   be assigned based on the version that will be installed
#' @param python_version The minimum required version of Python to use to create
#' the Python environment. Defaults to `NULL`. If `NULL`, it will check against
#' PyPi to get the minimum required Python version.
#' @param new_env If `TRUE`, any existing Python virtual environment and/or
#'   Conda environment specified by `envname` is deleted first.
#' @param method The installation method to use. If creating a new environment,
#'   `"auto"` (the default) is equivalent to `"virtualenv"`. Otherwise `"auto"`
#'   infers the installation method based on the type of Python environment
#'   specified by `envname`.
#' @param ... Passed on to [`reticulate::py_install()`]
#' @param as_job Runs the installation if using this function within the
#' RStudio IDE.
#' @param install_ml Installs ML related Python libraries. Defaults to TRUE. This
#' is mainly for machines with limited storage to avoid installing the rather
#' large 'torch' library if the ML features are not going to be used. This will
#' apply to any environment backed by 'Spark' version 3.5 or above.
#' @returns It returns no value to the R session. This function purpose is to
#' create the 'Python' environment, and install the appropriate set of 'Python'
#' libraries inside the new environment. During runtime, this function will send
#' messages to the console describing the steps that the function is
#' taking. For example, it will let the user know if it is getting the latest
#' version of the Python library from 'PyPi.org', and the result of such
#' query.
#' @export
install_pyspark <- function(
  version = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = TRUE,
  method = c("auto", "virtualenv", "conda"),
  as_job = TRUE,
  install_ml = FALSE,
  ...
) {
  install_as_job(
    main_library = "pyspark",
    spark_method = "pyspark_connect",
    backend = "pyspark",
    ml_version = "3.5",
    version = version,
    envname = envname,
    python_version = python_version,
    new_env = new_env,
    method = method,
    as_job = as_job,
    install_ml = install_ml,
    ... = ...
  )
}

#' Installs Databricks Connect and Python dependencies
#' @param version Version of 'databricks.connect' to install. Defaults to `NULL`.
#'  If `NULL`, it will check against PyPi to get the current library version.
#' @param cluster_id Target of the cluster ID that will be used with.
#' If provided, this value will be used to extract the cluster's
#' version
#' @rdname install_pyspark
#' @export
install_databricks <- function(
  version = NULL,
  cluster_id = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = TRUE,
  method = c("auto", "virtualenv", "conda"),
  as_job = TRUE,
  install_ml = FALSE,
  ...
) {
  if (!is.null(version) && !is.null(cluster_id)) {
    cli_div(theme = cli_colors())
    cli_alert_warning(
      paste0(
        "{.header Will use the value from }{.emph 'version'}, ",
        "{.header and ignoring }{.emph 'cluster_id'}"
      )
    )
    cli_end()
  }

  if (is.null(envname)) {
    if (is.null(version) && !is.null(cluster_id)) {
      version <- databricks_dbr_version(
        cluster_id = cluster_id,
        host = databricks_host(),
        token = databricks_token()
      )
    }
  }

  install_as_job(
    main_library = "databricks-connect",
    spark_method = "databricks_connect",
    backend = "databricks",
    ml_version = "14.1",
    version = version,
    envname = envname,
    python_version = python_version,
    new_env = new_env,
    method = method,
    as_job = as_job,
    install_ml = install_ml,
    ... = ...
  )
}

install_as_job <- function(
  main_library = NULL,
  spark_method = NULL,
  backend = NULL,
  ml_version = NULL,
  version = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = NULL,
  method = c("auto", "virtualenv", "conda"),
  as_job = TRUE,
  install_ml = TRUE,
  ...
) {
  args <- c(as.list(environment()), list(...))
  if (as_job && check_rstudio()) {
    install_code <- build_job_code(args)
    job_name <- paste0(
      "Installing '",
      main_library,
      "' version '",
      version,
      "'"
    )
    temp_file <- tempfile()
    writeLines(install_code, temp_file)
    invisible(
      jobRunScript(path = temp_file, name = job_name)
    )
    cli_div(theme = cli_colors())
    cli_alert_success("{.header Running installation as a RStudio job }")
    cli_end()
  } else {
    install_environment(
      main_library = main_library,
      spark_method = spark_method,
      backend = backend,
      ml_version = ml_version,
      version = version,
      envname = envname,
      python_version = python_version,
      new_env = new_env,
      method = method,
      install_ml = install_ml,
      ... = ...
    )
  }
}

install_environment <- function(
  main_library = NULL,
  spark_method = NULL,
  backend = NULL,
  ml_version = NULL,
  version = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = NULL,
  method = c("auto", "virtualenv", "conda"),
  install_ml = FALSE,
  install_packages = NULL,
  ...
) {
  cli_div(theme = cli_colors())
  library_info <- python_library_info(main_library, version)

  if (!is.null(library_info)) {
    if (is.null(python_version)) {
      python_version <- library_info$requires_python
    }
    version <- library_info$version
    ver_name <- version
  } else {
    if (!is.null(version)) {
      ver_name <- version_prep(version)
      if (version == ver_name) {
        version <- paste0(version, ".*")
      }
    } else {
      cli_abort(
        c(
          "No `version` provided, and none could be found",
          " " = "Please run again with a valid version number"
        ),
        call = NULL
      )
    }
  }
  python_number <- sub(">", "", python_version)
  python_number <- sub("=", "", python_number)
  python_number <- trimws(python_number)

  add_torch <- TRUE
  if (is.null(envname)) {
    ver_compare <- compareVersion(
      as.character(ver_name),
      as.character(ml_version)
    )
    if (ver_compare < 0) {
      add_torch <- FALSE
    }
    envname <- use_envname(
      backend = backend,
      version = ver_name,
      main_library = main_library,
      ask_if_not_installed = FALSE,
      python_version = python_version
    )
  }
  cli_alert_success(
    "{.header Automatically naming the environment:}{.emph '{envname}'}"
  )

  packages <- c(
    paste0(main_library, "==", version),
    "pandas!=2.1.0", # deprecation warnings
    "PyArrow",
    "grpcio",
    "google-api-python-client",
    "grpcio_status",
    "databricks-sdk",
    "zstandard"
  )

  if (add_torch && install_ml) {
    packages <- c(packages, pysparklyr_env$ml_libraries)
  }

  method <- match.arg(method)

  if (new_env) {
    if (method %in% c("auto", "virtualenv")) {
      tryCatch(virtualenv_remove(envname, confirm = FALSE), error = identity)
    }
    if (method %in% c("auto", "conda")) {
      conda <- list(...)$conda %||% "auto"
      while (
        !inherits(
          tryCatch(conda_python(envname, conda), error = identity),
          "error"
        )
      ) {
        conda_remove(envname, conda = conda)
      }
    }
  }
  if (
    new_env && method != "conda" && is.null(virtualenv_starter(python_version))
  ) {
    cli_abort(c(
      paste0(
        "{.header Python version} {.emph '{python_number}'}",
        " {.header or higher is required by some libraries.}"
      ),
      " " = paste0(
        "Use: {.run reticulate::install_python",
        "(version = '{python_number}:latest')} to install."
      )
    ))
  }

  if (dir.exists("/databricks/")) {
    # https://github.com/mlverse/pysparklyr/issues/11
    op <- options("reticulate.virtualenv.module" = "virtualenv")
    on.exit(options(op))
  }

  # conda_install() doesn't accept a version constraint for python_version
  if (method == "conda") {
    python_version <- python_number
  }

  if (!is.null(install_packages)) {
    packages <- install_packages
  }

  py_install(
    packages = packages,
    envname = envname,
    method = method,
    python_version = python_version,
    pip = TRUE,
    ...
  )
}

#' Lists installed Python libraries
#' @param list_all Flag that indicates to display all of the installed packages
#' or only the top two, namely, `pyspark` and `databricks.connect`
#' @returns Returns no value, only sends information to the console. The
#' information includes the current versions of 'sparklyr', and 'pysparklyr',
#' as well as the 'Python' environment currently loaded.
#' @export
installed_components <- function(list_all = FALSE) {
  pkgs <- py_list_packages(py_exe())
  db <- pkgs$package == "databricks-connect"
  ps <- pkgs$package == "pyspark"
  sel <- db | ps
  if (!list_all) {
    new_pkgs <- pkgs[sel, ]
  } else {
    new_pkgs <- rbind(pkgs[sel, ], pkgs[!sel, ])
  }
  cli_div(theme = cli_colors())
  cli_h3("R packages")
  cli_bullets(c(
    "*" = "{.header {.code sparklyr} ({packageVersion('sparklyr')}})"
  ))
  cli_bullets(c(
    "*" = "{.header {.code pysparklyr} ({packageVersion('pysparklyr')}})"
  ))
  cli_bullets(c(
    "*" = "{.header {.code reticulate} ({packageVersion('reticulate')}})"
  ))
  cli_h3("Python executable")
  cli_text("{.header {py_exe()}}")
  cli_h3("Python libraries")
  for (i in seq_len(nrow(new_pkgs))) {
    curr_row <- new_pkgs[i, ]
    cli_bullets(c(
      "*" = "{.header {curr_row$package} ({.header {curr_row$version}})}"
    ))
  }
  cli_end()
  invisible()
}

python_library_info <- function(
  library_name,
  library_version = NULL,
  verbose = TRUE,
  fail = TRUE,
  timeout = 2
) {
  msg_fail <- NULL
  msg_done <- NULL
  ret <- NULL
  if (verbose) {
    cli_div(theme = cli_colors())
    cli_progress_step(
      "{.header Retrieving version from PyPi.org}",
      msg_done = paste0(
        "{.header PyPi specs:} {.emph '{ret$name}'} {.header version} {ret$version},",
        " {.header requires Python }{ret$requires_python}"
      ),
      msg_failed = "{.header {msg_fail}}"
    )
  }

  resp <- query_pypi(library_name, library_version, timeout)

  if (inherits(resp, "try-error")) {
    # Not catastrophic, it will simply try to use the upstream name and version
    # provided by the user
    msg_fail <- "Failed to contact PyPi.org"
    if (verbose) {
      cli_progress_done(result = "failed")
    }
    ret <- NULL
  } else {
    if (!is.null(resp)) {
      # Happy path :D
      ret <- resp$info
      cli_progress_done()
    } else {
      msg_abort <- "{.header Library }{.emph {library_name}} {.header not found.}"
      msg_fail <- glue("Python library '{library_name}' not found")
      if (!is.null(library_version)) {
        # Quering PyPi again to see if at least the library name is valid
        resp2 <- query_pypi(library_name, timeout = timeout)
        if (!is.null(resp2)) {
          msg_fail <- glue(
            "Version '{library_version}' for '{library_name}' not found"
          )
          msg_abort <- c(
            "Version {.emph '{library_version}'} is not valid for {.emph '{library_name}'}",
            "i" = "{.header The most recent, valid, version is} {.emph '{resp2$info$version}'}"
          )
        }
      }
      if (!fail) {
        if (verbose) {
          cli_progress_done(result = "failed")
        }
        return(NULL)
      } else {
        if (verbose) {
          cli_progress_done(result = "clear")
          cli_progress_cleanup()
        }
        cli_abort(msg_abort, call = NULL)
      }
    }
  }
  ret
  # For possible future use
  # "https://packagemanager.posit.co/__api__/repos/5/packages/{library_name}"
}

query_pypi <- function(library_name, library_version = NULL, timeout) {
  url <- paste0("https://pypi.org/pypi/", library_name)
  if (!is.null(library_version)) {
    url <- paste0(url, "/", library_version)
  }
  url <- paste0(url, "/json")
  resp <- try(
    {
      tryCatch(
        url |>
          request() |>
          req_timeout(timeout) |>
          req_perform() |>
          resp_body_json(),
        httr2_http_404 = function(cnd) NULL
      )
    },
    silent = TRUE
  )
  resp
}

version_prep <- function(version) {
  if (version == "latest") {
    return(version)
  }
  version <- as.character(version)
  ver <- version |>
    strsplit("\\.") |>
    unlist()

  ver_name <- NULL
  ver_len <- length(ver)

  if (ver_len == 1) {
    cli_abort(c(
      "{.emph '{version}' }{.header is not a valid version}",
      "{.header - Please provide major & minor version (e.g. {version}.0) }"
    ))
  }

  if (ver_len == 0) {
    cli_abort("{.emph '{version}' }{.header is not a valid version}")
  }

  out <- paste0(ver[1:2], collapse = ".")

  if (ver_len > 3) {
    cli_abort(c(
      "{.emph '{version}' }{.header contains too many version levels.}",
      "{.header - Please provide a major/minor version (e.g. {out})}"
    ))
  }

  out
}

check_rstudio <- function() {
  check_rstudio <- try(RStudio.Version(), silent = TRUE)
  if (inherits(check_rstudio, "try-error")) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

check_interactive <- function() interactive()

build_job_code <- function(args) {
  args$as_job <- NULL
  args$method <- args$method[[1]]
  arg_list <- args |>
    imap(\(.x, .y) {
      if (inherits(.x, "character")) {
        x <- paste0("\"", .x, "\"")
      } else {
        x <- .x
      }
      paste0(.y, " = ", x)
    }) |>
    as.character() |>
    paste0(collapse = ", ")
  paste0(
    "pysparklyr:::install_environment(",
    arg_list,
    ")"
  )
}

new_env <- use_new_test_env()
test_that("It checks with PyPi if library version is NULL", {
  withr::with_envvar(
    new = c("WORKON_HOME" = new_env),
    {
      local_mocked_bindings(py_install = function(...) list(...))
      expect_message(
        install_environment(
          main_library = "pyspark",
          spark_method = "pyspark_connect",
          backend = "pyspark",
          ml_version = "3.5",
          new_env = FALSE,
          python = Sys.which("python")
        ),
        "Retrieving version from PyPi.org"
      )
    }
  )
})

test_that("Adds the ML libraries when prompted", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      local_mocked_bindings(py_install = function(...) list(...))
      x <- install_environment(
        main_library = "pyspark",
        spark_method = "pyspark_connect",
        backend = "pyspark",
        version = "3.5",
        ml_version = "3.5",
        new_env = FALSE,
        python = Sys.which("python"),
        # Arg(s) being tested
        install_ml = TRUE
      )
      x <- x[names(x) != "python_version"]
      expect_snapshot(x)
    }
  )
})

test_that("Fails when passing an invalid library version", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      local_mocked_bindings(py_install = function(...) list(...))
      expect_error(
        install_environment(
          main_library = "pyspark",
          new_env = FALSE,
          python = Sys.which("python"),
          # Arg(s) being tested
          version = "0.1",
        ),
        "Version '0.1' is not valid for 'pyspark'"
      )
    }
  )
})


test_that("install_as_job() is able to run", {
  local_mocked_bindings(
    check_rstudio = function(...) TRUE,
    jobRunScript = function(...) invisible(),
    install_environment = function(...) list(...)
  )
  expect_message(
    install_as_job(),
    "Running installation as a RStudio job"
  )
  expect_snapshot(
    install_as_job(as_job = TRUE)
  )
})

test_that("Installation runs even if no response from PyPi", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      local_mocked_bindings(
        py_install = function(...) list(...),
        python_library_info = function(...) NULL
      )
      x <- install_environment(
        main_library = "pyspark",
        spark_method = "pyspark_connect",
        backend = "pyspark",
        version = "3.5",
        ml_version = "3.5",
        new_env = FALSE,
        python = Sys.which("python")
      )
      x <- x[names(x) != "python_version"]
      expect_snapshot(x)
    }
  )
})

test_that("Fails when no library version is provided, and nothing comes back from PyPi", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      local_mocked_bindings(
        py_install = function(...) list(...),
        python_library_info = function(...) NULL
      )
      expect_error(
        install_environment(
          main_library = "pyspark",
          spark_method = "pyspark_connect",
          backend = "pyspark",
          ml_version = "3.5",
          new_env = FALSE,
          python = Sys.which("python")
        ),
        "No `version` provided, and none could be found"
      )
    }
  )
})

test_that("Fails when non-existent Python version is used", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      local_mocked_bindings(
        virtualenv_starter = function(...) {
          return(NULL)
        }
      )
      expect_error(
        install_environment(
          main_library = "pyspark",
          spark_method = "pyspark_connect",
          backend = "pyspark",
          version = "3.5",
          ml_version = "3.5",
          python = Sys.which("python"),
          # Arg(s) being tested
          new_env = TRUE,
          python_version = "10.1"
        ),
        "Python version '10.1' or higher is required by some libraries."
      )
    }
  )
})

test_that("Can get pypy.org info", {
  info <- python_library_info("databricks.connect", "13.0")
  expect_type(info, "list")
  expect_equal(info$summary, "Databricks Connect Client")

  expect_error(python_library_info("doesnt.exist", ""))
})

test_that("version_prep() outputs what's expected", {
  expect_error(version_prep(""))
  expect_error(version_prep("1"))
  expect_equal(version_prep("1.1"), "1.1")
  expect_equal(version_prep("1.1.1"), "1.1")
  expect_error(version_prep("1.1.1.1"))
})

test_that("Install code is correctly created", {
  expect_snapshot(build_job_code(list(a = 1)))
})

test_that("Databricks installation works", {
  local_mocked_bindings(install_as_job = function(...) list(...))

  out <- install_databricks(version = "14.1")
  expect_snapshot(out)

  expect_message(
    install_databricks(
      version = "14.1",
      cluster = Sys.getenv("DATABRICKS_CLUSTER_ID"),
      as_job = FALSE
    ),
    "Will use the value from 'version', and ignoring 'cluster_id'"
  )

  vcr::local_cassette("databricks-install")

  expect_message(
    withr::with_envvar(
      new = c(
        "DATABRICKS_HOST" = use_test_db_host(),
        "DATABRICKS_TOKEN" = "temptoken"
      ),
      {
        install_databricks(
          cluster = use_test_db_cluster(),
          as_job = FALSE
        )
      }
    )
  )
  expect_snapshot(install_databricks(version = "13.1"))
})

skip_on_ci()

test_that("Databricks installations work", {
  skip("Avoids multiple installations")
  env_paths <- map(1:3, \(.x) fs::path(tempdir(), random_table_name("env")))
  baseenv <- "r-sparklyr-databricks"
  version_1 <- "14.3"
  version_2 <- "13.3"
  cluster_id <- Sys.getenv("DATABRICKS_CLUSTER_ID")
  python_path <- Sys.which("python")

  withr::with_envvar(
    new = c("WORKON_HOME" = env_paths[[1]]),
    {
      install_databricks(
        cluster_id = cluster_id,
        install_ml = FALSE,
        as_job = FALSE,
        python = python_path
      )
      expect_equal(length(find_environments(baseenv)), 1)
    }
  )

  withr::with_envvar(
    new = c("WORKON_HOME" = env_paths[[2]]),
    {
      install_databricks(
        version_1,
        install_ml = FALSE,
        as_job = FALSE,
        python = python_path
      )
      expect_equal(length(find_environments(baseenv)), 1)
    }
  )

  withr::with_envvar(
    new = c("WORKON_HOME" = env_paths[[3]]),
    {
      install_databricks(
        version_1,
        install_ml = FALSE,
        as_job = FALSE,
        python = python_path
      )
      install_databricks(
        version_2,
        install_ml = FALSE,
        as_job = FALSE,
        python = python_path
      )
      install_databricks(
        cluster_id = cluster_id,
        install_ml = FALSE,
        as_job = FALSE,
        python = python_path
      )
      expect_equal(length(find_environments(baseenv)), 3)
    }
  )

  map(env_paths, fs::dir_delete)
})

# ---------------------------- tidyr pivot_longer ------------------------------
.strsep <- function(x, sep) {
  nchar <- nchar(x)
  pos <- map(sep, function(i) {
    if (i >= 0) {
      return(i)
    }
    pmax(0, nchar + i)
  })
  pos <- c(list(0), pos, list(nchar))

  map(seq_len(length(pos) - 1), function(i) {
    substr(x, pos[[i]] + 1, pos[[i + 1]])
  })
}

.slice_match <- function(x, i) {
  structure(
    x[i],
    match.length = attr(x, "match.length")[i],
    index.type = attr(x, "index.type"),
    useBytes = attr(x, "useBytes")
  )
}

.str_split_n <- function(x, pattern, n_max = -1) {
  m <- gregexpr(pattern, x, perl = TRUE)
  if (n_max > 0) {
    m <- lapply(m, function(x) .slice_match(x, seq_along(x) < n_max))
  }
  regmatches(x, m, invert = TRUE)
}

.list_indices <- function(x, max = 20) {
  if (length(x) > max) {
    x <- c(x[seq_len(max)], "...")
  }

  paste(x, collapse = ", ")
}

.simplify_pieces <- function(pieces, p, fill_left) {
  too_sml <- NULL
  too_big <- NULL
  n <- length(pieces)

  out <- lapply(seq(p), function(x) rep(NA, n))
  for (i in seq_along(pieces)) {
    x <- pieces[[i]]
    if (!(length(x) == 1 && is.na(x[[1]]))) {
      if (length(x) > p) {
        too_big <- c(too_big, i)

        for (j in seq(p)) {
          out[[j]][[i]] <- x[[j]]
        }
      } else if (length(x) < p) {
        too_sml <- c(too_sml, i)

        gap <- p - length(x)
        for (j in seq(p)) {
          if (fill_left) {
            out[[j]][[i]] <- (if (j >= gap) x[[j - gap]] else NA)
          } else {
            out[[j]][[i]] <- (if (j < length(x)) x[[j]] else NA)
          }
        }
      } else {
        for (j in seq(p)) {
          out[[j]][[i]] <- x[[j]]
        }
      }
    }
  }

  structure(list(
    strings = out,
    too_big = too_big,
    too_sml = too_sml
  ))
}

.str_split_fixed <- function(value, sep, n, extra = "warn", fill = "warn") {
  if (extra == "error") {
    warn(glue(
      "`extra = \"error\"` is deprecated. \\
       Please use `extra = \"warn\"` instead"
    ))
    extra <- "warn"
  }

  extra <- arg_match(extra, c("warn", "merge", "drop"))
  fill <- arg_match(fill, c("warn", "left", "right"))

  n_max <- if (extra == "merge") n else -1L
  pieces <- .str_split_n(value, sep, n_max = n_max)

  simp <- .simplify_pieces(pieces, n, fill == "left")

  n_big <- length(simp$too_big)
  if (extra == "warn" && n_big > 0) {
    idx <- .list_indices(simp$too_big)
    warn(glue(
      "Expected {n} pieces. ",
      "Additional pieces discarded in {n_big} rows [{idx}]."
    ))
  }

  n_sml <- length(simp$too_sml)
  if (fill == "warn" && n_sml > 0) {
    idx <- .list_indices(simp$too_sml)
    warn(glue(
      "Expected {n} pieces. ",
      "Missing pieces filled with `NA` in {n_sml} rows [{idx}]."
    ))
  }

  simp$strings
}

.str_separate <- function(x, into, sep, extra = "warn", fill = "warn") {
  if (!is.character(into)) {
    abort("`into` must be a character vector")
  }

  if (is.numeric(sep)) {
    out <- .strsep(x, sep)
  } else if (is_character(sep)) {
    out <- .str_split_fixed(x, sep, length(into), extra = extra, fill = fill)
  } else {
    abort("`sep` must be either numeric or character")
  }

  names(out) <- as_utf8_character(into)
  out <- out[!is.na(names(out))]

  as_tibble(out)
}

.str_extract <- function(x, into, regex, convert = FALSE) {
  stopifnot(
    is_string(regex),
    is_character(into)
  )

  out <- .str_match_first(x, regex)
  if (length(out) != length(into)) {
    stop(
      "`regex` should define ",
      length(into),
      " groups; ",
      ncol(matches),
      " found.",
      call. = FALSE
    )
  }

  # Handle duplicated names
  if (anyDuplicated(into)) {
    pieces <- split(out, into)
    into <- names(pieces)
    out <- map(pieces, pmap_chr, paste0, sep = "")
  }

  into <- as_utf8_character(into)

  non_na_into <- !is.na(into)
  out <- out[non_na_into]
  names(out) <- into[non_na_into]

  out <- as_tibble(out)

  if (convert) {
    out <- map(out, type.convert, as.is = TRUE)
  }

  out
}

.str_match_first <- function(string, regex) {
  loc <- regexpr(regex, string, perl = TRUE)
  loc <- .group_loc(loc)

  out <- lapply(
    seq_len(loc$matches),
    function(i) substr(string, loc$start[, i], loc$end[, i])
  )
  out[-1]
}

.group_loc <- function(x) {
  start <- cbind(as.vector(x), attr(x, "capture.start"))
  end <- start + cbind(attr(x, "match.length"), attr(x, "capture.length")) - 1L

  no_match <- start == -1L
  start[no_match] <- NA
  end[no_match] <- NA

  list(matches = ncol(start), start = start, end = end)
}

test_that("Internal functions work", {
  expect_snapshot(.strsep("domino", 3))
  expect_equal(.slice_match("hello", 1), "hello")
  expect_snapshot(.str_split_n("hello", "l", n_max = 3))
  expect_snapshot(.list_indices(c("one", "two", "three"), 2))
  expect_snapshot(.simplify_pieces(c("one", "two", "three"), 2, FALSE))
  expect_warning(.str_split_fixed("hello", "l", 1))
})

#' @export
sdf_schema.tbl_pyspark <- function(
  x,
  expand_nested_cols = NULL,
  expand_struct_cols = NULL
) {
  check_arg_supported(expand_nested_cols)
  check_arg_supported(expand_struct_cols)

  x_sdf <- python_sdf(x)
  x_schema <- x_sdf$schema
  x_schema$fields |>
    map(\(.x) {
      dt <- as.character(.x$dataType)
      dt <- substr(dt, 1, nchar(dt) - 2)
      list(name = .x$name, type = dt)
    }) |>
    set_names(x_schema$names)
}

databricks_host <- function(host = NULL, fail = TRUE) {
  if (!is.null(host)) {
    return(named_host(host, "argument"))
  }
  env_host <- Sys.getenv("DATABRICKS_HOST", unset = NA)
  connect_host <- Sys.getenv("CONNECT_DATABRICKS_HOST", unset = NA)
  if (!is.na(env_host)) {
    host <- named_host(env_host, "environment")
  }
  if (!is.na(connect_host)) {
    host <- named_host(connect_host, "environment_connect")
  }
  if (is.null(host)) {
    if (fail) {
      cli_abort(c(
        paste0(
          "No Host URL was provided, and",
          "the environment variable 'DATABRICKS_HOST' is not set."
        ),
        "Please add your Host to 'DATABRICKS_HOST' inside your .Renviron file."
      ))
    } else {
      host <- ""
    }
  }
  host
}

databricks_token <- function(token = NULL, fail = FALSE) {
  # if token provided, return
  # otherwise, search for token:
  # DATABRICKS_TOKEN > CONNECT_DATABRICKS_TOKEN > .rs.api.getDatabricksToken
  if (!is.null(token)) {
    return(set_names(token, "argument"))
  }
  # Checks the Environment Variables
  env_token <- Sys.getenv("DATABRICKS_TOKEN", unset = NA)
  if (!is.na(env_token)) {
    return(set_names(env_token, "environment"))
  }
  connect_token <- Sys.getenv("CONNECT_DATABRICKS_TOKEN", unset = NA)
  if (!is.na(connect_token)) {
    return(set_names(connect_token, "environment_connect"))
  }

  # Checks for OAuth Databricks token inside the Workbench configuration file
  wb_token <- workbench_databricks_token()
  if (!is.null(wb_token)) {
    return(set_names(wb_token, "workbench"))
  }

  # Checks for Posit Connect OAuth
  if (has_viewer_token(databricks_host())) {
    return(set_names(connect_viewer_token(databricks_host()), "viewer"))
  }

  if (fail) {
    rlang::abort(c(
      paste0(
        "No authentication token was identified: \n",
        " - No 'DATABRICKS_TOKEN' environment variable found \n",
        " - Not passed as a function argument"
      ),
      "Please add your Token to 'DATABRICKS_TOKEN' inside your .Renviron file."
    ))
  }
  return(NULL)
}

databricks_sdk_client <- function(
  sdk,
  host,
  token,
  serverless = FALSE,
  cluster_id = NULL,
  profile = NULL
) {
  # SDK behaviour
  # https://databricks-sdk-py.readthedocs.io/en/latest/authentication.html#default-authentication-flow
  if (token == "") {
    databricks_desktop_login(host = host, profile = profile)
    auth_type <- "databricks-cli"
  } else {
    auth_type <- "pat"
  }

  if (serverless) {
    serverless_compute_id <- "auto"
    cluster_id <- NULL
  } else {
    serverless_compute_id <- NULL
  }

  config <- sdk$core$Config(
    host = host,
    token = token,
    auth_type = auth_type,
    profile = profile,
    serverless_compute_id = serverless_compute_id,
    cluster_id = cluster_id
  )

  sdk$WorkspaceClient(config = config)
}

databricks_dbr_version_name <- function(
  cluster_id,
  client = NULL,
  host = NULL,
  token = NULL,
  silent = FALSE
) {
  bullets <- NULL
  version <- NULL
  cluster_info <- databricks_dbr_info(
    cluster_id = cluster_id,
    client = client,
    host = host,
    token = token,
    silent = silent
  )
  cluster_name <- substr(cluster_info$cluster_name, 1, 100)
  version <- databricks_extract_version(cluster_info)
  cli_progress_done()
  cli_end()
  list(version = version, name = cluster_name)
}

databricks_extract_version <- function(x) {
  sp_version <- x$spark_version
  if (!is.null(sp_version)) {
    sp_sep <- unlist(strsplit(sp_version, "\\."))
    version <- paste0(sp_sep[1], ".", sp_sep[2])
  } else {
    version <- ""
  }
  version
}

databricks_dbr_info <- function(
  cluster_id,
  client = NULL,
  host = NULL,
  token = NULL,
  silent = FALSE
) {
  cli_div(theme = cli_colors())

  if (!silent) {
    cli_progress_step(
      msg = "Retrieving info for cluster:}{.emph '{cluster_id}'",
      msg_done = "{.header Cluster:} {.emph '{cluster_id}'} | {.header DBR: }{.emph '{version}'}",
      msg_failed = "Failed contacting:}{.emph '{cluster_id}'"
    )
  }

  # client is used preferentially
  # host/token only should be used on initial setup
  out <- databricks_cluster_get(cluster_id, client, host, token)

  if (inherits(out, "try-error")) {
    cli_div(theme = cli_colors())
    invalid_host <- NULL
    invalid_token <- NULL
    invalid_cluster <- NULL
    invalid_msg <- " <<--- Possibly invalid"
    if (grepl("HTTP 404 Not Found", out)) {
      parse_host <- url_parse(client$config$host)
      invalid_host <- invalid_msg
      if (!is.null(parse_host$path)) {
        invalid_host <- glue(
          "<<--- Likely cause, last part in the URL: \"{parse_host$path}\""
        )
      }
    }
    if (grepl("HTTP 401 Unauthorized", out)) {
      invalid_token <- invalid_msg
    }
    if (grepl("HTTP 400 Bad Request", out)) {
      invalid_cluster <- invalid_msg
    }

    if (as.character(substr(out, 1, 26)) == "Error in req_perform(.) : ") {
      out <- substr(out, 27, nchar(out))
    }
    if (!silent) {
      cli_progress_done(result = "failed")
    }
    cli_abort(
      c(
        "{.header Connection with Databricks failed: }\"{trimws(out)}\"",
        "{.class  - Host: {.emph '{host}'}} {invalid_host}",
        "{.class  - Cluster ID: {.emph '{cluster_id}'}} {invalid_cluster}",
        "{.class  - Token: {.emph '<REDACTED>'}} {invalid_token}"
      ),
      call = NULL
    )
    out <- list()
  } else {
    version <- databricks_extract_version(out)
  }
  if (!silent) {
    cli_progress_done()
  }
  cli_end()
  out
}

databricks_dbr_version <- function(
  cluster_id,
  client = NULL,
  host = NULL,
  token = NULL
) {
  vn <- databricks_dbr_version_name(
    cluster_id = cluster_id,
    client = client,
    host = host,
    token = token
  )
  vn$version
}

databricks_cluster_get <- function(
  cluster_id,
  client = NULL,
  host = NULL,
  token = NULL
) {
  if (!is.null(client)) {
    try(
      client$clusters$get(cluster_id = cluster_id)$as_dict(),
      silent = TRUE
    )
  } else {
    try(
      paste0(
        host,
        "/api/2.0/clusters/get"
      ) |>
        request() |>
        req_auth_bearer_token(token) |>
        req_body_json(list(cluster_id = cluster_id)) |>
        req_perform() |>
        resp_body_json(),
      silent = TRUE
    )
  }
}

databricks_dbr_error <- function(error) {
  error_split <- error |>
    as.character() |>
    strsplit("\n\t") |>
    unlist()

  error_start <- substr(error_split, 1, 9)

  status_error <- NULL
  if (any(error_start == "status = ")) {
    status_error <- error_split[error_start == "status = "]
  }

  status_details <- NULL
  if (any(error_start == "details =")) {
    status_details <- error_split[error_start == "details ="]
  }

  status_tip <- NULL
  if (!is.null(status_error)) {
    if (grepl("UNAVAILABLE", status_error)) {
      status_tip <- "Possible cause = The cluster is not running, or not accessible"
    }
    if (grepl("FAILED_PRECONDITION", status_error)) {
      status_tip <- "Possible cause = The cluster is initializing. Try again later"
    }
  } else {
    status_error <- error
  }

  rlang::abort(
    c(
      "Spark connection error",
      status_tip,
      status_error,
      status_details
    )
  )
}

# from httr2
is_hosted_session <- function() {
  if (nzchar(Sys.getenv("COLAB_RELEASE_TAG"))) {
    return(TRUE)
  }
  Sys.getenv("RSTUDIO_PROGRAM_MODE") == "server" &&
    !grepl("localhost", Sys.getenv("RSTUDIO_HTTP_REFERER"), fixed = TRUE)
}

databricks_desktop_login <- function(host = NULL, profile = NULL) {
  # host takes priority over profile
  if (!is.null(host)) {
    method <- "--host"
    value <- host
  } else if (!is.null(profile)) {
    method <- "--profile"
    value <- profile
  } else {
    # todo rlang error?
    stop("must specifiy `host` or `profile`, neither were set")
  }

  cli_path <- Sys.getenv("DATABRICKS_CLI_PATH", "databricks")
  if (!is_hosted_session() && nchar(Sys.which(cli_path)) != 0) {
    # When on desktop, try using the Databricks CLI for auth.
    output <- suppressWarnings(
      system2(
        cli_path,
        c("auth", "login", method, value),
        stdout = TRUE,
        stderr = TRUE
      )
    )
  }
}

sanitize_host <- function(url, silent = FALSE) {
  url <- ifelse(!grepl("^https?://", url), paste0("https://", url), url)
  parsed_url <- url_parse(url)
  new_url <- url_parse("http://localhost")
  if (is.null(parsed_url$scheme)) {
    new_url$scheme <- "https"
    if (!is.null(parsed_url$path) && is.null(parsed_url$hostname)) {
      new_url$hostname <- parsed_url$path
    }
  } else {
    new_url$scheme <- parsed_url$scheme
    new_url$hostname <- parsed_url$hostname
  }
  new_url$path <- NULL
  ret <- url_build(new_url)
  if (endsWith(ret, "/")) {
    ret <- substr(ret, 1, nchar(ret) - 1)
  }
  ret
}

named_host <- function(url, name) {
  set_names(sanitize_host(url), name)
}

databricks_dbr_python <- function(version) {
  if (version >= "16.0") {
    "3.12"
  } else if (version >= "15.0") {
    "3.11"
  } else {
    "3.10"
  }
}

# https://docs.databricks.com/aws/en/release-notes/serverless#supported-spark-configuration-parameters
allowed_serverless_configs <- function() {
  c(
    "spark.sql.legacy.timeParserPolicy",
    "spark.sql.session.timeZone",
    "spark.sql.shuffle.partitions",
    "spark.sql.ansi.enabled"
  )
}

workbench_databricks_token <- function(host, cfg_file) {
  if (missing(host)) {
    host <- databricks_host(fail = FALSE)
  }
  if (missing(cfg_file)) {
    cfg_file <- Sys.getenv("DATABRICKS_CONFIG_FILE")
    if (!grepl("posit-workbench", cfg_file, fixed = TRUE)) {
      return(NULL)
    }
  }
  cfg <- readLines(cfg_file)
  # We don't attempt a full parse of the INI syntax supported by Databricks
  # config files, instead relying on the fact that this particular file will
  # always contain only one section.
  if (!any(grepl(host, cfg, fixed = TRUE))) {
    # The configuration doesn't actually apply to this host.
    return(NULL)
  }
  line <- grepl("token = ", cfg, fixed = TRUE)
  token <- gsub("token = ", "", cfg[line])
  if (nchar(token) == 0) {
    return(NULL)
  }
  token
}

test_coverage_enable <- function() {
  Sys.setenv("CODE_COVERAGE" = "true")
}

expect_same_remote_result <- function(.data, pipeline) {
  sc <- use_test_spark_connect()
  temp_name <- random_table_name("test_")
  spark_data <- dplyr::copy_to(sc, .data, temp_name)

  local <- pipeline(.data)

  remote <- try(
    spark_data |>
      pipeline() |>
      collect()
  )

  if (inherits(remote, "try-error")) {
    expect_equal(remote[[1]], "")
  } else {
    expect_equal(local, remote, ignore_attr = TRUE)
  }

  DBI::dbRemoveTable(sc, temp_name)
}

testthat_tbl <- function(name, data = NULL, repartition = 0L) {
  sc <- use_test_spark_connect()

  tbl <- tryCatch(dplyr::tbl(sc, name), error = identity)
  if (inherits(tbl, "error")) {
    if (is.null(data)) {
      data <- eval(as.name(name), envir = parent.frame())
    }
    tbl <- dplyr::copy_to(sc, data, name = name, repartition = repartition)
  }

  tbl
}

random_table_name <- function(prefix) {
  paste0(prefix, paste0(floor(runif(10, 0, 10)), collapse = ""))
}


skip_spark_min_version <- function(version) {
  sc <- use_test_spark_connect()
  sp_version <- spark_version(sc)
  comp_ver <- compareVersion(as.character(version), sp_version)
  if (comp_ver != -1) {
    skip(glue("Skips on Spark version {sp_version}"))
  }
}

skip_spark_max_version <- function(version) {
  sc <- use_test_spark_connect()
  sp_version <- spark_version(sc)
  comp_ver <- compareVersion(as.character(version), sp_version)
  if (comp_ver != 1) {
    skip(glue("Skips on Spark version {sp_version}"))
  }
}


test_remove_python_envs <- function(x = "") {
  found <- find_environments(x)
  cli_inform("Environments found: {length(found)}")

  invisible(
    lapply(
      found,
      function(x) {
        try(virtualenv_remove(x, confirm = FALSE), silent = TRUE)
      }
    )
  )

  invisible(
    lapply(
      found,
      function(x) {
        try(conda_remove(x), silent = TRUE)
      }
    )
  )
}

tests_disable_all <- function() {
  r_scripts <- dir_ls(test_path(), glob = "*.R")
  test_scripts <- r_scripts[substr(path_file(r_scripts), 1, 5) == "test-"]
  map(
    test_scripts,
    \(.x) {
      ln <- readLines(.x)
      writeLines(c("skip(\"temp\")", ln), con = .x)
    }
  )
}

tests_enable_all <- function() {
  r_scripts <- dir_ls(test_path(), glob = "*.R")
  test_scripts <- r_scripts[substr(path_file(r_scripts), 1, 5) == "test-"]
  map(
    test_scripts,
    \(.x) {
      ln <- readLines(.x)
      new_ln <- ln[ln != "skip(\"temp\")"]
      writeLines(new_ln, con = .x)
    }
  )
}

test_databricks_cluster_id <- function() {
  Sys.getenv("DATABRICKS_CLUSTER_ID")
}

test_databricks_cluster_version <- function() {
  if (is.null(.test_env$dbr)) {
    dbr <- try(
      databricks_dbr_version(
        cluster_id = test_databricks_cluster_id(),
        host = databricks_host(),
        token = databricks_token()
      ),
      silent = TRUE
    )
    if (inherits(dbr, "try-error")) {
      dbr <- "99.9"
    }
    .test_env$dbr <- dbr
  }
  .test_env$dbr
}

test_databricks_stump_env <- function() {
  env_name <- use_envname(
    version = test_databricks_cluster_version(),
    backend = "databricks",
    main_library = "databricks-connect"
  )
  env_path <- path(use_test_env(), env_name)
  if (names(env_name) != "exact") {
    py_install(
      package = "numpy",
      envname = env_path,
      pip = TRUE
    )
  }
  path(env_path, "bin", "python")
}

test_that("DBR error code returns as expected", {
  error <- paste0(
    "SparkConnectGrpcException('<_InactiveRpcError of RPC that terminated with:",
    "\n\tstatus = StatusCode.UNAVAILABLE\n\tdetails = 'RESOURCE_DOES_NOT_EXIST: No",
    " cluster found matching: asdfasdf'\n\tdebug_error_string = 'UNKNOWN:Error",
    " received from peer  {grpc_message:'RESOURCE_DOES_NOT_EXIST: No cluster",
    " found matching: asdfasdf', grpc_status:5,",
    " created_time:'2023-10-02T12:14:52.379226-05:00'}'\n>')"
  )

  expect_snapshot(databricks_dbr_error(error), error = TRUE)

  expect_snapshot(databricks_dbr_error(""), error = TRUE)
})

test_that("Databricks Host works", {
  expect_named(
    withr::with_envvar(
      new = c("DATABRICKS_HOST" = "test"),
      {
        databricks_host()
      }
    ),
    "environment"
  )
  expect_named(databricks_host("thisisatest"), "argument")

  expect_error(
    withr::with_envvar(
      new = c("DATABRICKS_HOST" = NA, "DATABRICKS_TOKEN" = NA),
      {
        databricks_host()
      }
    ),
    "No Host URL was provided"
  )

  expect_named(
    withr::with_envvar(
      new = c("DATABRICKS_HOST" = NA, "CONNECT_DATABRICKS_HOST" = "testing"),
      {
        databricks_host()
      }
    ),
    "environment_connect"
  )
})

test_that("Databricks Token works", {
  expect_named(
    withr::with_envvar(
      new = c("DATABRICKS_TOKEN" = "test"),
      {
        databricks_token()
      }
    ),
    "environment"
  )

  expect_named(databricks_token("thisisatest"), "argument")

  expect_error(
    withr::with_envvar(
      new = c("DATABRICKS_HOST" = NA, "DATABRICKS_TOKEN" = NA),
      {
        databricks_token(fail = TRUE)
      }
    ),
    "No authentication token was identified"
  )

  expect_named(
    withr::with_envvar(
      new = c("DATABRICKS_TOKEN" = NA, "CONNECT_DATABRICKS_TOKEN" = "testing"),
      {
        databricks_token()
      }
    ),
    "environment_connect"
  )
})

test_that("Get cluster version", {
  vcr::local_cassette("databricks-cluster-version")
  expect_equal(
    databricks_dbr_version(
      host = use_test_db_host(),
      token = "",
      cluster_id = use_test_db_cluster()
    ),
    "17.3"
  )
})

test_that("Cluster info runs as expected", {
  expect_error(databricks_dbr_version(""))
})

test_that("Misc tests", {
  expect_silent(databricks_desktop_login("xxxx", "xxxxx"))
  expect_snapshot(allowed_serverless_configs())
})

test_that("DBR Python comes back as expected", {
  expect_equal(databricks_dbr_python("17.0"), "3.12")
  expect_equal(databricks_dbr_python("15.0"), "3.11")
  expect_equal(databricks_dbr_python("14.0"), "3.10")
})

test_that("Host sanitation works", {
  clean_url <- "https://cloud.databricks.com"
  expect_equal(sanitize_host("cloud.databricks.com"), clean_url)
  expect_equal(sanitize_host("https://cloud.databricks.com"), clean_url)
  expect_equal(sanitize_host("https://cloud.databricks.com/"), clean_url)
  expect_equal(sanitize_host("https://cloud.databricks.com/?o=123#"), clean_url)
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import cli
#' @import DBI
#' @import dbplyr
#' @import fs
#' @import glue
#' @import httr2
#' @import reticulate
#' @importFrom dplyr filter mutate
#' @importFrom dplyr sample_n sample_frac slice_sample select tbl_ptype group_by
#' @importFrom dplyr tbl collect tibble same_src compute as_tibble group_vars
#' @importFrom glue glue
#' @importFrom lifecycle deprecated
#' @importFrom methods new is setOldClass
#' @importFrom processx process
#' @importFrom purrr map_chr discard transpose reduce map2 list_rbind flatten
#' @importFrom purrr map_lgl map_chr map pmap_chr imap iwalk walk imap_chr
#' @importFrom purrr as_mapper map_vec
#' @importFrom rlang enquo `!!` `!!!` quo_is_null sym warn abort `%||%`
#' @importFrom rlang exec arg_match as_utf8_character is_missing
#' @importFrom rlang is_string is_character parse_exprs set_names
#' @importFrom rstudioapi jobRunScript showQuestion getSourceEditorContext
#' @importFrom sparklyr ft_bucketed_random_projection_lsh ft_count_vectorizer ft_dct
#' @importFrom sparklyr ft_discrete_cosine_transform ft_vector_assembler ft_elementwise_product
#' @importFrom sparklyr ft_feature_hasher ft_idf ft_imputer ft_index_to_string
#' @importFrom sparklyr ft_interaction ft_min_max_scaler ft_minhash_lsh ft_ngram
#' @importFrom sparklyr ft_one_hot_encoder ft_pca ft_polynomial_expansion
#' @importFrom sparklyr ft_quantile_discretizer ft_regex_tokenizer ft_robust_scaler
#' @importFrom sparklyr ft_sql_transformer ft_vector_indexer ft_vector_slicer ft_word2vec
#' @importFrom sparklyr ft_tokenizer ft_stop_words_remover ft_hashing_tf ft_normalizer
#' @importFrom sparklyr ml_binary_classification_evaluator ml_cross_validator
#' @importFrom sparklyr ml_clustering_evaluator ml_regression_evaluator ml_kmeans
#' @importFrom sparklyr ml_decision_tree_regressor ml_bisecting_kmeans ml_aft_survival_regression
#' @importFrom sparklyr ml_gbt_classifier ml_gbt_regressor ml_isotonic_regression
#' @importFrom sparklyr ml_generalized_linear_regression
#' @importFrom sparklyr ml_linear_regression ft_r_formula ft_binarizer ft_bucketizer
#' @importFrom sparklyr ml_logistic_regression ft_standard_scaler ft_max_abs_scaler
#' @importFrom sparklyr ml_multiclass_classification_evaluator ml_decision_tree_classifier
#' @importFrom sparklyr ml_pipeline ml_predict ml_transform ml_fit
#' @importFrom sparklyr ml_random_forest_classifier ml_random_forest_regressor ft_string_indexer
#' @importFrom sparklyr ml_save ml_load spark_jobj spark_install_find spark_apply
#' @importFrom sparklyr sdf_copy_to spark_connect_method spark_log random_string
#' @importFrom sparklyr spark_connection connection_is_open hive_context
#' @importFrom sparklyr spark_dataframe spark_web sdf_register sdf_schema
#' @importFrom sparklyr spark_ide_connection_updated spark_version
#' @importFrom sparklyr spark_ide_objects spark_ide_columns sdf_read_column
#' @importFrom sparklyr spark_read_csv spark_read_parquet spark_read_text
#' @importFrom sparklyr spark_read_json spark_read_orc
#' @importFrom sparklyr spark_session invoke invoke_new invoke_static
#' @importFrom sparklyr spark_table_name spark_integ_test_skip spark_ide_preview
#' @importFrom sparklyr spark_write_csv spark_write_parquet spark_write_text
#' @importFrom sparklyr spark_write_delta
#' @importFrom sparklyr spark_write_orc spark_write_json spark_write_table
#' @importFrom stats terms setNames
#' @importFrom tidyr pivot_longer
#' @importFrom tidyselect matches
#' @importFrom tidyselect tidyselect_data_has_predicates
#' @importFrom utils capture.output installed.packages menu
#' @importFrom utils head type.convert compareVersion
#' @importFrom uuid UUIDgenerate
#' @importFrom vctrs vec_as_names
#' @importFrom connectcreds has_viewer_token connect_viewer_token
## usethis namespace: end

## mockable bindings: start
## mockable bindings: end
NULL

globalVariables("RStudio.Version")

pysparklyr_env <- new.env()
pysparklyr_env$temp_prefix <- "sparklyr_tmp_"
temp_prefix <- function() pysparklyr_env$temp_prefix
pysparklyr_env$ml_libraries <- c("torch", "torcheval", "scikit-learn")
pysparklyr_env$artifacts <- list()

initialize_connection <- function(
  conn,
  master_label,
  con_class,
  cluster_id = NULL,
  serverless = FALSE,
  method = NULL,
  config = NULL,
  misc = NULL,
  quote = NULL
) {
  warnings <- import("warnings")
  warnings$filterwarnings(
    "ignore",
    message = "is_datetime64tz_dtype is deprecated",
    module = "pyspark"
  )
  warnings$filterwarnings(
    "ignore",
    message = "is_categorical_dtype is deprecated",
    module = "pyspark"
  )
  warnings$filterwarnings(
    "ignore",
    message = "'SparkSession' object has no attribute 'setLocalProperty'",
    module = "pyspark"
  )
  warnings$filterwarnings(
    "ignore",
    message = "Index.format is deprecated and will be removed in a future version"
  )

  session <- conn$getOrCreate()
  get_version <- try(session$version, silent = TRUE)
  if (inherits(get_version, "try-error")) {
    databricks_dbr_error(get_version)
  }

  if (!is.null(config)) {
    # remove sparklyr default configs (at least for now)
    config <- list_diff(config, sparklyr::spark_config())

    # remove any other sparklyr configs
    config <- config[!grepl("^sparklyr", names(config))]

    # add pyspark configs if not on serverless
    # drop spark sql configs that are unsupported on serverless
    if (!serverless) {
      config <- c(config, pyspark_config())
    } else {
      # ensure only `spark.sql.<...>` configs are considered
      sql_configs <- names(config)[grepl("^spark.sql", names(config))]
      allowed_confs <- sql_configs %in% allowed_serverless_configs()

      # drop configs that aren't supported
      dropped <- sql_configs[!allowed_confs]

      if (length(dropped) > 0) {
        cli::cli_alert_warning(
          text = "Unsupported configs on serverless were dropped:"
        )
        cli::cli_li(paste0("{.envvar ", dropped, "}"))
        cli::cli_end()
      }
      config <- config[!names(config) %in% dropped]
    }

    iwalk(config, function(x, y) session$conf$set(y, x))
  }

  # do we need this `spark_context` object?
  spark_context <- list(spark_context = session)

  # browser()
  structure(
    list(
      master = master_label,
      cluster_id = cluster_id,
      config = config,
      method = method,
      session = session,
      state = spark_context,
      serverless = serverless,
      misc = misc,
      quote = quote,
      con = structure(list(), class = c("spark_connection", "DBIConnection")),
      connection_id = UUIDgenerate()
    ),
    class = c(
      con_class,
      "pyspark_connection",
      "spark_connection",
      "DBIConnection"
    )
  )
}

python_conn <- function(x) {
  py_object <- "python.builtin.object"
  ret <- NULL
  if (inherits(x$state$spark_context, py_object)) {
    ret <- x$state$spark_context
  }
  if (is.null(ret) && inherits(x[[1]]$session$sparkSession, py_object)) {
    ret <- x[[1]]$session$sparkSession
  }
  if (is.null(ret)) {
    cli_abort("Could not match Python Connection to: {class(x)}")
  }
  ret
}

build_user_agent <- function() {
  product <- NULL
  in_rstudio <- FALSE
  in_connect <- FALSE

  env_var <- Sys.getenv("SPARK_CONNECT_USER_AGENT", unset = NA)
  if (!is.na(env_var)) {
    return(env_var)
  }

  if (current_product_connect()) {
    product <- "posit-connect"
  }

  if (is.null(product)) {
    pr <- NULL
    if (check_rstudio()) {
      rstudio_version <- int_rstudio_version()
      edition <- rstudio_version$edition
      mode <- rstudio_version$mode
      version <- rstudio_version$long_version
      pr <- "rstudio"
      if (length(edition) == 0) {
        edition <- ""
      }

      if (length(mode) == 0) {
        mode <- ""
      }
      if (edition == "Professional") {
        if (mode == "server") {
          pr <- "workbench-rstudio"
        } else {
          pr <- "rstudio-pro"
        }
      }
      if (Sys.getenv("R_CONFIG_ACTIVE") == "rstudio_cloud") {
        pr <- "cloud-rstudio"
      }
    }
    if (Sys.getenv("POSITRON", unset = "0") == "1") {
      pr <- "positron"
      version <- Sys.getenv("POSITRON_VERSION", unset = NA)
    }
    if (!is.null(pr)) {
      product <- glue("posit-{pr}/{version}")
    }
  }

  glue(
    "sparklyr/{packageVersion('sparklyr')} {product}",
    .null = NULL
  )
}

int_rstudio_version <- function() {
  out <- try(RStudio.Version(), silent = TRUE)
  if (!inherits(out, "try-error")) {
    return(out)
  }
  return(NULL)
}

connection_label <- function(x) {
  x <- x[[1]]
  ret <- "Connection"
  method <- NULL
  con <- spark_connection(x)
  if (is.null(con)) {
    method <- x
  } else {
    method <- con$method
  }
  if (!is.null(method)) {
    if (method == "spark_connect" | method == "pyspark") {
      ret <- "Spark Connect"
    }
    if (method == "databricks_connect" | method == "databricks") {
      ret <- "Databricks Connect"
    }
  }
  ret
}

#' Read Spark configuration
#' @returns A list object with the initial configuration that will be used for
#' the Connect session.
#' @export
pyspark_config <- function() {
  list(
    "spark.sql.session.localRelationCacheThreshold" = 1048576L,
    "spark.sql.execution.arrow.pyspark.enabled" = "true",
    "spark.sql.execution.arrow.sparkr.enabled" = "true"
  )
}

test_that("User agent builder basic", {
  withr::with_envvar(
    new = c("RSTUDIO_PRODUCT" = NA),
    {
      x <- build_user_agent()
      expect_equal(substr(x, 1, 8), "sparklyr")
    }
  )
})

test_that("User agent builder acknowledges when product is Connectr", {
  withr::with_envvar(
    new = c("RSTUDIO_PRODUCT" = "CONNECT"),
    {
      expect_equal(
        build_user_agent(),
        glue("sparklyr/{packageVersion('sparklyr')} posit-connect")
      )
    }
  )
})

test_that("User agent builder works with Connect env var", {
  withr::with_envvar(
    new = c("SPARK_CONNECT_USER_AGENT" = "testagent"),
    {
      expect_equal(build_user_agent(), "testagent")
    }
  )
})

use_test_rs_version <- function() {
  x <- list()
  x$mode <- "server"
  x$edition <- "Professional"
  x$long_version <- "2023.12.0"
  x
}

test_that("User agent works on RStudio Workbench", {
  withr::with_envvar(
    new = c("RSTUDIO_PRODUCT" = NA),
    {
      local_mocked_bindings(
        check_rstudio = function(...) TRUE,
        int_rstudio_version = function(...) use_test_rs_version()
      )
      expect_equal(
        build_user_agent(),
        glue(
          "sparklyr/{packageVersion('sparklyr')} posit-workbench-rstudio/2023.12.0"
        )
      )
    }
  )
})

test_that("User agent works on RStudio Pro", {
  withr::with_envvar(
    new = c("RSTUDIO_PRODUCT" = NA),
    {
      local_mocked_bindings(
        check_rstudio = function(...) TRUE,
        int_rstudio_version = function(...) {
          x <- use_test_rs_version()
          x$mode <- "desktop"
          x
        }
      )
      expect_equal(
        build_user_agent(),
        glue(
          "sparklyr/{packageVersion('sparklyr')} posit-rstudio-pro/2023.12.0"
        )
      )
    }
  )
})


test_that("User agent builder works with Connect env var", {
  withr::with_envvar(
    new = c("R_CONFIG_ACTIVE" = "rstudio_cloud"),
    {
      local_mocked_bindings(
        check_rstudio = function(...) TRUE,
        int_rstudio_version = function(...) {
          x <- use_test_rs_version()
          x$mode <- "desktop"
          x
        }
      )
      expect_equal(
        build_user_agent(),
        glue(
          "sparklyr/{packageVersion('sparklyr')} posit-cloud-rstudio/2023.12.0"
        )
      )
    }
  )
})

test_that("User agent builder for Positron", {
  withr::with_envvar(
    new = c(
      "RSTUDIO_PRODUCT" = NA,
      "POSITRON" = "1",
      "POSITRON_VERSION" = "12345"
    ),
    {
      expect_equal(
        build_user_agent(),
        glue(
          "sparklyr/{packageVersion('sparklyr')} posit-positron/12345"
        )
      )
    }
  )
})

import_check <- function(x, envname, silent = FALSE) {
  cli_div(theme = cli_colors())
  env_found <- !is.na(envname)
  env_loaded <- NA
  is_uv <- FALSE
  is_reticulate <- FALSE
  uv_managed <- NA
  look_for_env <- TRUE

  env_found_names <- names(env_found) %||% ""

  if (env_found_names == "unavailable") {
    env_found <- FALSE
  }

  cli_msg <- "Attempting to load {.emph '{envname}'}"

  env_is_file <- FALSE
  if (envname == "") {
    env_path <- ""
    is_reticulate <- TRUE
    uv_managed <- "false"
    envname <- "Deferring to {reticulate}"
    cli_msg <- NULL
  } else {
    if (file_exists(envname)) {
      env_contents <- try(dir_ls(envname), silent = TRUE)
      # Workaround to fs::is_file() not working
      if (inherits(env_contents, "try-error")) {
        env_is_file <- TRUE
      }
      env_path <- envname
    } else {
      env_path <- env_python(envname)
      if (is.na(env_path)) {
        env_path <- ""
        is_uv <- TRUE
        uv_managed <- "true"
        envname <- "Managed `uv` environment"
        cli_msg <- NULL
      }
    }
  }

  if (!silent) {
    cli_progress_step(
      msg = cli_msg,
      msg_done = "{.header Python environment:} {.emph '{envname}'}",
      msg_failed = "Problem using {.emph '{envname}'}"
    )
  }

  if (env_is_file) {
    look_for_env <- FALSE
    use_python(envname)
    env_loaded <- TRUE
  }

  if (look_for_env) {
    if (py_available()) {
      # If there is a Python environment already loaded
      if (env_found) {
        find_env <- env_python(envname)
        if (is.na(find_env)) {
          find_env <- ""
        }
        if (find_env == py_exe()) {
          env_loaded <- TRUE
        } else {
          env_loaded <- FALSE
        }
      }
    } else {
      # If there is NO Python environment already loaded
      if (env_found) {
        # If the envname is found, we try to use it
        envir_type <- env_type(envname)
        if (is.na(envir_type)) {
          envir_type <- ""
        }
        if (envir_type == "virtualenv") {
          try_load <- try(use_virtualenv(envname), silent = TRUE)
        } else {
          try_load <- try(use_condaenv(envname), silent = TRUE)
        }
        if (!inherits(try_load, "try-error")) {
          env_loaded <- TRUE
        }
      }
    }
  }

  if (is.na(env_loaded)) {
    if (env_path != "") {
      py_executable <- py_exe() %||% ""
      env_loaded <- env_path == py_executable
    } else {
      env_loaded <- FALSE
    }
  }

  withr::with_envvar(
    c("RETICULATE_USE_MANAGED_VENV" = uv_managed),
    out <- try(import(x), silent = TRUE)
  )

  if (inherits(out, "try-error")) {
    if (env_found) {
      if (env_loaded) {
        # found & loaded
        rlang::abort(
          paste0(
            "Issue loading Python library '",
            x,
            "'\n",
            as.character(out)
          ),
          call = NULL
        )
      } else {
        cli_progress_done(result = "failed")
        cli_abort(
          c(
            "{.emph '{x}' }{.header is not available current Python environment.}",
            if (!is_reticulate) {
              c(
                paste(
                  "{.header - The }{.emph '{envname}'} {.header Python",
                  " environment is installed, but it is not loaded.}"
                ),
                paste(
                  "{.header - Restart your R session, and avoid",
                  " initializing Python before using} {.emph '{x}'}"
                )
              )
            }
          ),
          call = NULL
        )
      }
    } else {
      cli_abort(
        c(
          "{.emph '{x}' }{.header is not available current Python environment.}",
          paste(
            "- The {.emph '{envname}'} Python environment is not installed."
          )
        ),
        call = NULL
      )
    }
    cli_alert_danger(glue("`reticulate` error:\n {out[[1]]}"))
  } else {
    if (!env_loaded && !is_uv && !is_reticulate) {
      cli_progress_done(result = "failed")
      cli_bullets(c(
        " " = "{.header A different Python is already loaded: }{.emph '{py_exe()}'}",
        " " = "{.emph '{x}'} {.header was found and loaded from that environment}"
      ))
      cli_end()
    }
  }
  out
}

env_type <- function(envname) {
  ret <- NA
  if (virtualenv_exists(envname)) {
    ret <- "virtualenv"
  }
  if (is.na(ret)) {
    check_conda <- try(conda_python(envname), silent = TRUE)
    if (!inherits(check_conda, "try-error")) {
      ret <- "conda"
    }
  }
  ret
}

env_python <- function(envname) {
  ret <- NA
  type <- env_type(envname)
  if (!is.na(type)) {
    if (type == "virtualenv") {
      ret <- virtualenv_python(envname)
    }
    if (type == "conda") {
      ret <- conda_python(envname)
    }
  }
  ret
}

test_that("import_check() exception tests", {
  sc <- use_test_spark_connect()

  exe_location <- reticulate::py_exe()
  exe_folders <- unlist(strsplit(exe_location, "/"))
  env_name <- exe_folders[length(exe_folders) - 2]

  expect_error(import_check("not.real", env_name))
})

.test_env <- new.env()
.test_env$sc <- NULL
.test_env$lr_model <- NULL
.test_env$env <- NULL
.test_env$started <- NULL
.test_env$dbr <- NULL
.test_env$databricks <- NULL
.test_env$tune_grid <- NULL

use_test_env <- function() {
  if (is.null(.test_env$env)) {
    .test_env$env <- use_new_test_env()
  }
  .test_env$env
}

use_new_test_env <- function() {
  base <- fs::path_expand("~/test-spark")
  x <- fs::path(base, random_table_name("env"))
  fs::dir_create(x)
}

use_new_temp_env <- function() {
  fs::path(tempdir(), random_table_name("env"))
}

use_test_version_spark <- function() {
  version <- Sys.getenv("SPARK_VERSION", unset = NA)
  if (is.na(version)) {
    version <- "4.0"
  }
  version
}

use_test_scala_spark <- function() {
  version <- Sys.getenv("SCALA_VERSION", unset = NA)
  if (is.na(version)) {
    version <- "2.13"
  }
  version
}

use_test_python_version <- function() {
  version <- Sys.getenv("PYTHON_VERSION", unset = NA)
  if (is.na(version)) {
    version <- "3.10"
  }
  version
}

use_test_connect_start <- function() {
  if (is.null(.test_env$started)) {
    env_path <- use_test_python_environment()
    version <- use_test_version_spark()
    cli_h2("Spark Connect: {version}")
    cli_inform("PYTHON_VERSION_MISMATCH: {env_path}")
    cli_inform("PYSPARK_PYTHON: {env_path}")
    cli_inform("WORKON_HOME: {use_test_env()}")
    cli_inform("SCALA_VERSION: {use_test_scala_spark()}")
    cli_inform("PYTHON_VERSION: {use_test_python_version()}")
    cli_h2("")
    print(reticulate::py_list_packages())

    withr::with_envvar(
      new = c(
        "PYSPARK_PYTHON" = env_path,
        "PYTHON_VERSION_MISMATCH" = env_path,
        "PYSPARK_DRIVER_PYTHON" = env_path,
        "PYTHON_VERSION" = use_test_python_version()
      ),
      {
        spark_connect_service_start(
          version = version,
          scala_version = use_test_scala_spark()
        )
      }
    )
    .test_env$started <- 0
    cli_inform("-- Waiting a couple of seconds to let the service start")
    Sys.sleep(2)
    cli_inform("-- Done")
  } else {
    invisible()
  }
}

use_test_spark_connect <- function() {
  if (is.null(.test_env$sc)) {
    conf <- pyspark_config()
    conf$spark.python.worker.memory <- "50m"
    use_test_connect_start()
    cli_h1("Connecting to Spark cluster")
    withr::with_envvar(
      new = c(
        "WORKON_HOME" = use_test_env(),
        "PYSPARK_PYTHON" = use_test_python_environment(),
        "PYTHON_VERSION_MISMATCH" = use_test_python_environment(),
        "PYSPARK_DRIVER_PYTHON" = use_test_python_environment()
      ),
      {
        .test_env$sc <- sparklyr::spark_connect(
          master = "sc://localhost",
          method = "spark_connect",
          version = use_test_version_spark(),
          config = conf,
          envname = use_test_python_environment()
        )
      }
    )
  }
  .test_env$sc
}

use_test_table_mtcars <- function() {
  sc <- use_test_spark_connect()
  if (!"mtcars" %in% dbListTables(sc)) {
    ret <- dplyr::copy_to(sc, mtcars, overwrite = TRUE)
  } else {
    ret <- dplyr::tbl(sc, "mtcars")
  }
  ret
}

use_test_table_ovarian <- function() {
  use_test_table(
    x = readRDS(test_path("_data/ovarian.rds")),
    name = "ovarian"
  )
}

use_test_table_iris <- function() {
  sc <- use_test_spark_connect()
  if (!"iris" %in% dbListTables(sc)) {
    ret <- dplyr::copy_to(sc, iris, overwrite = TRUE)
  } else {
    ret <- dplyr::tbl(sc, "iris")
  }
  ret
}

use_test_table <- function(x, name) {
  sc <- use_test_spark_connect()
  if (!name %in% dbListTables(sc)) {
    ret <- dplyr::copy_to(sc, x, name = name, overwrite = TRUE)
  } else {
    ret <- dplyr::tbl(sc, name)
  }
  ret
}

use_test_table_simple <- function() {
  use_test_table(
    x = data.frame(x = c(2, 2, 4, NA, 4), y = 1:5),
    name = "simple_table"
  )
}

use_test_table_reviews <- function() {
  reviews <- data.frame(
    x = "This has been the best TV I've ever used. Great screen, and sound."
  )
  sc <- use_test_spark_connect()
  if (!"reviews" %in% dbListTables(sc)) {
    ret <- dplyr::copy_to(sc, reviews, overwrite = TRUE)
  } else {
    ret <- dplyr::tbl(sc, "reviews")
  }
  ret
}

use_test_lr_model <- function() {
  if (is.null(.test_env$lr_model)) {
    tbl_mtcars <- use_test_table_mtcars()
    .test_env$lr_model <- ml_logistic_regression(
      tbl_mtcars,
      am ~ .,
      max_iter = 10
    )
  }
  .test_env$lr_model
}

use_test_python_environment <- function(use_uv = TRUE) {
  if (is.null(.test_env$target)) {
    withr::with_envvar(
      new = c("WORKON_HOME" = use_test_env()),
      {
        version <- use_test_version_spark()
        if (use_uv) {
          reticulate::py_require("torch")
          env <- use_envname(
            backend = "pyspark",
            main_library = "pyspark",
            version = version,
            messages = TRUE,
            python_version = use_test_python_version(),
            ask_if_not_installed = FALSE
          )
          reticulate::import("pyspark")
          target <- reticulate::py_exe()
        } else {
          env_avail <- names(env)
          target <- path(use_test_env(), env)
          if (!dir_exists(target)) {
            if (env_avail != "exact") {
              cli_h1("Creating Python environment")
              install_pyspark(
                version = version,
                as_job = FALSE,
                python = Sys.which("python"),
                install_ml = FALSE
              )
              env <- use_envname(
                backend = "pyspark",
                main_library = "pyspark",
                version = version
              )
            }
          }
          target <- path(env, "bin", "python")
        }
      }
    )
    .test_env$target <- target
  }
  .test_env$target
}

use_test_ml_installed <- function() {
  ml_libraries <- pysparklyr_env$ml_libraries
  installed_libraries <- py_list_packages()$package
  find_ml <- map_lgl(ml_libraries, \(.x) .x %in% installed_libraries)
  all(find_ml)
}

use_test_install_ml <- function() {
  if (!use_test_ml_installed()) {
    try(
      py_install(pysparklyr_env$ml_libraries),
      silent = TRUE
    )
    try(
      py_require(pysparklyr_env$ml_libraries),
      silent = TRUE
    )
  }
}

# Needed to ensure `vcr` tests are 100% consistent with the
# cluster number from which they were originally ran from
use_test_db_cluster <- function() {
  return("1117-200943-l1nvd2bl")
}
use_test_db_host <- function() {
  return("https://adb-3256282566390055.15.azuredatabricks.net")
}

skip_ml_missing <- function() {
  skip_if(!use_test_ml_installed(), "ML Python libraries not installed")
}

skip_ml_not_missing <- function() {
  skip_if(use_test_ml_installed(), "ML Python libraries installed")
}

test_that("spark_apply() works", {
  tbl_mtcars <- use_test_table_mtcars()
  expect_s3_class(
    spark_apply(
      tbl_mtcars,
      nrow,
      group_by = "am",
      columns = "am double, x long"
    ),
    "tbl_spark"
  )
  skip_spark_min_version(3.5)
  expect_s3_class(
    spark_apply(tbl_mtcars, function(x) x),
    "tbl_spark"
  )
  # expect_s3_class(
  #   spark_apply(tbl_mtcars, function(x) x, fetch_result_as_sdf = FALSE),
  #   "data.frame"
  # )
  expect_s3_class(
    spark_apply(tbl_mtcars, function(x) x, arrow_max_records_per_batch = 5000),
    "tbl_spark"
  )
  expect_s3_class(
    spark_apply(tbl_mtcars, ~.x),
    "tbl_spark"
  )
  expect_s3_class(
    spark_apply(dplyr::filter(tbl_mtcars, am == 0), ~.x),
    "tbl_spark"
  )
  expect_error(
    spark_apply(tbl_mtcars, function(x, y) y, context = 1)
  )
  expect_s3_class(
    spark_apply(tbl_mtcars, function(x, y) y, context = 1, group_by = "am"),
    "tbl_spark"
  )
})

test_that("Errors are output by specific params", {
  expect_error(spark_apply(tbl_mtcars, nrow, packages = "test"))
  expect_error(spark_apply(tbl_mtcars, nrow, context = ""))
  expect_error(spark_apply(tbl_mtcars, nrow, auto_deps = TRUE))
  expect_error(spark_apply(tbl_mtcars, nrow, partition_index_param = "test"))
})

test_that("head() works", {
  tbl_mtcars <- use_test_table_mtcars()

  expect_equal(
    tbl_mtcars |>
      head(5) |>
      collect() |>
      nrow(),
    5
  )
})

test_that("copy_to() works", {
  tbl_ordered <- use_test_table_mtcars() |>
    arrange(mpg, qsec, hp)

  expect_snapshot(tbl_ordered)

  expect_snapshot(print(head(tbl_ordered)))
})

test_that("Sampling functions works", {
  tbl_mtcars <- use_test_table_mtcars()
  tbl_n <- tbl_mtcars |>
    sample_n(5) |>
    collect() |>
    count() |>
    pull()

  expect_equal(tbl_n, 5)

  tbl_frac <- tbl_mtcars |>
    sample_frac(0.2) |>
    collect() |>
    count() |>
    pull()

  expect_lt(tbl_frac, 30)

  expect_error(
    sample_frac(tbl_mtcars, size = 0.5, weight = mpg),
    "`weight` is not supported"
  )
})

test_that("Misc functions", {
  tbl_am <- use_test_table_mtcars() |>
    group_by(am) |>
    filter(mpg == max(mpg, na.rm = TRUE)) |>
    select(am)

  expect_silent(compute(tbl_am, name = "am"))

  expect_silent(compute(tbl_am, name = NULL))

  tbl_join <- use_test_table_mtcars() |>
    left_join(tbl_am, by = "am") |>
    arrange(mpg, qsec, hp)

  expect_error(tbl_ptype(tbl_am))

  expect_snapshot(tbl_am[1])

  expect_snapshot(tbl_join)
})

test_that("sdf_copy_to() workks", {
  sc <- use_test_spark_connect()
  test_df <- data.frame(a = 1:1000, b = 1:1000)
  expect_s3_class(
    sdf_copy_to(sc, test_df, name = "test_df"),
    "tbl_pyspark"
  )
  expect_error(
    sdf_copy_to(sc, test_df, name = "test_df"),
    "Temp table test_df already exists, use `overwrite = TRUE` to replace"
  )
  expect_s3_class(
    sdf_copy_to(sc, test_df, name = "test_df", overwrite = TRUE),
    "tbl_pyspark"
  )
  expect_s3_class(
    sdf_copy_to(
      sc,
      test_df,
      name = "test_df",
      overwrite = TRUE,
      repartition = 2
    ),
    "tbl_pyspark"
  )
})

test_that("sdf_register() works", {
  tbl_mtcars <- use_test_table_mtcars()
  sc <- use_test_spark_connect()
  obj <- python_sdf(tbl_mtcars)
  py_obj <- as_spark_pyobj(obj, sc)
  expect_s3_class(
    sdf_register(py_obj),
    "tbl_pyspark"
  )
})

test_that("I() works", {
  tbl_mtcars <- use_test_table_mtcars()
  sc <- use_test_spark_connect()
  expect_s3_class(
    {
      tbl(sc, I("mtcars")) |>
        head() |>
        collect()
    },
    "data.frame"
  )
})

test_that("Writing requirements works", {
  req_folder <- path(tempdir(), random_table_name("req"))
  req_file <- path(req_folder, "requirements.txt")
  dir_create(req_folder)
  sc <- use_test_spark_connect()

  expect_message(
    requirements_write(destfile = req_file),
    "Python library requirements"
  )

  expect_error(
    requirements_write(destfile = req_file),
    "File 'requirements.txt' already exists"
  )

  expect_message(
    requirements_write(destfile = req_file, overwrite = TRUE),
    "Python library requirements"
  )
})

skip_spark_min_version("4")
test_that("Databricks Connect", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env(),
      "DATABRICKS_HOST" = "testhost",
      "DATABRICKS_TOKEN" = "testtoken"
    ),
    {
      local_mocked_bindings(
        initialize_connection = function(...) {
          return(list(...))
        },
        databricks_dbr_info = function(...) {
          return(list(cluster_name = "test_host"))
        },
        import_check = function(...) {
          out <- list()
          out$DatabricksSession$builder$sdkConfig <- function(...) {
            x <- list()
            x$userAgent <- function(...) {
              return(list())
            }
            x
          }
          out$core$Config <- function(...) {
            x <- list()
            x$token <- "testtoken"
          }
          out
        },
        databricks_sdk_client = function(...) {
          return(NULL)
        }
      )

      sc_out <- spark_connect_method.spark_method_databricks_connect(
        method = "databricks_connect",
        master = NULL,
        envname = use_test_python_environment(),
        version = "17.1",
        cluster_id = "test_cluster"
      )
      expect_equal(sc_out$master_label, "test_host (test_cluster)")
      expect_equal(sc_out$cluster_id, "test_cluster")
    }
  )
})

test_that("Snowpark Connect (Snowflake)", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env()
    ),
    {
      local_mocked_bindings(
        initialize_connection = function(...) {
          return(list(...))
        },
        import_check = function(...) {
          out <- list()
          out$Session$builder$configs <- function(...) {
            list(...)
          }
          out
        },
        databricks_sdk_client = function(...) {
          return(NULL)
        }
      )
      sc_out <- spark_connect_method.spark_method_snowpark_connect(
        method = "snowpark_connect",
        master = "testaccount",
        connection_parameters = list(
          user = "test@user.com",
          password = "testtoken",
          warehouse = "testwh",
          database = "testdb",
          schema = "testschema"
        )
      )
      expect_snapshot(sc_out)
      expect_error(
        spark_connect_method.spark_method_snowpark_connect(
          method = "snowpark_connect",
          connection_parameters = list(
            user = "test@user.com",
            password = "testtoken",
            warehouse = "testwh",
            database = "testdb",
            schema = "testschema"
          )
        )
      )
    }
  )
})

test_that("installed_components() output properly", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_test_env()),
    {
      expect_message(installed_components())
    }
  )
})

skip_on_ci()
skip_spark_min_version(4.0)

test_that("Binarizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_binarizer(sc, "a", "b")))
  expect_snapshot(class(ft_binarizer(ml_pipeline(sc), "a", "b")))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_binarizer("mpg", "mpg_new", threshold = 20) |>
      use_test_pull(TRUE)
  )
})

test_that("Bucket Random Projection LSH works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_bucketed_random_projection_lsh(sc)))
  expect_snapshot(
    class(ft_bucketed_random_projection_lsh(ml_pipeline(sc)))
  )
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_bucketed_random_projection_lsh("vec_x", "lsh_x", bucket_length = 1) |>
      use_test_pull()
  )
})

test_that("Bucketizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_bucketizer(sc, "a", "b", c(1, 2, 3))))
  expect_snapshot(class(ft_bucketizer(ml_pipeline(sc), "a", "b", c(1, 2, 3))))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_bucketizer("mpg", "mpg_new", splits = c(0, 10, 20, 30, 40)) |>
      use_test_pull(TRUE)
  )
})

test_that("Count vectorizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_count_vectorizer(sc)))
  expect_snapshot(class(ft_count_vectorizer(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer(input_col = "x", output_col = "token_x") |>
      ft_count_vectorizer("token_x", "cv_x") |>
      use_test_pull()
  )
})

test_that("DCT works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_dct(sc)))
  expect_snapshot(class(ft_dct(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_dct("vec_x", "dct_x") |>
      use_test_pull()
  )
})

test_that("Discrete Cosine works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_discrete_cosine_transform(sc)))
  expect_snapshot(class(ft_discrete_cosine_transform(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_discrete_cosine_transform("vec_x", "dct_x") |>
      use_test_pull()
  )
})

test_that("Elementwise Product works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_elementwise_product(sc)))
  expect_snapshot(class(ft_elementwise_product(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_elementwise_product("vec_x", "elm_x", scaling_vec = c(1:3)) |>
      use_test_pull()
  )
})

test_that("Feature Hasher works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_feature_hasher(sc)))
  expect_snapshot(class(ft_feature_hasher(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_feature_hasher(c("mpg", "wt", "cyl")) |>
      use_test_pull()
  )
})

test_that("Hashing TF works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_hashing_tf(ml_pipeline(sc))))
  expect_snapshot(class(ft_hashing_tf(sc)))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer(input_col = "x", output_col = "token_x") |>
      ft_hashing_tf(
        input_col = "token_x",
        output_col = "hashed_x",
        binary = TRUE,
        num_features = 1024
      ) |>
      use_test_pull()
  )
})

test_that("IDF works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_idf(sc)))
  expect_snapshot(class(ft_idf(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_idf("vec_x", "idf_x") |>
      use_test_pull()
  )
})

test_that("Imputer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_imputer(sc)))
  expect_snapshot(class(ft_imputer(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_simple() |>
      ft_imputer(list(c("x")), list(c("new_x"))) |>
      use_test_pull()
  )
})

test_that("Index-to-string works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_index_to_string(ml_pipeline(sc))))
  expect_snapshot(class(ft_index_to_string(sc)))
  expect_snapshot(
    use_test_table_iris() |>
      ft_string_indexer("Species", "species_idx") |>
      ft_index_to_string("species_idx", "species_x") |>
      use_test_pull(TRUE)
  )
})

test_that("Interaction works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_interaction(sc)))
  expect_snapshot(class(ft_interaction(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_interaction(c("mpg", "wt"), c("mpg_wt")) |>
      use_test_pull()
  )
})

test_that("Max Abs Scaler works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_max_abs_scaler(sc)))
  expect_snapshot(class(ft_max_abs_scaler(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_max_abs_scaler("vec_x", "rs_x") |>
      use_test_pull()
  )
})

test_that("Min Hash LSH works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_minhash_lsh(sc)))
  expect_snapshot(class(ft_minhash_lsh(ml_pipeline(sc))))
  expect_snapshot(
    use_test_iris_va() |>
      ft_minhash_lsh("vec_x", "hash_x") |>
      use_test_pull() |>
      round() |>
      table()
  )
})

test_that("N-gram works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_ngram(sc)))
  expect_snapshot(class(ft_ngram(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer("x", "token_x") |>
      ft_ngram("token_x", "ngram_x") |>
      dplyr::pull()
  )
})

test_that("Normalizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_hashing_tf(ml_pipeline(sc))))
  expect_snapshot(class(ft_hashing_tf(sc)))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer(input_col = "x", output_col = "token_x") |>
      ft_stop_words_remover(input_col = "token_x", output_col = "stop_x") |>
      ft_hashing_tf(
        input_col = "stop_x",
        output_col = "hashed_x",
        binary = TRUE,
        num_features = 1024
      ) |>
      ft_normalizer(
        input_col = "hashed_x",
        output_col = "normal_x"
      ) |>
      use_test_pull()
  )
})

test_that("One hot encoder works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_one_hot_encoder(sc)))
  expect_snapshot(class(ft_one_hot_encoder(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_simple() |>
      ft_one_hot_encoder(list(c("y")), list(c("ohe_x"))) |>
      use_test_pull()
  )
})

test_that("PCA works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_pca(sc)))
  expect_snapshot(class(ft_pca(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_pca("vec_x", "pca_x", k = 2) |>
      use_test_pull()
  )
})

test_that("Polynomial expansion works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_polynomial_expansion(sc)))
  expect_snapshot(class(ft_polynomial_expansion(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_polynomial_expansion("vec_x", "pe_x", degree = 2) |>
      use_test_pull()
  )
})

test_that("Quantile discretizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_quantile_discretizer(sc)))
  expect_snapshot(class(ft_quantile_discretizer(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_simple() |>
      ft_quantile_discretizer(c("y"), c("ohe_x")) |>
      use_test_pull()
  )
})

test_that("R Formula works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_r_formula(ml_pipeline(sc))))
  expect_snapshot(class(ft_r_formula(sc)))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_r_formula(mpg ~ ., features_col = "test") |>
      dplyr::select(test) |>
      use_test_pull()
  )
})

test_that("Regex Tokenizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_regex_tokenizer(sc)))
  expect_snapshot(class(ft_regex_tokenizer(ml_pipeline(sc))))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_regex_tokenizer("x", "new_x") |>
      dplyr::pull()
  )
})

test_that("Standard Scaler works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_standard_scaler(sc)))
  expect_snapshot(class(ft_standard_scaler(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_standard_scaler("vec_x", "rs_x") |>
      use_test_pull()
  )
})

test_that("Robust Scaler works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_robust_scaler(sc)))
  expect_snapshot(class(ft_robust_scaler(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_robust_scaler("vec_x", "rs_x") |>
      use_test_pull()
  )
})

test_that("SQL transformer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_sql_transformer(sc)))
  expect_snapshot(class(ft_sql_transformer(ml_pipeline(sc))))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_sql_transformer("select * from __THIS__ where mpg > 20") |>
      use_test_pull()
  )
})

test_that("Stop words remover works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_tokenizer(ml_pipeline(sc))))
  expect_snapshot(class(ft_tokenizer(sc)))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer(input_col = "x", output_col = "token_x") |>
      ft_stop_words_remover(input_col = "token_x", output_col = "stop_x") |>
      dplyr::pull()
  )
})

test_that("String indexer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_string_indexer(ml_pipeline(sc))))
  expect_snapshot(class(ft_string_indexer(sc)))
  expect_snapshot(
    use_test_table_iris() |>
      ft_string_indexer("Species", "species_idx") |>
      use_test_pull(TRUE)
  )
})

test_that("Tokenizer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_tokenizer(ml_pipeline(sc))))
  expect_snapshot(class(ft_tokenizer(sc)))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer(input_col = "x", output_col = "token_x") |>
      dplyr::pull()
  )
})

test_that("Vector assembler works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_vector_assembler(ml_pipeline(sc))))
  expect_snapshot(class(ft_vector_assembler(sc)))
  expect_snapshot(
    use_test_table_mtcars() |>
      ft_vector_assembler(
        input_cols = c("mpg", "wt", "cyl"),
        output_col = "vec_x"
      ) |>
      use_test_pull()
  )
})

test_that("Vector indexer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_vector_indexer(ml_pipeline(sc))))
  expect_snapshot(class(ft_vector_indexer(sc)))
  expect_equal(
    use_test_mtcars_va() |>
      ft_vector_indexer("vec_x", "index_x") |>
      use_test_pull() |>
      nrow(),
    32
  )
})

test_that("Vector slicer works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_vector_slicer(ml_pipeline(sc))))
  expect_snapshot(class(ft_vector_slicer(sc)))
  expect_snapshot(
    use_test_mtcars_va() |>
      ft_vector_slicer("vec_x", "index_x", indices = list(1L)) |>
      use_test_pull()
  )
})

test_that("Word2Vec works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ft_word2vec(ml_pipeline(sc))))
  expect_snapshot(class(ft_word2vec(sc)))
  expect_snapshot(
    use_test_table_reviews() |>
      ft_tokenizer("x", "token_x") |>
      ft_word2vec("token_x", "word_x", min_count = 1) |>
      use_test_pull()
  )
})

skip_spark_min_version(4.0)

test_that("Logistic regression works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_logistic_regression(ml_pipeline(sc))))
  expect_snapshot(class(ml_logistic_regression(sc)))
  model <- use_test_table_mtcars() |>
    ml_logistic_regression(am ~ .)
  expect_snapshot(class(model))
  fitted <- model |>
    ml_predict(use_test_table_mtcars()) |>
    pull()
  expect_snapshot(table(fitted))
})

test_that("Linear regression works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_linear_regression(ml_pipeline(sc))))
  expect_snapshot(class(ml_linear_regression(sc)))
  model <- use_test_table_mtcars() |>
    ml_linear_regression(wt ~ .)
  expect_snapshot(class(model))
  expect_snapshot(model)
  fitted <- model |>
    ml_predict(use_test_table_mtcars()) |>
    use_test_pull() |>
    round(2) |>
    sort()
  expect_snapshot(fitted)
})

test_that("Random Forest Classifier works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_random_forest_classifier(ml_pipeline(sc))))
  expect_snapshot(class(ml_random_forest_classifier(sc)))
  model <- tbl_iris |>
    ft_string_indexer("Species", "species_idx") |>
    ml_random_forest_classifier(species_idx ~ Petal_Length + Petal_Width)
  expect_snapshot(class(model))
  x <- tbl_iris |>
    ft_string_indexer("Species", "species_idx") |>
    ml_predict(model, dataset = _) |>
    use_test_pull() |>
    table()
  expect_length(x, 3)
})

test_that("Random Forest Regressor works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  expect_snapshot(class(ml_random_forest_regressor(ml_pipeline(sc))))
  expect_snapshot(class(ml_random_forest_regressor(sc)))
  model <- tbl_mtcars |>
    ml_random_forest_regressor(mpg ~ ., seed = 100)
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_mtcars |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

test_that("Decision Tree Classifier works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_decision_tree_classifier(ml_pipeline(sc))))
  expect_snapshot(class(ml_decision_tree_classifier(sc)))
  model <- tbl_iris |>
    ft_string_indexer("Species", "species_idx") |>
    ml_decision_tree_classifier(
      species_idx ~ Petal_Length + Petal_Width,
      seed = 100
    )
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_iris |>
    ft_string_indexer("Species", "species_idx") |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_snapshot(table(x$prediction))
})

test_that("Decision Tree Regressor works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  expect_snapshot(class(ml_decision_tree_regressor(ml_pipeline(sc))))
  expect_snapshot(class(ml_decision_tree_regressor(sc)))
  model <- tbl_mtcars |>
    ml_decision_tree_regressor(mpg ~ ., seed = 100)
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_mtcars |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

test_that("Kmeans works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_kmeans(sc)))
  expect_snapshot(class(ml_kmeans(ml_pipeline(sc))))
  model <- tbl_iris |>
    ml_kmeans(Species ~ .)
  preds <- model |>
    ml_predict(tbl_iris) |>
    use_test_pull() |>
    table()
  expect_length(preds, 2)
})

test_that("Bisecting Kmeans works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_bisecting_kmeans(sc)))
  expect_snapshot(class(ml_bisecting_kmeans(ml_pipeline(sc))))
  model <- tbl_iris |>
    ml_bisecting_kmeans(Species ~ .)
  preds <- model |>
    ml_predict(tbl_iris) |>
    dplyr::pull() |>
    table() |>
    sort() |>
    as.vector()
  expect_snapshot(preds)
})

test_that("AFT Survival works", {
  sc <- use_test_spark_connect()
  tbl_ovarian <- use_test_table_ovarian()
  expect_snapshot(class(ml_aft_survival_regression(ml_pipeline(sc))))
  expect_snapshot(class(ml_aft_survival_regression(sc)))
  model <- tbl_ovarian |>
    ml_aft_survival_regression(
      futime ~ ecog_ps + rx + age + resid_ds,
      censor_col = "fustat"
    )
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_ovarian |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

test_that("GBT classifiers works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  expect_snapshot(class(ml_gbt_classifier(ml_pipeline(sc))))
  expect_snapshot(class(ml_gbt_classifier(sc)))
  model <- tbl_mtcars |>
    ml_gbt_classifier(am ~ .)
  expect_snapshot(class(model))
  x <- tbl_mtcars |>
    ml_predict(model, dataset = _) |>
    use_test_pull()
  expect_snapshot(table(x))
})

test_that("GBT Regressor works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  expect_snapshot(class(ml_gbt_regressor(ml_pipeline(sc))))
  expect_snapshot(class(ml_gbt_regressor(sc)))
  model <- tbl_mtcars |>
    ml_gbt_regressor(mpg ~ ., seed = 100)
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_mtcars |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

test_that("Isotonic regression works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_isotonic_regression(ml_pipeline(sc))))
  expect_snapshot(class(ml_isotonic_regression(sc)))
  model <- tbl_iris |>
    ml_isotonic_regression(Petal_Length ~ Petal_Width)
  expect_snapshot(model)
  expect_snapshot(class(model))
  x <- tbl_iris |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

test_that("Generalized linear regression works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  expect_snapshot(class(ml_generalized_linear_regression(ml_pipeline(sc))))
  expect_snapshot(class(ml_generalized_linear_regression(sc)))
  model <- tbl_mtcars |>
    ml_generalized_linear_regression(mpg ~ .)
  expect_snapshot(class(model))
  x <- tbl_mtcars |>
    ml_predict(model, dataset = _) |>
    collect()
  expect_true("prediction" %in% names(x))
})

skip_on_ci()
skip_spark_min_version(3.5)

test_that("Pipeline fits and predicts", {
  reticulate::py_require("torch")
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_test_env()
    ),
    {
      use_test_install_ml()
      sc <- use_test_spark_connect()
      tbl_mtcars <- use_test_table_mtcars()

      new_pipeline <- sc |>
        ml_pipeline() |>
        ft_max_abs_scaler(
          input_col = "features",
          output_col = "scaled_features"
        ) |>
        ml_logistic_regression(features_col = "scaled_features", max_iter = 10)

      prepd <- ml_prepare_dataset(tbl_mtcars, am ~ .)

      expect_silent(fitted <- ml_fit(new_pipeline, prepd))

      expect_snapshot(class(fitted))

      pipeline_folder <- path(tempdir(), random_table_name("ml_"))
      expect_silent(ml_save(new_pipeline, pipeline_folder))
      expect_s3_class(
        ml_connect_load(sc, pipeline_folder),
        "ml_connect_pipeline"
      )

      fitted_folder <- path(tempdir(), random_table_name("ml_"))
      expect_silent(ml_save(fitted, fitted_folder))

      skip("Skip until Issue #132 is resolved")
      loaded <- ml_connect_load(sc, fitted_folder)
      expect_s3_class(loaded, "ml_connect_pipeline_model")
      expect_snapshot(colnames(ml_transform(loaded, prepd)))
    }
  )
})

test_that("Write text works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  file_name <- tempfile()
  text_tbl <- tbl_mtcars |>
    mutate(x = as.character(mpg)) |>
    select(x)
  expect_silent(spark_write_text(text_tbl, file_name, overwrite = ))
  expect_s3_class(spark_read_text(sc, file_name), "tbl_pyspark")
})

test_that("Write table works", {
  tbl_mtcars <- use_test_table_mtcars()
  expect_silent(
    spark_write_table(tbl_mtcars, random_table_name("mtcars"))
  )
  if (dir_exists(test_path("spark-warehouse"))) {
    dir_delete(test_path("spark-warehouse"))
  }
})

test_that("CSV works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  file_name <- tempfile()
  expect_silent(spark_write_csv(tbl_mtcars, file_name))
  expect_snapshot(
    spark_read_csv(
      sc = sc,
      name = "csv_1",
      path = file_name,
      overwrite = TRUE,
      repartition = 2
    )
  )
  expect_snapshot(
    spark_read_csv(
      sc = sc,
      name = "csv_2",
      path = file_name,
      overwrite = TRUE,
      columns = paste0(names(mtcars), "t")
    )
  )
  expect_snapshot(
    spark_read_csv(
      sc = sc,
      name = "csv_3",
      path = file_name,
      overwrite = TRUE,
      memory = TRUE
    )
  )
})

test_that("Parquet works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  file_name <- tempfile()
  expect_silent(spark_write_parquet(tbl_mtcars, file_name))
  expect_snapshot(
    spark_read_parquet(sc, "csv_1", file_name, overwrite = TRUE)
  )
})

test_that("ORC works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  file_name <- tempfile()
  expect_silent(spark_write_orc(tbl_mtcars, file_name))
  expect_snapshot(
    spark_read_orc(sc, "csv_1", file_name, overwrite = TRUE)
  )
})

test_that("JSON works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()
  file_name <- tempfile()
  expect_silent(spark_write_json(tbl_mtcars, file_name))
  expect_snapshot(
    spark_read_json(sc, "csv_1", file_name, overwrite = TRUE)
  )
})

test_that("Other tests", {
  expect_equal(
    substr(gen_sdf_name("12-3-4"), 1, 5),
    "1234_"
  )
})

skip_on_ci()
skip_spark_min_version("4")

test_that("Functions work", {
  reticulate::py_require("torch")
  tbl_mtcars <- use_test_table_mtcars()
  model <- use_test_lr_model()
  expect_length(
    suppressMessages(capture.output(print(model))),
    24
  )

  expect_s3_class(spark_jobj(model), "spark_pyobj")

  expect_s3_class(ml_predict(model, tbl_mtcars), "tbl_pyspark")

  expect_snapshot(
    colnames(transform_impl(model, tbl_mtcars, prep = TRUE))
  )

  expect_snapshot(
    colnames(transform_impl(model, tbl_mtcars, prep = TRUE, remove = TRUE))
  )
})

skip_spark_min_version(4)

test_that("Binary evaluation works", {
  sc <- use_test_spark_connect()
  tbl_iris <- use_test_table_iris()
  expect_snapshot(class(ml_binary_classification_evaluator(sc)))
  tbl_mtcars <- use_test_table_mtcars()
  model <- tbl_mtcars |>
    ml_logistic_regression(am ~ .)
  preds <- ml_predict(model, tbl_mtcars)
  expect_snapshot(
    ml_binary_classification_evaluator(preds, label_col = "am")
  )
})

test_that("Regression evaluation works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ml_regression_evaluator(sc)))
  tbl_mtcars <- use_test_table_mtcars()
  model <- tbl_mtcars |>
    ml_linear_regression(wt ~ .)
  preds <- ml_predict(model, tbl_mtcars)
  expect_snapshot(
    ml_regression_evaluator(preds, label_col = "wt", metric_name = "r2")
  )
})

test_that("Multiclass evaluation works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ml_multiclass_classification_evaluator(sc)))
  tbl_iris <- use_test_table_iris() |>
    ft_string_indexer("Species", "species_idx")
  model <- tbl_iris |>
    ml_random_forest_classifier(
      species_idx ~ Sepal_Length + Sepal_Width + Petal_Length + Petal_Width
    )
  preds <- ml_predict(model, tbl_iris)
  expect_snapshot(
    ml_multiclass_classification_evaluator(preds, label_col = "species_idx")
  )
})

test_that("Clustering evaluation works", {
  sc <- use_test_spark_connect()
  expect_snapshot(class(ml_clustering_evaluator(sc)))
  tbl_iris <- use_test_table_iris() |>
    ft_string_indexer("Species", "species_idx")
  model <- tbl_iris |>
    ml_kmeans(
      species_idx ~ Sepal_Length + Sepal_Width + Petal_Length + Petal_Width,
      seed = 1
    )
  preds <- ml_predict(model, tbl_iris)
  expect_snapshot(
    preds |>
      mutate(prediction = as.numeric(prediction)) |>
      compute() |>
      ml_regression_evaluator(label_col = "species_idx")
  )
})

skip_spark_min_version(4)

test_that("Cross validator works", {
  sc <- use_test_spark_connect()
  tbl_mtcars <- use_test_table_mtcars()

  pipeline <- sc |>
    ml_pipeline() |>
    ft_binarizer("mpg", "mpg2", 20) |>
    ft_r_formula(mpg2 ~ .) |>
    ml_logistic_regression()

  grid <- list(
    logistic_regression = list(
      reg_param = seq(0, 1, length = 3)
    )
  )

  cv <- ml_cross_validator(
    x = sc,
    estimator = pipeline,
    estimator_param_maps = grid,
    evaluator = ml_binary_classification_evaluator(sc),
    seed = 100
  )

  expect_snapshot(class(cv))

  tuning_model <- ml_fit(cv, tbl_mtcars)
  expect_snapshot(class(tuning_model))

  metrics <- ml_validation_metrics(tuning_model)

  expect_snapshot(use_test_pull(metrics))

  grid2 <- list(
    linear_regression = list(
      reg_param = seq(0, 1, length = 3)
    )
  )

  expect_error(
    cv <- ml_cross_validator(
      x = sc,
      estimator = pipeline,
      estimator_param_maps = grid2,
      evaluator = ml_binary_classification_evaluator(sc),
      seed = 100
    )
  )
})

test_that("Pivot longer", {
  sc <- use_test_spark_connect()
  local_pivot <- tibble::tribble(
    ~id , ~z_1 , ~y_1 , ~x_1 , ~z_2 , ~y_2 , ~x_2 ,
    "A" ,    1 ,    2 ,    3 ,    4 ,    5 ,    6 ,
    "B" ,    7 ,    8 ,    9 ,   10 ,   11 ,   12 ,
  )
  tbl_pivot <- copy_to(sc, local_pivot)

  expect_snapshot(
    tbl_pivot %>%
      tidyr::pivot_longer(
        -id,
        names_to = c(".value", "n"),
        names_sep = "_"
      ) %>%
      collect()
  )
})

trivial_sdf <- testthat_tbl(
  "testthat_tidyr_pivot_longer_trivial_sdf",
  data = tibble::tibble(x_y = 1)
)

test_that("can pivot all cols to long", {
  expect_same_remote_result(
    tibble::tibble(x = 1:2, y = 3:4),
    function(.x) .x |> tidyr::pivot_longer(x:y)
  )
})

test_that("values interleaved correctly", {
  expect_same_remote_result(
    tibble::tibble(x = c(1, 2), y = c(10, 20), z = c(100, 200)),
    function(.x) .x |> tidyr::pivot_longer(1:3)
  )
})

test_that("can drop missing values", {
  expect_same_remote_result(
    tibble::tibble(x = c(1, NA), y = c(NA, 2)),
    function(.x) .x |> tidyr::pivot_longer(x:y, values_drop_na = TRUE)
  )
})

test_that("preserves original keys", {
  expect_same_remote_result(
    tibble::tibble(x = 1:2, y = 2L, z = 1:2),
    function(.x) .x |> tidyr::pivot_longer(y:z)
  )
})

test_that("can handle missing combinations", {
  expect_same_remote_result(
    tibble::tribble(
      ~id , ~x_1 , ~x_2 , ~y_2 ,
      "A" ,    1 ,    2 , "a"  ,
      "B" ,    3 ,    4 , "b"  ,
    ),
    function(.x) {
      .x |>
        tidyr::pivot_longer(
          -id,
          names_to = c(".value", "n"),
          names_sep = "_"
        )
    }
  )
})

test_that("original col order is preserved", {
  expect_same_remote_result(
    tibble::tribble(
      ~id , ~z_1 , ~y_1 , ~x_1 , ~z_2 , ~y_2 , ~x_2 ,
      "A" ,    1 ,    2 ,    3 ,    4 ,    5 ,    6 ,
      "B" ,    7 ,    8 ,    9 ,   10 ,   11 ,   12 ,
    ),
    function(.x) {
      .x |>
        tidyr::pivot_longer(
          -id,
          names_to = c(".value", "n"),
          names_sep = "_"
        )
    }
  )
})

test_that("can pivot duplicated names to .value", {
  expect_same_remote_result(
    tibble::tibble(x = 1, a_1 = 1, a_2 = 2, b_1 = 3, b_2 = 4),
    function(.x) {
      .x |> tidyr::pivot_longer(-x, names_to = c(".value", NA), names_sep = "_")
    }
  )

  expect_same_remote_result(
    tibble::tibble(x = 1, a_1 = 1, a_2 = 2, b_1 = 3, b_2 = 4),
    function(.x) {
      .x |>
        tidyr::pivot_longer(
          -x,
          names_to = c(".value", NA),
          names_pattern = "(.)_(.)"
        )
    }
  )

  expect_same_remote_result(
    tibble::tibble(x = 1, a_1 = 1, a_2 = 2, b_1 = 3, b_2 = 4),
    function(.x) {
      .x |>
        tidyr::pivot_longer(-x, names_to = ".value", names_pattern = "(.)_.")
    }
  )
})

test_that(".value can be at any position in `names_to`", {
  sc <- use_test_spark_connect()
  samp_sdf <- copy_to(
    sc,
    tibble::tibble(
      i = 1:4,
      y_t1 = rnorm(4),
      y_t2 = rnorm(4),
      z_t1 = rep(3, 4),
      z_t2 = rep(-2, 4)
    )
  )
  samp_sdf2 <- dplyr::rename(
    samp_sdf,
    t1_y = y_t1,
    t2_y = y_t2,
    t1_z = z_t1,
    t2_z = z_t2
  )

  pv <- lapply(
    list(
      tidyr::pivot_longer(
        samp_sdf,
        -i,
        names_to = c(".value", "time"),
        names_sep = "_"
      ),
      tidyr::pivot_longer(
        samp_sdf2,
        -i,
        names_to = c("time", ".value"),
        names_sep = "_"
      )
    ),
    collect
  )

  expect_identical(pv[[1]], pv[[2]])
})

test_that("reporting data type mismatch", {
  sc <- use_test_spark_connect()
  sdf <- copy_to(sc, tibble::tibble(abc = 1, xyz = "b"))
  err <- capture_error(tidyr::pivot_longer(sdf, tidyr::everything()))

  expect_true(grepl("data type mismatch", err$message, fixed = TRUE))
})

test_that("grouping is preserved", {
  expect_same_remote_result(
    tibble::tibble(g = 1, x1 = 1, x2 = 2),
    function(.x) {
      .x |>
        dplyr::group_by(g) %>%
        tidyr::pivot_longer(x1:x2, names_to = "x", values_to = "v")
    }
  )
})

#' @export
pivot_longer.tbl_pyspark <- function(
  data,
  cols,
  ...,
  cols_vary = "fastest",
  names_to = "name",
  names_prefix = NULL,
  names_sep = NULL,
  names_pattern = NULL,
  names_ptypes = NULL,
  names_transform = NULL,
  names_repair = NULL,
  values_to = "value",
  values_drop_na = FALSE,
  values_ptypes = NULL,
  values_transform = NULL
) {
  check_arg_supported(names_ptypes)
  check_arg_supported(names_transform)
  check_arg_supported(names_repair)
  check_arg_supported(values_ptypes)
  check_arg_supported(values_transform)

  sc <- spark_connection(data)

  spark_df <- tbl_pyspark_sdf(data)

  # This part is so that we can run tidyeval on the resulting
  # Spark Data Frame. We're converting the resulting SDF back
  # into SQL by making it a temp view, and then accesing it via
  # SQL, so we can use `dplyr` commands. We don't do this for
  # the original `data` variable, becuase it `pivot_longer()` may
  # be called after one or several `dplyr` commands. So we have to
  # operate the piped commands first.
  temp_name <- tbl_temp_name()
  te_df <- tbl_pyspark_temp(spark_df, data, temp_name)
  col_names <- te_df |>
    select(!!enquo(cols)) |>
    colnames()

  if (length(col_names) == 0) {
    abort("'cols' must select at least one column.")
  }

  # Determining what columns are NOT selected
  col_all <- spark_df$columns
  col_dif <- col_all
  for (col in col_names) {
    col_dif <- col_dif[col_dif != col]
  }

  # Becuase of how PySpark `unpivot` works, if there are no unselected columns
  # I chose to pass the first one, and then remove it after performing the
  # un-pivot operation. Not ideal, but it works
  remove_first <- FALSE
  if (length(col_dif) == 0) {
    col_dif <- col_all[1]
    remove_first <- TRUE
  }

  output_names <- names_to
  if (!is.null(names_sep)) {
    output_names <- .str_separate(col_names, names_to, sep = names_sep)
  } else {
    if (!is.null(names_pattern)) {
      output_names <- .str_extract(col_names, names_to, regex = names_pattern)
    }
  }

  val_vals <- output_names[".value"][[1]]

  u_val <- unique(val_vals)

  if (length(u_val) > 1) {
    if (length(names_to) > 2) {
      abort("Only two level is supported")
    }

    if (names_to[[1]] == ".value") {
      val_no <- 1
      nm_no <- 2
    } else {
      val_no <- 2
      nm_no <- 1
    }

    if (length(names_to) > 1) {
      set_names_to <- names_to[[nm_no]]
    } else {
      set_names_to <- "temp222"
    }

    all_pv <- NULL
    for (i in seq_along(u_val)) {
      nm_rm <- list()
      c_val <- u_val[[i]]
      nm_rm[[val_no]] <- c_val
      nm_rm[[nm_no]] <- names_sep
      nm_rm <- paste0(nm_rm, collapse = "")
      cur_cols <- col_names[which(val_vals == c_val)]
      sel_df <- spark_df$select(c(col_dif, cur_cols))
      for (j in seq_along(cur_cols)) {
        ren_cols <- sub(nm_rm, "", cur_cols)
        sel_df <- sel_df$withColumnRenamed(
          existing = cur_cols[[j]],
          new = sub(nm_rm, "", ren_cols[[j]])
        )
      }

      check_same_types(x = sel_df, col_names = ren_cols)

      all_pv[[i]] <- un_pivot(
        x = sel_df,
        ids = col_dif,
        values = ren_cols,
        names_to = set_names_to,
        values_to = c_val,
        remove_first = remove_first,
        values_drop_na = values_drop_na,
        repair = names_repair
      )
    }

    out <- NULL
    for (i in seq_along(all_pv)) {
      no_i <- length(all_pv) - i + 1
      if (no_i > 1) {
        if (is.null(out)) {
          out <- all_pv[[no_i]]
        }
        next_df <- all_pv[[(no_i - 1)]]
        out <- out$join(
          other = next_df,
          on = as.list(c(col_dif, set_names_to)),
          how = "full"
        )
      }
    }

    output_cols <- colnames(output_names)
    output_cols <- output_cols[output_cols != ".value"]

    out <- out$select(as.list(c(col_dif, output_cols, u_val)))
  } else {
    check_same_types(x = spark_df, col_names = col_names)

    out <- un_pivot(
      x = spark_df,
      ids = col_dif,
      values = col_names,
      names_to = names_to,
      values_to = values_to,
      remove_first = remove_first,
      values_drop_na = values_drop_na,
      repair = names_repair
    )
  }

  # Cleaning up by removing the temp view with the operations
  # that occurred before pivoting
  dbRemoveTable(sc, temp_name)

  # Creating temp view with the pivoting results
  out <- tbl_pyspark_temp(out, sc)
  gr_out <- group_vars(data)
  if (length(gr_out) > 0) {
    out <- group_by(out, !!!rlang::syms(gr_out))
  }

  out
}

un_pivot <- function(
  x,
  ids,
  values,
  names_to,
  values_to,
  remove_first,
  values_drop_na,
  repair
) {
  out <- x$unpivot(
    ids = as.list(ids),
    values = as.list(values),
    variableColumnName = names_to,
    valueColumnName = values_to
  )

  if (remove_first) {
    out <- out$select(list(names_to, values_to))
  }

  if (values_drop_na) {
    out <- out$dropna(subset = values_to)
  }

  col_names <- out$columns

  if (length(unique(col_names)) != length(col_names)) {
    new_names <- vec_as_names(col_names, repair = repair)
  }

  out
}

check_same_types <- function(x, col_names) {
  dtypes <- x$select(col_names)$dtypes
  char_types <- map_chr(dtypes, \(.x) .x[[2]])
  char_match <- length(unique(char_types)) == 1

  if (!char_match) {
    type_names <- map_chr(dtypes, \(.x) paste0(.x[[1]], " <", .x[[2]], ">"))
    abort(glue(
      "There is a data type mismatch: {paste0(type_names, collapse = ' ')}"
    ))
  }
}

# -------------------------- Linear Regression  --------------------------------
ml_linear_regression_impl <- function(
  x,
  formula = NULL,
  fit_intercept = TRUE,
  elastic_net_param = 0,
  reg_param = 0,
  max_iter = 100,
  weight_col = NULL,
  loss = "squaredError",
  solver = "auto",
  standardization = TRUE,
  tol = 1e-6,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "LinearRegression",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "linear_regression"
  )
}
#' @export
ml_linear_regression.pyspark_connection <- ml_linear_regression_impl
#' @export
ml_linear_regression.ml_connect_pipeline <- ml_linear_regression_impl
#' @export
ml_linear_regression.tbl_pyspark <- ml_linear_regression_impl


# ------------------------- Logistic Regression  -------------------------------
ml_logistic_regression_impl <- function(
  x,
  formula = NULL,
  fit_intercept = NULL,
  elastic_net_param = NULL,
  reg_param = NULL,
  max_iter = 100,
  threshold = NULL,
  thresholds = NULL,
  tol = 1e-06,
  weight_col = NULL,
  aggregation_depth = NULL,
  lower_bounds_on_coefficients = NULL,
  lower_bounds_on_intercepts = NULL,
  upper_bounds_on_coefficients = NULL,
  upper_bounds_on_intercepts = NULL,
  features_col = "features",
  label_col = "label",
  family = NULL,
  prediction_col = "prediction",
  probability_col = "probability",
  raw_prediction_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "LogisticRegression",
    has_fit = TRUE,
    ml_type = "classification",
    ml_fn = "logistic_regression"
  )
}
#' @export
ml_logistic_regression.pyspark_connection <- ml_logistic_regression_impl
#' @export
ml_logistic_regression.ml_connect_pipeline <- ml_logistic_regression_impl
#' @export
ml_logistic_regression.tbl_pyspark <- ml_logistic_regression_impl

# ----------------------- Random Forest Classifier  ----------------------------
ml_random_forest_classifier_impl <- function(
  x,
  formula = NULL,
  num_trees = 10,
  subsampling_rate = 1,
  max_depth = 5,
  min_instances_per_node = 1,
  feature_subset_strategy = "auto",
  impurity = "gini",
  min_info_gain = 0,
  max_bins = 32,
  seed = NULL,
  thresholds = NULL,
  checkpoint_interval = 10,
  cache_node_ids = FALSE,
  max_memory_in_mb = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  probability_col = "probability",
  raw_prediction_col = "rawPrediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "RandomForestClassifier",
    has_fit = TRUE,
    ml_type = "classification",
    ml_fn = "random_forest_classifier"
  )
}
#' @export
ml_random_forest_classifier.pyspark_connection <- ml_random_forest_classifier_impl
#' @export
ml_random_forest_classifier.ml_connect_pipeline <- ml_random_forest_classifier_impl
#' @export
ml_random_forest_classifier.tbl_pyspark <- ml_random_forest_classifier_impl

# ----------------------- Random Forest Regressor  -----------------------------
ml_random_forest_regressor_impl <- function(
  x,
  formula = NULL,
  num_trees = 20,
  subsampling_rate = 1,
  max_depth = 5,
  min_instances_per_node = 1,
  feature_subset_strategy = "auto",
  impurity = "variance",
  min_info_gain = 0,
  max_bins = 32,
  seed = NULL,
  checkpoint_interval = 10,
  cache_node_ids = FALSE,
  max_memory_in_mb = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "RandomForestRegressor",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "random_forest_regressor"
  )
}
#' @export
ml_random_forest_regressor.pyspark_connection <- ml_random_forest_regressor_impl
#' @export
ml_random_forest_regressor.ml_connect_pipeline <- ml_random_forest_regressor_impl
#' @export
ml_random_forest_regressor.tbl_pyspark <- ml_random_forest_regressor_impl

# ----------------------- Decision Tree Classifier  ----------------------------
ml_decision_tree_classifier_impl <- function(
  x,
  formula = NULL,
  max_depth = 5,
  max_bins = 32,
  min_instances_per_node = 1,
  min_info_gain = 0,
  impurity = "gini",
  seed = NULL,
  thresholds = NULL,
  cache_node_ids = FALSE,
  checkpoint_interval = 10,
  max_memory_in_mb = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  probability_col = "probability",
  raw_prediction_col = "rawPrediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "DecisionTreeClassifier",
    has_fit = TRUE,
    ml_type = "classification",
    ml_fn = "decision_tree_classifier"
  )
}
#' @export
ml_decision_tree_classifier.pyspark_connection <- ml_decision_tree_classifier_impl
#' @export
ml_decision_tree_classifier.ml_connect_pipeline <- ml_decision_tree_classifier_impl
#' @export
ml_decision_tree_classifier.tbl_pyspark <- ml_decision_tree_classifier_impl

# ----------------------- Decision Tree Regressor  ----------------------------
ml_decision_tree_regressor_impl <- function(
  x,
  formula = NULL,
  max_depth = 5,
  max_bins = 32,
  min_instances_per_node = 1,
  min_info_gain = 0,
  impurity = "variance",
  seed = NULL,
  cache_node_ids = FALSE,
  checkpoint_interval = 10,
  max_memory_in_mb = NULL,
  variance_col = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "DecisionTreeRegressor",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "decision_tree_regressor"
  )
}
#' @export
ml_decision_tree_regressor.pyspark_connection <- ml_decision_tree_regressor_impl
#' @export
ml_decision_tree_regressor.ml_connect_pipeline <- ml_decision_tree_regressor_impl
#' @export
ml_decision_tree_regressor.tbl_pyspark <- ml_decision_tree_regressor_impl

# --------------------------------- Kmeans  ------------------------------------
ml_kmeans_impl <- function(
  x,
  formula = NULL,
  k = 2,
  max_iter = 20,
  tol = 1e-4,
  init_steps = 2,
  init_mode = "k-means||",
  seed = NULL,
  features_col = "features",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "KMeans",
    has_fit = TRUE,
    ml_type = "clustering",
    ml_fn = "kmeans"
  )
}
#' @export
ml_kmeans.pyspark_connection <- ml_kmeans_impl
#' @export
ml_kmeans.ml_connect_pipeline <- ml_kmeans_impl
#' @export
ml_kmeans.tbl_pyspark <- ml_kmeans_impl

# --------------------------- Bisecting Kmeans  --------------------------------
ml_bisecting_kmeans_impl <- function(
  x,
  formula = NULL,
  k = 4,
  max_iter = 20,
  seed = NULL,
  min_divisible_cluster_size = 1,
  features_col = "features",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "BisectingKMeans",
    has_fit = TRUE,
    ml_type = "clustering",
    ml_fn = "bisecting_kmeans"
  )
}
#' @export
ml_bisecting_kmeans.pyspark_connection <- ml_bisecting_kmeans_impl
#' @export
ml_bisecting_kmeans.ml_connect_pipeline <- ml_bisecting_kmeans_impl
#' @export
ml_bisecting_kmeans.tbl_pyspark <- ml_bisecting_kmeans_impl

# --------------------------- AFT Survival  ------------------------------------
ml_aft_survival_regression_impl <- function(
  x,
  formula = NULL,
  censor_col = "censor",
  quantile_probabilities = c(
    0.01,
    0.05,
    0.1,
    0.25,
    0.5,
    0.75,
    0.9,
    0.95,
    0.99
  ),
  fit_intercept = TRUE,
  max_iter = 100L,
  tol = 1e-06,
  aggregation_depth = 2,
  quantiles_col = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "AFTSurvivalRegression",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "aft_survival_regressor"
  )
}
#' @export
ml_aft_survival_regression.pyspark_connection <- ml_aft_survival_regression_impl
#' @export
ml_aft_survival_regression.ml_connect_pipeline <- ml_aft_survival_regression_impl
#' @export
ml_aft_survival_regression.tbl_pyspark <- ml_aft_survival_regression_impl

# ------------------------------ GBT Classifier  -------------------------------
ml_gbt_classifier_impl <- function(
  x,
  formula = NULL,
  max_iter = 20,
  max_depth = 5,
  step_size = 0.1,
  subsampling_rate = 1,
  feature_subset_strategy = "auto",
  min_instances_per_node = 1L,
  max_bins = 32,
  min_info_gain = 0,
  loss_type = "logistic",
  seed = NULL,
  thresholds = NULL,
  checkpoint_interval = 10,
  cache_node_ids = FALSE,
  max_memory_in_mb = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  probability_col = NULL,
  raw_prediction_col = NULL,
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "GBTClassifier",
    has_fit = TRUE,
    ml_type = "classification",
    ml_fn = "gbt_classifier"
  )
}
#' @export
ml_gbt_classifier.pyspark_connection <- ml_gbt_classifier_impl
#' @export
ml_gbt_classifier.ml_connect_pipeline <- ml_gbt_classifier_impl
#' @export
ml_gbt_classifier.tbl_pyspark <- ml_gbt_classifier_impl

# ------------------------------- GBT Regressor  -------------------------------
ml_gbt_regressor_impl <- function(
  x,
  formula = NULL,
  max_iter = 20,
  max_depth = 5,
  step_size = 0.1,
  subsampling_rate = 1,
  feature_subset_strategy = "auto",
  min_instances_per_node = 1,
  max_bins = 32,
  min_info_gain = 0,
  loss_type = "squared",
  seed = NULL,
  checkpoint_interval = 10,
  cache_node_ids = FALSE,
  max_memory_in_mb = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "GBTRegressor",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "gbt_regressor"
  )
}
#' @export
ml_gbt_regressor.pyspark_connection <- ml_gbt_regressor_impl
#' @export
ml_gbt_regressor.ml_connect_pipeline <- ml_gbt_regressor_impl
#' @export
ml_gbt_regressor.tbl_pyspark <- ml_gbt_regressor_impl

# --------------------------- Isotonic Regression  -----------------------------
ml_isotonic_regression_impl <- function(
  x,
  formula = NULL,
  feature_index = 0,
  isotonic = TRUE,
  weight_col = NULL,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "IsotonicRegression",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "isotonic_regressor"
  )
}
#' @export
ml_isotonic_regression.pyspark_connection <- ml_isotonic_regression_impl
#' @export
ml_isotonic_regression.ml_connect_pipeline <- ml_isotonic_regression_impl
#' @export
ml_isotonic_regression.tbl_pyspark <- ml_isotonic_regression_impl

# ----------------------- Generalized Linear Regression  -----------------------
ml_generalized_linear_regression_impl <- function(
  x,
  formula = NULL,
  family = "gaussian",
  link = NULL,
  fit_intercept = TRUE,
  offset_col = NULL,
  link_power = NULL,
  link_prediction_col = NULL,
  reg_param = 0,
  max_iter = 25,
  weight_col = NULL,
  solver = "irls",
  tol = 1e-6,
  variance_power = 0,
  features_col = "features",
  label_col = "label",
  prediction_col = "prediction",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "GeneralizedLinearRegression",
    has_fit = TRUE,
    ml_type = "regression",
    ml_fn = "generalized_linear_regressor"
  )
}
#' @export
ml_generalized_linear_regression.pyspark_connection <- ml_generalized_linear_regression_impl
#' @export
ml_generalized_linear_regression.ml_connect_pipeline <- ml_generalized_linear_regression_impl
#' @export
ml_generalized_linear_regression.tbl_pyspark <- ml_generalized_linear_regression_impl

#' Writes the 'requirements.txt' file, containing the needed Python libraries
#' @description This is a helper function that it is meant to be used for deployments
#' of the document or application. By default, `deploy_databricks()` will run this
#' function the first time you use that function to deploy content to Posit Connect.
#' @param envname The name of, or path to, a Python virtual environment.
#' @param destfile Target path for the requirements file. Defaults to 'requirements.txt'.
#' @param overwrite Replace the contents of the file if it already exists?
#' @param ... Additional arguments passed to `reticulate::py_list_packages()`
#' @returns No value is returned to R. The output is a text file with the
#' list of Python libraries.
#' @export
requirements_write <- function(
  envname = NULL,
  destfile = "requirements.txt",
  overwrite = FALSE,
  ...
) {
  cli_div(theme = cli_colors())

  if (file_exists(destfile)) {
    if (overwrite) {
      file_delete(destfile)
    } else {
      cli_abort(c(
        "{.header File }{.emph '{path_file(destfile)}'} {.header already exists}",
        " " = "{.header Use} {.code overwrite = TRUE} {.header if you wish to replace}"
      ))
    }
  }

  pkgs <- py_list_packages(
    envname = envname,
    ...
  )

  writeLines(
    c(
      "# Automatically created by `sparklyr`",
      "",
      pkgs$requirement
    ),
    con = destfile
  )

  txt_pkgs_lbl <- c(
    "pyspark",
    "databricks.connect",
    "databricks-connect"
  ) |>
    map_chr(\(.x) {
      x <- pkgs$version[pkgs$package == .x]
      if (length(x) == 0) {
        x <- ""
      } else {
        x <- glue("'{.x}' '{x}'")
      }
      x
    }) |>
    discard(\(.x) .x == "") |>
    reduce(paste, collapse = ",", .init = "")

  if (txt_pkgs_lbl != "") {
    line1 <- "{.header Python library requirements for}{.emph {txt_pkgs_lbl}}"
  } else {
    line1 <- "{.header Python library requirements}"
  }

  cli_inform(c(
    "i" = line1,
    " " = "{.header File}: {path_expand(destfile)}"
  ))
  cli_end()
  invisible()
}

#' @export
spark_apply.tbl_pyspark <- function(
  x,
  f,
  columns = NULL,
  memory = TRUE,
  group_by = NULL,
  packages = NULL,
  context = NULL,
  name = NULL,
  barrier = NULL,
  fetch_result_as_sdf = TRUE,
  partition_index_param = "",
  arrow_max_records_per_batch = NULL,
  auto_deps = FALSE,
  ...
) {
  rpy2_installed()
  cli_div(theme = cli_colors())
  if (!is.null(packages)) {
    cli_abort("`packages` is not yet supported for this backend")
  }
  if (!is.null(context)) {
    if (!is.null(group_by)) {
      if (inherits(context, "list")) {
        cli_abort("R lists are not supported for `context`")
      }
    } else {
      cli_abort("`context` is only supported for calls using `group_by`")
    }
  }
  if (auto_deps) {
    cli_abort("`auto_deps` is not supported for this backend")
  }
  if (partition_index_param != "") {
    cli_abort("`partition_index_param` is not supported for this backend")
  }
  if (!is.null(arrow_max_records_per_batch)) {
    sc <- python_sdf(x)$sparkSession
    conf_name <- "spark.sql.execution.arrow.maxRecordsPerBatch"
    conf_curr <- sc$conf$get(conf_name)
    conf_req <- as.character(arrow_max_records_per_batch)
    if (conf_curr != conf_req) {
      cli_div(theme = cli_colors())
      cli_inform(
        "{.header Changing {.emph {conf_name}} to: {prettyNum(conf_req, big.mark = ',')}}"
      )
      cli_end()
      sc$conf$set(conf_name, conf_req)
    }
  }
  cli_end()
  sa_in_pandas(
    x = x,
    .f = f,
    .schema = columns,
    .group_by = group_by,
    .as_sdf = fetch_result_as_sdf,
    .name = name,
    .barrier = barrier,
    .context = context,
    ... = ...
  )
}

sa_in_pandas <- function(
  x,
  .f,
  ...,
  .schema = NULL,
  .schema_arg = "columns",
  .group_by = NULL,
  .as_sdf = TRUE,
  .name = NULL,
  .context = NULL,
  .barrier = NULL
) {
  schema_msg <- FALSE
  if (is.null(.schema)) {
    r_fn <- .f |>
      sa_function_to_string(
        .r_only = TRUE,
        .group_by = .group_by,
        .colnames = NULL,
        .context = .context,
        ... = ...
      ) |>
      rlang::parse_expr() |>
      eval()
    r_df <- x |>
      head(10) |>
      collect()
    if (is.null(.context)) {
      r_exec <- r_fn(r_df)
    } else {
      r_exec <- r_fn(r_df, .context)
    }

    col_names <- colnames(r_exec)
    col_names <- gsub("\\.", "_", col_names)
    colnames(r_exec) <- col_names
    .schema <- r_exec |>
      imap(\(.x, .y) {
        x_class <- class(.x)
        if ("POSIXt" %in% x_class) {
          x_class <- "timestamp"
        }
        if (x_class == "character") {
          x_class <- "string"
        }
        if (x_class == "numeric") {
          x_class <- "double"
        }
        if (x_class == "integer") {
          x_class <- "long"
        }
        paste0(.y, " ", x_class)
      }) |>
      paste0(collapse = ", ")
    schema_msg <- TRUE
  } else {
    fields <- unlist(strsplit(.schema, ","))
    col_names <- map_chr(fields, \(.x) unlist(strsplit(trimws(.x), " "))[[1]])
    col_names <- gsub("\\.", "_", col_names)
  }
  .f |>
    sa_function_to_string(
      .group_by = .group_by,
      .colnames = col_names,
      .context = .context,
      ... = ...
    ) |>
    py_run_string()
  main <- reticulate::import_main()
  if (inherits(x, "tbl_spark")) {
    df <- python_sdf(x)
  } else {
    df <- x
  }
  if (is.null(df)) {
    df <- x |>
      compute() |>
      python_sdf()
  }
  if (!is.null(.group_by)) {
    # TODO: Add support for multiple grouping columns
    renamed_gp <- paste0("_", .group_by)
    w_gp <- df$withColumn(colName = renamed_gp, col = df[.group_by])
    tbl_gp <- w_gp$groupby(renamed_gp)
    if (is.null(.context)) {
      p_df <- tbl_gp$applyInPandas(
        main$r_apply,
        schema = .schema
      )
    } else {
      p_df <- tbl_gp$applyInPandas(
        main$r_apply(.context),
        schema = .schema
      )
    }
  } else {
    p_df <- df$mapInPandas(
      main$r_apply,
      schema = .schema,
      barrier = .barrier %||% FALSE
    )
  }
  if (.as_sdf) {
    ret <- tbl_pyspark_temp(
      x = p_df,
      conn = spark_connection(x),
      tmp_name = .name
    )
  } else {
    ret <- to_pandas_cleaned(p_df)
  }
  if (schema_msg) {
    schema_arg <- .schema_arg
    schema <- .schema
    cli_div(theme = cli_colors())
    cli_inform(c(
      "{.header To increase performance, use the following schema:}",
      "{.emph {schema_arg} = \"{schema}\" }"
    ))
    cli_end()
  }
  ret
}

sa_function_to_string <- function(
  .f,
  .group_by = NULL,
  .r_only = FALSE,
  .colnames = NULL,
  .context = NULL,
  ...
) {
  path_scripts <- pkg_path("udf")
  udf_fn <- ifelse(is.null(.group_by), "map", "apply")
  if (!is.null(.context)) {
    udf_fn <- glue("{udf_fn}-context")
  }
  fn_r <- paste0(
    readLines(path(path_scripts, glue("udf-{udf_fn}.R"))),
    collapse = ""
  )
  fn_python <- paste0(
    readLines(path(path_scripts, glue("udf-{udf_fn}.py"))),
    collapse = "\n"
  )
  if (!is.null(.group_by)) {
    fn_r <- gsub(
      "gp_field <- 'am'",
      paste0("gp_field <- '", .group_by, "'"),
      fn_r
    )
  }
  if (is.null(.colnames)) {
    .colnames <- "NULL"
  } else {
    .colnames <- paste0("'", .colnames, "'", collapse = ", ")
  }
  fn_r <- gsub(
    "col_names <- c\\('am', 'x'\\)",
    paste0("col_names <- c(", .colnames, ")"),
    fn_r
  )
  fn <- as_mapper(.f = .f, ... = ...)
  if (is.character(.f)) {
    fn_str <- .f
  } else {
    fn_str <- paste0(deparse(fn), collapse = "\n")
  }
  if (inherits(fn, "rlang_lambda_function")) {
    fn_str <- paste0(
      "function(...) {x <- (",
      fn_str,
      "); x(...)}"
    )
  }
  fn_str <- gsub("\"", "'", fn_str)
  fn_rep <- "function\\(\\.\\.\\.\\) 1"
  fn_r_new <- gsub(fn_rep, fn_str, fn_r)
  if (.r_only) {
    ret <- fn_r_new
  } else {
    ret <- gsub(fn_rep, fn_r_new, fn_python)
  }
  ret
}

rpy2_installed <- function(envname = NULL) {
  py_check_installed(
    envname = envname,
    libraries = "rpy2",
    msg = "Required 'rpy2' Python library is missing"
  )
}

ml_formula <- function(f, data) {
  col_data <- colnames(data)

  temp_tbl <- rep(1, times = length(col_data)) |>
    set_names(col_data) |>
    as.list() |>
    as.data.frame()

  f_terms <- terms(f, data = temp_tbl)

  f_factors <- attr(f_terms, "factors")

  feat_names <- colnames(f_factors)
  row_f <- rownames(f_factors)

  for (i in feat_names) {
    in_data <- i %in% col_data
    if (!in_data) {
      cli_abort(
        c(
          "Formula resulted in an invalid parameter set\n",
          "- Only '+' is supported."
        ),
        call = NULL
      )
    }
  }

  label <- NULL
  for (i in row_f) {
    if (!(i %in% feat_names)) {
      label <- c(label, i)
    }
  }
  list(
    label = label,
    features = feat_names
  )
}

snake_to_camel <- function(x) {
  s <- unlist(strsplit(x, "_"))
  x <- paste(
    toupper(substring(s, 1, 1)),
    substring(s, 2),
    sep = "",
    collapse = ""
  )
  paste(
    tolower(substring(x, 1, 1)),
    substring(x, 2),
    sep = "",
    collapse = ""
  )
}

ml_connect_not_supported <- function(args, not_supported = c()) {
  x <- map(
    not_supported,
    \(.x) {
      if (.x %in% names(args)) {
        arg <- args[names(args) == .x]
        if (!is.null(arg[[1]])) {
          ret <- paste0("{.emph - `", names(arg), "`}")
        } else {
          ret <- ""
        }
      }
    }
  ) |>
    discard(is.null) |>
    as.character()

  x <- x[x != ""]

  if (length(x) > 0) {
    cli_abort(c(
      "The following argument(s) are not supported by Spark Connect:",
      x,
      "Set it/them to `NULL` and try again."
    ))
  }
}

ml_execute <- function(args, python_library, fn, sc) {
  py_lib <- import(python_library)

  # Removes any variable with a "." prefix
  args <- args[substr(names(args), 1, 1) != "."]

  args$x <- NULL

  args <- discard(args, is.null)

  new_names <- args |>
    names() |>
    map_chr(snake_to_camel)

  new_args <- set_names(args, new_names)

  invisible(
    jobj <- do.call(
      what = py_lib[fn],
      args = new_args
    )
  )

  as_spark_pyobj(jobj, sc)
}

get_params <- function(x) {
  py_model <- python_obj_get(x)
  m_params <- py_model$params
  m_map <- py_model$extractParamMap()
  m_map_names <- m_map |>
    names() |>
    strsplit("__") |>
    map(\(.x) .x[[2]]) |>
    as.character() |>
    sort()

  m_param_names <- m_params |>
    map(\(.x) .x$name) |>
    as.character()

  m_map_names |>
    map(\(.x) {
      c_param_name <- m_params[which(.x == m_param_names)]
      c_map_name <- m_map[which(.x == m_map_names)]
      list(
        name = .x,
        value = c_map_name[[1]],
        description = c_param_name[[1]]$doc
      )
    })
}

ml_installed <- function(envname = NULL) {
  py_check_installed(
    envname = envname,
    libraries = pysparklyr_env$ml_libraries,
    msg = "Required Python libraries to run ML functions are missing"
  )
}

ml_get_params <- function(x) {
  py_x <- get_spark_pyobj(x)
  params <- invoke(py_x, "params")
  params |>
    map(\(.x) {
      nm <- .x$name
      nm <- paste0("get", toupper(substr(nm, 1, 1)), substr(nm, 2, nchar(nm)))
      tr <- try(invoke(py_x, nm), silent = TRUE)
      if (inherits(tr, "spark_pyobj")) {
        tr <- tr |>
          python_obj_get() |>
          py_to_r()
      }
      out <- ifelse(inherits(tr, "try-error"), list(), list(tr))
      set_names(out, .x$name)
    }) |>
    flatten() |>
    discard(is.null)
}

#' @export
print.ml_connect_estimator <- function(x, ...) {
  pyobj <- python_obj_get(x)
  msg <- ml_get_last_item(class(pyobj)[[1]])
  cli_div(theme = cli_colors())
  cli_inform("<{.header {msg}}>")
  cli_end()
}

test_that("ml_formula() works", {
  expect_snapshot(ml_formula(am ~ mpg, mtcars))
  expect_error(
    ml_formula(am ~ mpg * cyl, mtcars),
    "Formula resulted in an invalid parameter set"
  )
})

test_that("snake_to_camel() works", {
  expect_equal(
    snake_to_camel("var_one"),
    "varOne"
  )
})

test_that("ml_connect_not_supported() works", {
  expect_silent(
    ml_connect_not_supported(
      args = list(),
      not_supported = c(
        "elastic_net_param",
        "reg_param",
        "threshold",
        "aggregation_depth",
        "fit_intercept",
        "raw_prediction_col",
        "uid",
        "weight_col"
      )
    )
  )

  expect_error(
    ml_connect_not_supported(
      args = list(reg_param = 1),
      not_supported = c(
        "elastic_net_param",
        "reg_param",
        "threshold",
        "aggregation_depth",
        "fit_intercept",
        "raw_prediction_col",
        "uid",
        "weight_col"
      ),
      "The following argument(s) are not supported by Spark Connect:"
    )
  )
})

test_that("ml_installed() works on simulated interactive session", {
  skip_if_not_databricks()
  test_env <- test_databricks_stump_env() |>
    path_dir() |>
    path_dir()
  print(test_env)
  local_mocked_bindings(
    check_interactive = function(...) TRUE,
    check_rstudio = function(...) TRUE,
    menu = function(...) {
      return(1)
    },
    py_install = function(...) invisible()
  )
  expect_snapshot(
    ml_installed(envname = test_env)
  )
})

ml_binary_classification_evaluator_impl <- function(
  x,
  label_col = "label",
  raw_prediction_col = "rawPrediction",
  metric_name = "areaUnderROC",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "BinaryClassificationEvaluator",
    has_fit = FALSE,
    ml_type = "evaluation"
  )
}
#' @export
ml_binary_classification_evaluator.pyspark_connection <- ml_binary_classification_evaluator_impl
#' @export
ml_binary_classification_evaluator.tbl_pyspark <- ml_binary_classification_evaluator_impl

ml_multiclass_classification_evaluator_impl <- function(
  x,
  label_col = "label",
  prediction_col = "prediction",
  metric_name = "f1",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "MulticlassClassificationEvaluator",
    has_fit = FALSE,
    ml_type = "evaluation"
  )
}
#' @export
ml_multiclass_classification_evaluator.pyspark_connection <- ml_multiclass_classification_evaluator_impl
#' @export
ml_multiclass_classification_evaluator.tbl_pyspark <- ml_multiclass_classification_evaluator_impl

ml_clustering_evaluator_impl <- function(
  x,
  features_col = "features",
  prediction_col = "prediction",
  metric_name = "silhouette",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "ClusteringEvaluator",
    has_fit = FALSE,
    ml_type = "evaluation"
  )
}
#' @export
ml_clustering_evaluator.pyspark_connection <- ml_clustering_evaluator_impl
#' @export
ml_clustering_evaluator.tbl_pyspark <- ml_clustering_evaluator_impl

ml_regression_evaluator_impl <- function(
  x,
  label_col = "label",
  prediction_col = "prediction",
  metric_name = "rmse",
  uid = NULL,
  ...
) {
  ml_process_fn(
    args = c(as.list(environment()), list(...)),
    fn = "RegressionEvaluator",
    has_fit = FALSE,
    ml_type = "evaluation"
  )
}
#' @export
ml_regression_evaluator.pyspark_connection <- ml_regression_evaluator_impl
#' @export
ml_regression_evaluator.tbl_pyspark <- ml_regression_evaluator_impl

use_envname <- function(
  envname = NULL,
  backend = "pyspark",
  version = NULL,
  messages = FALSE,
  match_first = FALSE,
  ignore_reticulate_python = FALSE,
  ask_if_not_installed = FALSE,
  main_library = NULL,
  python_version = NULL
) {
  if (is.null(main_library) && !is.null(backend)) {
    cli_abort("Backend `{backend}` not valid")
  }
  cli_div(theme = cli_colors())

  ret_python <- reticulate_python_check(ignore_reticulate_python, unset = FALSE)

  if (ret_python != "") {
    return(set_names(ret_python, "env_var"))
  }

  if (!is.null(envname)) {
    return(set_names(envname, "argument"))
  }

  version_from_pypi <- FALSE
  if (is.null(version)) {
    if (!is.null(main_library)) {
      lib_info <- python_library_info(main_library, fail = FALSE, verbose = FALSE)
      if (!is.null(lib_info)) {
        version <- lib_info$version
        version_from_pypi <- TRUE
      }
    }
    if (is.null(version)) {
      cli_abort("A cluster {.code version} is required, please provide one")
    }
  }

  env_base <- glue("r-sparklyr-{backend}-")
  run_code <- glue("pysparklyr::install_{backend}(version = \"{version}\")")
  run_full <- "{.header Run: {.run {run_code}} to install.}"

  con_label <- connection_label(backend)
  sp_version <- version_prep(version)
  envname <- as.character(glue("{env_base}{sp_version}"))
  envs <- find_environments(env_base)

  match_one <- length(envs) > 0
  match_exact <- length(envs[envs == envname]) > 0
  install_ver <- version

  if (!is.null(main_library) && !match_exact) {
    lib_info <- python_library_info(main_library, fail = FALSE, verbose = FALSE)
    if (!is.null(lib_info)) {
      latest_ver <- lib_info$version
      if (version == "latest") {
        version <- latest_ver
      }
      vers <- compareVersion(latest_ver, version)
      install_recent <- vers == 1
      # For cases when the cluster's version is higher than the latest library
      if (vers == -1) {
        envname <- as.character(glue("{env_base}{latest_ver}"))
        install_ver <- latest_ver
      }
    }
  } else {
    install_recent <- TRUE
  }

  msg_default <- paste0(
    "{.header You do not have a Python environment that matches your",
    " {.emph {con_label}} cluster}"
  )

  msg_1 <- NULL
  msg_2 <- NULL
  msg_yes <- NULL
  msg_no <- NULL

  # There were 0 environments found
  if (!match_one && !match_exact) {
    ret_name <- if (version_from_pypi) "latest" else "unavailable"
    ret <- set_names(envname, ret_name)
    msg_1 <- msg_default
    msg_no <- " - Will use the default Python environment"
  }

  # Found an exact match
  if (match_one && match_exact) {
    ret <- set_names(envname, "exact")
  }

  # There are environments, but no exact match, and argument says
  # to choose the most recent environment
  if (match_one && !match_exact && match_first) {
    ret <- set_names(envs[1], "first")
    if (install_recent) {
      msg_1 <- msg_default
      msg_no <- glue(" - Will use alternate environment ({ret})")
    } else {
      ask_if_not_installed <- FALSE
      run_full <- NULL
      msg_1 <- paste0(
        "{.header Library {.emph {con_label}} version {.emph {version}} is not ",
        "yet available}"
      )
    }
  }

  # There are environments, but no exact match
  if (match_one && !match_exact && !match_first) {
    msg_1 <- msg_default
    msg_no <- " - Will use the default Python environment"
    ret_name <- if (version_from_pypi) "latest" else "unavailable"
    ret <- set_names(envname, ret_name)
  }

  ret_name <- names(ret)
  if (messages && ret_name != "exact") {
    if (ask_if_not_installed) {
      cli_alert_warning(msg_1)
      cli_bullets(c(
        " " = msg_2,
        " " = "{.header Do you wish to install {con_label} version {install_ver}?}"
      ))
      choice <- menu(
        choices = c(
          paste0("Yes", msg_yes),
          paste0("No", msg_no),
          "Cancel"
        )
      )
      if (choice == 1) {
        ret <- set_names(envname, "prompt")
        exec(
          .fn = glue("install_{backend}"),
          version = version,
          as_job = FALSE
        )
      }
      if (choice == 2) {
        ret <- set_names(ret, "prompt")
      }
      if (choice == 3) {
        stop_quietly()
      }
    } else {
      if (ret_name %in% c("unavailable", "latest")) {
        reqs <- python_requirements(
          backend = backend,
          main_library = main_library,
          version = version,
          python_version = python_version,
          install_ml = FALSE,
          add_torch = FALSE
        )
        reticulate::py_require(
          packages = reqs$packages,
          python_version = reqs$python_version
        )
      }
      if (ret_name == "first") {
        cli_alert_warning(msg_1)
        cli_bullets(c(
          " " = msg_2,
          " " = run_full
        ))
      }
    }
    cli_end()
  }
  ret
}

find_environments <- function(x) {
  conda_names <- tryCatch(conda_list()$name, error = function(e) character())
  ve_names <- virtualenv_list()
  all_names <- c(ve_names, conda_names)
  sub_names <- substr(all_names, 1, nchar(x))
  matched <- all_names[sub_names == x]
  sorted <- sort(matched, decreasing = TRUE)
  sorted
}

python_requirements <- function(
  backend = NULL,
  main_library = NULL,
  ml_version = NULL,
  version = NULL,
  python_version = NULL,
  install_ml = FALSE,
  add_torch = FALSE
) {
  cli_div(theme = cli_colors())

  if (is.null(python_version) && backend == "databricks") {
    python_version <- databricks_dbr_python(version)
  }

  library_info <- python_library_info(
    library_name = main_library,
    library_version = version,
    verbose = is.null(python_version)
  )

  if (!is.null(library_info)) {
    if (is.null(python_version)) {
      python_version <- library_info$requires_python
    }
    version <- library_info$version
    ver_name <- version
  } else {
    if (!is.null(version)) {
      ver_name <- version_prep(version)
      if (version == ver_name) {
        version <- paste0(version, ".*")
      }
    } else {
      cli_abort(
        c(
          "No `version` provided, and none could be found",
          " " = "Please run again with a valid version number"
        ),
        call = NULL
      )
    }
  }

  requires_dist <- as.character(library_info$requires_dist)
  packages <- c(
    paste0(main_library, "==", version),
    if (length(requires_dist)) {
      with_extra <- grepl("; extra", requires_dist)
      extra_str <- strsplit(requires_dist[with_extra], "; extra")
      extra_str <- lapply(extra_str, function(x) x[[1]])
      extra_str <- as.character(extra_str)
      extra_str <- unique(extra_str)
      c(requires_dist[!with_extra], extra_str)
    } else {
      c(
        "pandas!=2.1.0", # deprecation warnings
        "PyArrow",
        "grpcio",
        "google-api-python-client",
        "grpcio_status",
        "databricks-sdk",
        "zstandard"
      )
    }
  )

  if (add_torch && install_ml) {
    packages <- c(packages, pysparklyr_env$ml_libraries)
  }

  packages <- c(packages, "pip")

  list(
    packages = packages,
    python_version = python_version
  )
}

test_that("Install environment", {
  withr::with_envvar(
    new = c(
      "WORKON_HOME" = use_new_temp_env()
    ),
    {
      expect_output(
        install_pyspark("3.4", as_job = FALSE)
      )
    }
  )
})

test_that("Use first one", {
  env_path <- use_new_temp_env()
  dir_create(path(env_path, "r-sparklyr-pyspark-3.4"))
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      expect_message(
        x <- use_envname(
          main_library = "pyspark",
          version = "1.1",
          messages = TRUE,
          match_first = TRUE,
          ask_if_not_installed = FALSE
        ),
        "You do not have a Python environment that matches"
      )
      expect_named(x, "first")
    }
  )
})

test_that("Error if 'use_first' is not TRUE", {
  env_path <- use_new_temp_env()
  dir_create(path(env_path, "r-sparklyr-pyspark-3.4"))
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      expect_message(
        x <- use_envname(
          version = "100.0",
          messages = TRUE,
          match_first = TRUE,
          main_library = "pyspark",
          ask_if_not_installed = FALSE
        ),
        "Library Spark Connect version 100.0 is not yet available"
      )
    }
  )
})

test_that("Error if 'use_first' is not TRUE", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_temp_env()),
    {
      expect_error(
        x <- use_envname(
          version = "1.1",
          messages = TRUE,
          match_first = FALSE,
          ask_if_not_installed = FALSE
        )
      )
    }
  )
})

test_that("'Ask to install', simulates menu selection 'Yes'", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_temp_env()),
    {
      local_mocked_bindings(
        check_interactive = function(...) TRUE,
        check_rstudio = function(...) TRUE,
        menu = function(...) {
          return(1)
        },
        exec = function(...) list(...)
      )
      expect_equal(
        use_envname(
          main_library = "pyspark",
          version = "1.1",
          messages = TRUE,
          match_first = FALSE,
          ask_if_not_installed = TRUE
        ),
        rlang::set_names("r-sparklyr-pyspark-1.1", "prompt")
      )
    }
  )
})

test_that("'Ask to install', simulates menu selection 'No'", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_temp_env()),
    {
      local_mocked_bindings(
        check_interactive = function(...) TRUE,
        check_rstudio = function(...) TRUE,
        menu = function(...) {
          return(2)
        },
        exec = function(...) list(...)
      )
      expect_equal(
        use_envname(
          main_library = "pyspark",
          version = "1.1",
          messages = TRUE,
          match_first = FALSE,
          ask_if_not_installed = TRUE
        ),
        rlang::set_names("r-sparklyr-pyspark-1.1", "prompt")
      )
    }
  )
})

test_that("'Ask to install', simulates menu selection 'Cancel'", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_temp_env()),
    {
      local_mocked_bindings(
        check_interactive = function(...) TRUE,
        check_rstudio = function(...) TRUE,
        menu = function(...) {
          return(3)
        },
        exec = function(...) list(...)
      )
      expect_error(
        use_envname(
          version = "1.1",
          messages = TRUE,
          match_first = FALSE,
          ask_if_not_installed = TRUE
        )
      )
    }
  )
})

test_that("Expect error when no 'version' is provided and no 'main_library'", {
  expect_error(use_envname(version = NULL, main_library = NULL))
})

test_that("Auto-detect version from PyPI when version is NULL", {
  env_path <- use_new_temp_env()
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      local_mocked_bindings(
        python_library_info = function(
          library_name,
          library_version = NULL,
          verbose = TRUE,
          fail = TRUE,
          timeout = 2
        ) {
          list(
            version = "3.5.0",
            requires_python = ">=3.8",
            name = library_name
          )
        }
      )
      x <- use_envname(
        main_library = "pyspark",
        version = NULL,
        messages = FALSE,
        match_first = FALSE,
        ask_if_not_installed = FALSE
      )
      expect_named(x, "latest")
      expect_equal(as.character(x), "r-sparklyr-pyspark-3.5")
    }
  )
})

test_that("Auto-detect uses latest version for databricks.connect", {
  env_path <- use_new_temp_env()
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      local_mocked_bindings(
        python_library_info = function(
          library_name,
          library_version = NULL,
          verbose = TRUE,
          fail = TRUE,
          timeout = 2
        ) {
          list(
            version = "14.2.0",
            requires_python = ">=3.8",
            name = library_name
          )
        }
      )
      x <- use_envname(
        main_library = "databricks.connect",
        version = NULL,
        backend = "databricks",
        messages = FALSE,
        match_first = FALSE,
        ask_if_not_installed = FALSE
      )
      expect_named(x, "latest")
      expect_equal(as.character(x), "r-sparklyr-databricks-14.2")
    }
  )
})

test_that("Falls back to error when PyPI query fails and no version provided", {
  withr::with_envvar(
    new = c("WORKON_HOME" = use_new_temp_env()),
    {
      local_mocked_bindings(
        python_library_info = function(
          library_name,
          library_version = NULL,
          verbose = TRUE,
          fail = TRUE,
          timeout = 2
        ) {
          return(NULL)
        }
      )
      expect_error(
        use_envname(
          main_library = "pyspark",
          version = NULL,
          messages = FALSE
        ),
        "A cluster.*version.*is required"
      )
    }
  )
})

test_that("Explicit version returns 'unavailable' not 'latest'", {
  env_path <- use_new_temp_env()
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      x <- use_envname(
        main_library = "pyspark",
        version = "3.5.0",
        messages = FALSE,
        match_first = FALSE,
        ask_if_not_installed = FALSE
      )
      expect_named(x, "unavailable")
      expect_equal(as.character(x), "r-sparklyr-pyspark-3.5")
    }
  )
})

test_that("Auto-detect finds exact match when environment exists", {
  env_path <- use_new_temp_env()
  dir_create(path(env_path, "r-sparklyr-pyspark-3.5"))
  withr::with_envvar(
    new = c("WORKON_HOME" = env_path),
    {
      local_mocked_bindings(
        python_library_info = function(
          library_name,
          library_version = NULL,
          verbose = TRUE,
          fail = TRUE,
          timeout = 2
        ) {
          list(
            version = "3.5.0",
            requires_python = ">=3.8",
            name = library_name
          )
        }
      )
      x <- use_envname(
        main_library = "pyspark",
        version = NULL,
        messages = FALSE,
        match_first = FALSE,
        ask_if_not_installed = FALSE
      )
      expect_named(x, "exact")
      expect_equal(as.character(x), "r-sparklyr-pyspark-3.5")
    }
  )
})


test_that("Requirements work", {
  reqs <- python_requirements(
    backend = "pyspark",
    main_library = "pyspark",
    version = "3.4"
  )
  expect_equal(c("packages", "python_version"), names(reqs))
})

suppressPackageStartupMessages(library(dplyr))
suppressPackageStartupMessages(library(sparklyr))
suppressPackageStartupMessages(library(cli))

## Clean up at end
withr::defer({
  # Disconnecting from Spark
  try(spark_disconnect_all(), silent = TRUE)

  # Stopping Spark Connect service
  try(spark_connect_service_stop(use_test_version_spark()), silent = TRUE)

  # Kill process if still running
  if (exists("spark_connect_process", envir = .GlobalEnv)) {
    prs <- get("spark_connect_process", envir = .GlobalEnv)
    if (prs$is_alive()) {
      try(prs$kill(), silent = TRUE)
    }
    rm("spark_connect_process", envir = .GlobalEnv)
  }

  # Deleting main Python environment (commented out as it may be too aggressive)
  # try(fs::dir_delete(use_test_env()), silent = TRUE)
}, testthat::teardown_env())

#' @export
spark_connect_method.spark_method_spark_connect <- function(
  x,
  method,
  master,
  spark_home,
  config = pyspark_config(),
  app_name,
  version = NULL,
  hadoop_version,
  extensions,
  scala_version,
  ...
) {
  version <- version %||% Sys.getenv("SPARK_VERSION")

  if (version == "") {
    cli_abort("Spark `version` is required, please provide")
  }

  args <- list(...)
  envname <- args$envname

  envname <- use_envname(
    backend = "pyspark",
    main_library = "pyspark",
    version = version,
    envname = envname,
    messages = TRUE,
    match_first = TRUE,
    python_version = args$python_version
  )

  if (is.null(envname)) {
    return(invisible())
  }

  pyspark <- import_check("pyspark", envname)
  pyspark_sql <- pyspark$sql
  conn <- pyspark_sql$SparkSession$builder$remote(master)
  con_class <- "connect_spark"
  master_label <- glue("Spark Connect - {master}")

  initialize_connection(
    conn = conn,
    master_label = master_label,
    con_class = con_class,
    cluster_id = NULL,
    method = method,
    config = config
  )
}

setOldClass(
  c("connect_spark", "pyspark_connection", "spark_connection")
)

ml_process_fn <- function(
  args,
  fn,
  has_fit = TRUE,
  ml_type = "",
  ml_fn = NULL
) {
  ml_installed()
  x <- args$x
  conn <- spark_connection(x)
  if (spark_version(conn) >= "4.0.0") {
    python_library <- glue("pyspark.ml.{ml_type}")
  } else {
    python_library <- glue("pyspark.ml.connect.{ml_type}")
  }

  x <- args$x
  args <- args[names(args) != "x"]
  jobj <- ml_execute(
    args = args[names(args) != "formula"],
    python_library = python_library,
    fn = fn,
    sc = spark_connection(x)
  )
  args[["uid"]] <- invoke(jobj, "uid")
  prep_obj <- c(
    list(
      uid = invoke(jobj, "uid"),
      param_map = ml_get_params(jobj),
      .jobj = jobj
    )
  )
  obj_class <- c(
    glue("ml_{ml_fn}"),
    "ml_connect_estimator",
    "ml_estimator",
    "ml_pipeline_stage"
  )
  stage <- structure(
    prep_obj,
    class = obj_class
  )
  if (inherits(x, "pyspark_connection")) {
    return(stage)
  }
  if (inherits(x, "ml_connect_pipeline")) {
    return(ml_connect_add_stage(x, stage))
  }
  if (inherits(x, "tbl_pyspark")) {
    tbl_prep <- ml_prep_dataset(
      x = x,
      formula = args[["formula"]],
      label_col = args[["label_col"]],
      features_col = args[["features_col"]] %||% args[["raw_prediction_col"]],
      lf = "only",
      additional = args[["censor_col"]],
      fitting = TRUE
    )
    if (has_fit) {
      stage <- ml_fit_impl(stage, tbl_prep)
    }
    if (ml_type == "evaluation") {
      ret <- stage |>
        invoke("evaluate", python_obj_get(x)) |>
        python_obj_get() |>
        py_to_r()
    } else {
      attrs <- attributes(tbl_prep)
      ret <- structure(
        list(
          pipeline = stage,
          features = attrs$features,
          label = attrs$label
        ),
        class = c(
          "ml_connect_model",
          glue("ml_model_{ml_fn}"),
          glue("ml_model_{ml_type}"),
          "ml_model_prediction",
          "ml_model"
        )
      )
    }
    return(ret)
  }
}

ml_process_transformer <- function(args, fn, has_fit = TRUE) {
  ml_installed()
  x <- args$x
  conn <- spark_connection(x)
  args <- args[names(args) != "x"]
  if (!is.null(args[["formula"]])) {
    if (rlang::is_formula(args[["formula"]])) {
      args[["formula"]] <- deparse(args[["formula"]])
    }
  }
  if (spark_version(conn) >= "4.0.0") {
    python_library <- "pyspark.ml.feature"
  } else {
    python_library <- "pyspark.ml.connect.feature"
  }
  jobj <- ml_execute(
    args = args,
    python_library = python_library,
    fn = fn,
    sc = conn
  )
  prep_obj <- list(
    uid = invoke(jobj, "uid"),
    param_map = ml_get_params(jobj),
    .jobj = jobj
  )
  obj_class <- "ml_pipeline_stage"
  stage <- structure(
    prep_obj,
    class = obj_class
  )
  if (inherits(x, "pyspark_connection")) {
    caller <- deparse(rlang::caller_call())
    ml_fn <- unlist(strsplit(caller, "\\."))[1]
    class(stage) <- c(ml_fn, "ml_transformer", class(stage))
    return(stage)
  }
  if (inherits(x, "ml_connect_pipeline")) {
    return(ml_connect_add_stage(x, stage))
  }
  if (inherits(x, "tbl_pyspark")) {
    tbl_prep <- ml_prep_dataset(
      x = x,
      label_col = args[["label_col"]],
      features_col = args[["input_col"]],
      lf = "all"
    )
    if (has_fit) {
      stage <- ml_fit_impl(stage, tbl_prep)
    }
    ret <- transform_impl(
      x = stage,
      dataset = tbl_prep,
      prep = FALSE,
      conn = conn,
      as_df = TRUE
    )
  }
  return(ret)
}

#' @export
spark_write_delta.tbl_pyspark <- function(
  x,
  path,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "delta",
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list()
  )
}

#' @export
spark_write_table.tbl_pyspark <- function(
  x,
  name,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  args <- list(...)
  save_action <- ifelse(identical(mode, "append"), "insertInto", "saveAsTable")
  pyspark_write_generic(
    x = x,
    path = name,
    format = args$format,
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list(),
    save_action = save_action,
    expand_path = FALSE
  )
}

#' @export
spark_write_csv.tbl_pyspark <- function(
  x,
  path,
  header = TRUE,
  delimiter = ",",
  quote = "\"",
  escape = "\\",
  charset = "UTF-8",
  null_value = NULL,
  options = list(),
  mode = NULL,
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "csv",
    partition_by = partition_by,
    mode = mode,
    options = options,
    args = list(
      header = header,
      delimiter = delimiter,
      quote = quote,
      charset = charset
    )
  )
}

#' @export
spark_write_parquet.tbl_pyspark <- function(
  x,
  path,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "parquet",
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list()
  )
}

#' @export
spark_write_text.tbl_pyspark <- function(
  x,
  path,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "text",
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list()
  )
}

#' @export
spark_write_orc.tbl_pyspark <- function(
  x,
  path,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "orc",
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list()
  )
}

#' @export
spark_write_json.tbl_pyspark <- function(
  x,
  path,
  mode = NULL,
  options = list(),
  partition_by = NULL,
  ...
) {
  pyspark_write_generic(
    x = x,
    path = path,
    format = "json",
    mode = mode,
    partition_by = partition_by,
    options = options,
    args = list()
  )
}

pyspark_write_generic <- function(
  x,
  path,
  format = NULL,
  mode,
  partition_by,
  options,
  args,
  save_action = "save",
  expand_path = TRUE
) {
  query <- tbl_pyspark_sdf(x)

  if (is.null(partition_by)) {
    query_prep <- query$repartition(1L)$write
  } else {
    query_prep <- query$write$partitionBy(partition_by)
  }

  opts <- c(args, options)

  path <- ifelse(expand_path, path_expand(path), path)

  if (!is.null(format)) {
    x <- py_invoke(query_prep, "format", format)
  } else {
    x <- query_prep
  }

  if (!is.null(mode)) {
    x <- py_invoke(x, "mode", mode)
  }

  invisible(
    x |>
      py_invoke_options(options = opts) |>
      py_invoke(save_action, path)
  )
}

py_invoke <- function(x, fun, ...) {
  x |>
    py_get_attr(fun) |>
    py_call(...)
}

py_invoke_option <- function(x, option, value) {
  if (!is.null(value)) {
    out <- py_invoke(x, "option", option, value)
  } else {
    out <- x
  }
  out
}

py_invoke_options <- function(x, options) {
  for (i in seq_along(options)) {
    curr_option <- options[i]
    x <- py_invoke_option(x, names(curr_option), curr_option[[1]])
  }
  x
}

#' @export
spark_read_csv.pyspark_connection <- function(
  sc,
  name = NULL,
  path = name,
  header = TRUE,
  columns = NULL,
  infer_schema = is.null(columns),
  delimiter = ",",
  quote = "\"",
  escape = "\\",
  charset = "UTF-8",
  null_value = NULL,
  options = list(),
  repartition = 0,
  memory = TRUE,
  overwrite = TRUE,
  ...
) {
  pyspark_read_generic(
    sc = sc,
    path = path,
    name = name,
    format = "csv",
    memory = memory,
    repartition = repartition,
    overwrite = overwrite,
    options = options,
    args = list(
      header = header,
      schema = columns,
      inferSchema = infer_schema,
      sep = delimiter,
      quote = quote,
      escape = escape,
      encoding = charset,
      nullValue = null_value
    )
  )
}

#' @export
spark_read_text.pyspark_connection <- function(
  sc,
  name = NULL,
  path = name,
  repartition = 0,
  memory = TRUE,
  overwrite = TRUE,
  options = list(),
  whole = FALSE,
  ...
) {
  pyspark_read_generic(
    sc = sc,
    path = path,
    name = name,
    format = "text",
    memory = memory,
    repartition = repartition,
    overwrite = overwrite,
    options = options,
    args = list(
      wholetext = whole
    )
  )
}

#' @export
spark_read_json.pyspark_connection <- function(
  sc,
  name = NULL,
  path = name,
  options = list(),
  repartition = 0,
  memory = TRUE,
  overwrite = TRUE,
  columns = NULL,
  ...
) {
  pyspark_read_generic(
    sc = sc,
    path = path,
    name = name,
    format = "json",
    memory = memory,
    repartition = repartition,
    overwrite = overwrite,
    options = options,
    args = list(
      columns = columns
    )
  )
}

#' @export
spark_read_orc.pyspark_connection <- function(
  sc,
  name = NULL,
  path = name,
  options = list(),
  repartition = 0,
  memory = TRUE,
  overwrite = TRUE,
  columns = NULL,
  schema = NULL,
  ...
) {
  pyspark_read_generic(
    sc = sc,
    path = path,
    name = name,
    format = "orc",
    memory = memory,
    repartition = repartition,
    overwrite = overwrite,
    options = options,
    args = list(
      columns = columns,
      schema = schema
    )
  )
}

#' @export
spark_read_parquet.pyspark_connection <- function(
  sc,
  name = NULL,
  path = name,
  options = list(),
  repartition = 0,
  memory = TRUE,
  overwrite = TRUE,
  columns = NULL,
  schema = NULL,
  ...
) {
  con <- python_conn(sc)

  new_schema <- NULL
  if (!is.null(columns)) {
    new_schema <- columns
  }
  if (!is.null(schema)) {
    new_schema <- schema
  }
  if (!is.null(new_schema)) {
    sch <- con$read$schema(new_schema)
    read <- sch$parquet(path)
  } else {
    read <- con$read$parquet(path)
  }

  pyspark_read_generic(
    sc = sc,
    path = NULL,
    name = name,
    format = NULL,
    memory = memory,
    repartition = repartition,
    overwrite = overwrite,
    options = NULL,
    args = NULL,
    py_obj = read
  )
}


pyspark_read_generic <- function(
  sc,
  path,
  name,
  format,
  memory,
  repartition,
  overwrite,
  args,
  options = list(),
  py_obj = NULL
) {
  opts <- c(args, options)
  rename_fields <- FALSE
  schema <- args$schema
  if (!is.null(schema)) {
    if (!inherits(schema, "list")) {
      rename_fields <- TRUE
    } else {
      abort("Complex schemas not yet supported")
    }
  }

  if (is.null(py_obj)) {
    x <- python_conn(sc)$read |>
      py_invoke_options(options = opts) |>
      py_invoke(format, path_expand(path))
  } else {
    x <- py_obj
  }

  repartition <- as.integer(repartition)
  if (repartition > 0) {
    x$repartition(repartition)
  }

  if (overwrite && name %in% dbListTables(sc)) {
    dbRemoveTable(sc, name)
  }

  if (is.null(name) | identical(path, name)) {
    name <- gen_sdf_name(path)
  }

  if (rename_fields) {
    cur_names <- x$columns
    for (i in seq_along(cur_names)) {
      x <- x$withColumnRenamed(
        existing = cur_names[[i]],
        new = schema[[i]]
      )
    }
  }

  if (memory && !sc$serverless) {
    if (repartition > 1) {
      storage_level <- import("pyspark.storagelevel")
      x$createOrReplaceTempView(name)
      x$persist(storage_level$StorageLevel$MEMORY_AND_DISK)
      out <- tbl(sc, name)
    } else {
      out <- x |>
        tbl_pyspark_temp(sc) |>
        cache_query(name = name)
    }
  } else {
    x$createOrReplaceTempView(name)
    out <- tbl(sc, name)
  }

  spark_ide_connection_updated(sc, name)
  out
}

gen_sdf_name <- function(path) {
  x <- path |>
    path_file() |>
    path_ext_remove()

  random_string(gsub("[[:punct:]]", "", x))
}

#' @export
spark_connect_method.spark_method_snowpark_connect <- function(
  x,
  method,
  master,
  spark_home,
  config = NULL,
  app_name,
  version = NULL,
  hadoop_version,
  extensions,
  scala_version,
  ...
) {
  if (missing(master) || is.null(master)) {
    master <- Sys.getenv("SNOWFLAKE_ACCOUNT", unset = NA)
    if (is.na(master)) {
      cli_abort(paste(
        "Please provide a `master` argument. It needs to be your Snowflake's",
        "'Account Identifier', which can be found in the portal."
      ))
    }
  }
  args <- list(...)
  envname <- use_envname(
    backend = "snowflake",
    main_library = "snowflake-snowpark-python",
    version = version %||% "latest",
    envname = args$envname,
    messages = TRUE,
    match_first = TRUE,
    python_version = args$python_version
  )
  if (is.null(envname)) {
    return(invisible())
  }
  pyspark <- import_check("snowflake.snowpark", envname)
  connection_parameters <- args$connection_parameters %||% list()
  connection_parameters$account <- master
  if (is.null(connection_parameters$password)) {
    # Checks to see if there is a Posit Workbench token available
    snowflake_home <- Sys.getenv("SNOWFLAKE_HOME", unset = NA)
    if (!is.na(snowflake_home)) {
      if (grepl("workbench", snowflake_home)) {
        token <- try(
          workbench_snowflake_token(master, snowflake_home),
          silent = TRUE
        )
        if (!inherits(token, "try-error")) {
          connection_parameters$authenticator <- "oauth"
          connection_parameters$token <- token
        }
      }
    }
  }
  missing_path <- NULL
  if (is.null(connection_parameters$warehouse)) {
    missing_path <- "warehouse"
  }
  if (is.null(connection_parameters$database)) {
    missing_path <- c(missing_path, "database")
  }
  if (is.null(connection_parameters$schema)) {
    missing_path <- c(missing_path, "schema")
  }
  if (!is.null(missing_path)) {
    missing_path <- paste0("'", missing_path, "'")
    cli_alert_warning(
      "Argument{?s} {.pkg {missing_path}} will be needed to easily navigate Snowflake"
    )
    cli_bullets(
      c(" " = "Please use the `connection_parameters` argument to pass them.")
    )
  }
  conn <- pyspark$Session$builder$configs(connection_parameters)
  con_class <- "connect_snowflake"
  master_label <- glue("Snowpark Connect - {master}")
  initialize_connection(
    conn = conn,
    master_label = master_label,
    con_class = con_class,
    cluster_id = NULL,
    method = method,
    config = NULL,
    misc = list(
      sql_catalogs = "show databases",
      sql_tables_schema = "show tables in {schema}",
      sql_tables_catalog_schema = "show tables in {catalog}.{schema}",
      sql_schemas_catalog = "show schemas in database {catalog}",
      sql_schemas = "show schemas"
    ),
    quote = ""
  )
}

setOldClass(
  c("connect_snowflake", "pyspark_connection", "spark_connection")
)

# Copy of code from `odbc` which gets the Snowflake token from Workbench
# https://github.com/r-dbi/odbc/blob/main/R/driver-snowflake.R
workbench_snowflake_token <- function(account, sf_home) {
  cfg <- readLines(file.path(sf_home, "connections.toml"))
  # We don't attempt a full parse of the TOML syntax, instead relying on the
  # fact that this file will always contain only one section.
  if (!any(grepl(account, cfg, fixed = TRUE))) {
    # The configuration doesn't actually apply to this account.
    return(NULL)
  }
  line <- grepl("token = ", cfg, fixed = TRUE)
  token <- gsub("token = ", "", cfg[line])
  if (nchar(token) == 0) {
    return(NULL)
  }
  # Drop enclosing quotes.
  gsub("\"", "", token)
}

#' @export
print.spark_pyobj <- function(x, ...) {
  pyspark_obj <- x$pyspark_obj
  pyspark_class <- class(pyspark_obj)[[1]]
  pyspark_print <- py_str(pyspark_obj)
  cli_div(theme = cli_colors())
  cli_text("{.header - {.emph PySpark object}}")
  cli_bullets(
    c(
      " " = "{.class {pyspark_class}}",
      " " = "{.info {pyspark_print}}"
    )
  )
  cli_end()
}

#' @export
sdf_read_column.spark_pyjobj <- function(x, column) {
  col_df <- x |>
    spark_dataframe() |>
    invoke("select", column) |>
    collect()

  col_df[[column]]
}

#' @export
spark_version.spark_pyobj <- function(sc) {
  python_conn(sc)$version
}

#' @export
spark_connection.spark_pyobj <- function(x, ...) {
  x[["connection"]]
}

#' @export
spark_dataframe.spark_pyobj <- function(x, ...) {
  x
}

#' @export
collect.spark_pyobj <- function(x, ...) {
  to_pandas_cleaned(x$pyspark_obj)
}

as_spark_pyobj <- function(obj, conn, class = NULL) {
  structure(
    list(
      pyspark_obj = obj,
      connection = conn,
      class = class
    ),
    class = c("spark_pyobj", "spark_jobj")
  )
}
setOldClass(c("spark_pyobj", "spark_jobj"))

get_spark_pyobj <- function(obj) {
  if (inherits(obj, "spark_pyobj")) {
    return(obj)
  }
  out <- obj[[".jobj"]]
  if (is.null(out)) {
    out <- obj[["pipeline"]]
  }
  if (is.null(out)) {
    stop("No `spark_pyobj` found")
  }
  out
}

to_pandas_cleaned <- function(x) {
  fields <- x$dtypes
  orig_types <- map_chr(fields, \(.x) .x[[2]])

  dec_types <- map_lgl(orig_types, \(.x) grepl("decimal\\(", .x))

  if (sum(dec_types) > 0) {
    # Has to check if SQL functions can be imported because some backend
    # connection Python packages does not include it
    sf <- try(import("pyspark.sql.functions"), silent = TRUE)
    if (!inherits(sf, "try-error")) {
      for (field in fields[dec_types]) {
        fn <- field[[1]]
        x <- x$withColumn(fn, sf$col(fn)$cast("double"))
      }
    }
  }

  pandas_tbl <- x$toPandas()

  if (is.data.frame(pandas_tbl)) {
    collected <- pandas_tbl
  } else {
    # Pandas 3.0 conversion makes encases all columns inside lists
    collected <- try(
      {
        as.data.frame(pandas_tbl$values) |>
          lapply(\(col) {
            map_vec(
              col,
              \(x) {
                if (length(x) == 0 || is.nan(x)) {
                  NA
                } else {
                  x[[1]]
                }
              }
            )
          }) |>
          set_names(x$columns)
      },
      silent = TRUE
    )
    if (inherits(collected, "try-error")) {
      collected <- x$toArrow()
      collected <- collected$to_data_frame()
    }
  }

  # Handle NULL columns by converting them to appropriate empty vectors
  for (i in seq_along(collected)) {
    if (is.null(collected[[i]])) {
      collected[[i]] <- character(0)
    }
    # Convert arrow_binary to list
    if (inherits(collected[[i]], "arrow_binary")) {
      collected[[i]] <- as.list(collected[[i]])
    }
    # Convert UTC timestamps to local timezone
    if (inherits(collected[[i]], "POSIXct")) {
      # If the timestamp is in UTC, convert to local timezone
      if (!is.null(attr(collected[[i]], "tzone")) && attr(collected[[i]], "tzone") == "UTC") {
        attr(collected[[i]], "tzone") <- ""
      }
    }
    # Convert pandas Arrow-backed arrays to regular R vectors
    if (inherits(collected[[i]], "python.builtin.object")) {
      col_class <- try(class(collected[[i]]), silent = TRUE)
      if (!inherits(col_class, "try-error")) {
        # Check if it's an Arrow-backed array from pandas
        if (any(grepl("ArrowStringArray|ArrowExtensionArray", col_class))) {
          # Convert to regular character vector using numpy
          collected[[i]] <- try(
            as.character(collected[[i]]$to_numpy()),
            silent = TRUE
          )
          # Fallback to Python list conversion if to_numpy fails
          if (inherits(collected[[i]], "try-error")) {
            collected[[i]] <- as.character(collected[[i]]$tolist())
          }
        }
      }
    }
  }

  collected <- collected |>
    dplyr::as_tibble()

  col_types <- map_chr(
    collected,
    \(.x) {
      classes <- class(.x)
      classes[[1]]
    }
  )

  clean_col <- function(x, subset = TRUE) {
    if (length(x) == 0 || is.nan(x)) {
      x <- NA
    } else if (subset) {
      x <- x[[1]]
    }
    x
  }

  for (i in seq_len(ncol(collected))) {
    py_type <- orig_types[i]
    r_type <- col_types[i]
    col <- collected[[i]]
    collected[[i]] <-
      if (py_type == "date" && r_type == "list") {
        as.Date(map_vec(col, clean_col), origin = "1970-01-01")
      } else if (py_type == "date" && r_type == "character") {
        as.Date(col, origin = "1970-01-01")
      } else if (py_type == "date" && r_type %in% c("numeric", "integer", "logical")) {
        as.Date(col, origin = "1970-01-01")
      } else if (py_type == "boolean" && r_type == "list") {
        map_lgl(col, clean_col)
      } else if (py_type == "boolean" && r_type == "character") {
        as.logical(col)
      } else if (r_type == "integer" && py_type %in% c("bigint", "long")) {
        # Convert bigint/long to numeric since R integer is 32-bit
        as.numeric(col)
      } else if (r_type == "numeric" || r_type == "integer") {
        if (py_type %in% c("tinyint", "smallint", "int")) {
          ptype <- integer()
        } else {
          ptype <- numeric()
        }
        map_vec(col, clean_col, FALSE, .ptype = ptype)
      } else {
        col
      }
  }

  out <- tibble(collected)

  # Snowflake returns field names in double quotes
  colnames(out) <- gsub("\"", "", colnames(out))

  attr(out, "pandas.index") <- NULL
  out
}

# DBR error code returns as expected

    Code
      databricks_dbr_error(error)
    Condition
      Error in `databricks_dbr_error()`:
      ! Spark connection error
      * Possible cause = The cluster is not running, or not accessible
      * status = StatusCode.UNAVAILABLE
      * details = 'RESOURCE_DOES_NOT_EXIST: No cluster found matching: asdfasdf'

---

    Code
      databricks_dbr_error("")
    Condition
      Error in `databricks_dbr_error()`:
      ! Spark connection error
      * 

# Misc tests

    Code
      allowed_serverless_configs()
    Output
      [1] "spark.sql.legacy.timeParserPolicy" "spark.sql.session.timeZone"       
      [3] "spark.sql.shuffle.partitions"      "spark.sql.ansi.enabled"           

# Tests deploy_databricks() happy path cases

    Code
      deploy_databricks()
    Output
      $appDir
      NULL
      
      $lint
      [1] FALSE
      
      $python
      NULL
      
      $version
      [1] "17.3"
      
      $backend
      [1] "databricks"
      
      $main_library
      [1] "databricks-connect"
      
      $envVars
      [1] "DATABRICKS_HOST"  "DATABRICKS_TOKEN"
      
      $env_var_message
                                    i                                 
           "{.header Host URL:} test" "{.header Token:} '<REDACTED>'" 
      
      $account
      NULL
      
      $server
      NULL
      
      $confirm
      [1] FALSE
      

---

    Code
      deploy_databricks(host = "another1", token = "token")
    Output
      $appDir
      NULL
      
      $lint
      [1] FALSE
      
      $python
      NULL
      
      $version
      [1] "17.3"
      
      $backend
      [1] "databricks"
      
      $main_library
      [1] "databricks-connect"
      
      $envVars
      [1] "CONNECT_DATABRICKS_HOST"  "CONNECT_DATABRICKS_TOKEN"
      
      $env_var_message
                                    i                                 
       "{.header Host URL:} another1" "{.header Token:} '<REDACTED>'" 
      
      $account
      NULL
      
      $server
      NULL
      
      $confirm
      [1] FALSE
      

# Tests deploy_databricks() error cases

    Code
      deploy_databricks()
    Condition
      Error:
      ! Cluster setup errors:
        - No host URL was provided or found. Please either set the 'DATABRICKS_HOST' environment variable, or pass the `host` argument.

# Pipeline fits and predicts

    Code
      class(fitted)
    Output
      [1] "ml_connect_pipeline_model" "ml_pipeline_model"        
      [3] "ml_transformer"            "ml_connect_transformer"   
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

# Adds the ML libraries when prompted

    Code
      x
    Output
      $packages
       [1] "pyspark==3.5.0"           "pandas!=2.1.0"           
       [3] "PyArrow"                  "grpcio"                  
       [5] "google-api-python-client" "grpcio_status"           
       [7] "databricks-sdk"           "zstandard"               
       [9] "torch"                    "torcheval"               
      [11] "scikit-learn"            
      
      $envname
                   unavailable 
      "r-sparklyr-pyspark-3.5" 
      
      $method
      [1] "auto"
      
      $pip
      [1] TRUE
      

# install_as_job() is able to run

    Code
      install_as_job(as_job = TRUE)
    Message
      v Running installation as a RStudio job 

# Installation runs even if no response from PyPi

    Code
      x
    Output
      $packages
      [1] "pyspark==3.5.*"           "pandas!=2.1.0"           
      [3] "PyArrow"                  "grpcio"                  
      [5] "google-api-python-client" "grpcio_status"           
      [7] "databricks-sdk"           "zstandard"               
      
      $envname
                   unavailable 
      "r-sparklyr-pyspark-3.5" 
      
      $method
      [1] "auto"
      
      $pip
      [1] TRUE
      

# Install code is correctly created

    Code
      build_job_code(list(a = 1))
    Output
      [1] "pysparklyr:::install_environment(a = 1)"

# Databricks installation works

    Code
      out
    Output
      $main_library
      [1] "databricks-connect"
      
      $spark_method
      [1] "databricks_connect"
      
      $backend
      [1] "databricks"
      
      $ml_version
      [1] "14.1"
      
      $version
      [1] "14.1"
      
      $envname
      NULL
      
      $python_version
      NULL
      
      $new_env
      [1] TRUE
      
      $method
      [1] "auto"       "virtualenv" "conda"     
      
      $as_job
      [1] TRUE
      
      $install_ml
      [1] FALSE
      

---

    Code
      install_databricks(version = "13.1")
    Output
      $main_library
      [1] "databricks-connect"
      
      $spark_method
      [1] "databricks_connect"
      
      $backend
      [1] "databricks"
      
      $ml_version
      [1] "14.1"
      
      $version
      [1] "13.1"
      
      $envname
      NULL
      
      $python_version
      NULL
      
      $new_env
      [1] TRUE
      
      $method
      [1] "auto"       "virtualenv" "conda"     
      
      $as_job
      [1] TRUE
      
      $install_ml
      [1] FALSE
      

# Logistic regression works

    Code
      class(ml_logistic_regression(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_logistic_regression(sc))
    Output
      [1] "ml_logistic_regression" "ml_connect_estimator"   "ml_estimator"          
      [4] "ml_pipeline_stage"     

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"             "ml_model_logistic_regression"
      [3] "ml_model_classification"      "ml_model_prediction"         
      [5] "ml_model"                    

---

    Code
      table(fitted)
    Output
      fitted
       0  1 
      19 13 

# Linear regression works

    Code
      class(ml_linear_regression(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_linear_regression(sc))
    Output
      [1] "ml_linear_regression" "ml_connect_estimator" "ml_estimator"        
      [4] "ml_pipeline_stage"   

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"           "ml_model_linear_regression"
      [3] "ml_model_regression"        "ml_model_prediction"       
      [5] "ml_model"                  

---

    Code
      model
    Message
      
      -- MLib model: LinearRegressionModel --
      
      -- Coefficients: 
    Output
      [x] Intercept:   -0.231   [x] qsec:         0.2  
      [x] mpg:         -0.042   [x] vs:          -0.066
      [x] cyl:         -0.057   [x] am:           0.018
      [x] disp:         0.007   [x] gear:        -0.094
      [x] hp:          -0.003   [x] carb:         0.249
      [x] drat:        -0.09    
    Message
      
      -- Summary: 
    Output
      [x] coefficientStandardErrors:  0.021, 0.11, 0.001, 0.002, ...
      [x] devianceResiduals:          -0.348, 0.548                 
      [x] explainedVariance:          0.876                         
      [x] featuresCol:                features                      
      [x] labelCol:                   label                         
      [x] meanAbsoluteError:          0.185                         
      [x] meanSquaredError:           0.052                         
      [x] objectiveHistory:           0                             
      [x] pValues:                    0.063, 0.608, 0, 0.166, 0.6...
      [x] predictionCol:              prediction                    
      [x] r2:                         0.944                         
      [x] r2adj:                      0.918                         
      [x] rootMeanSquaredError:       0.227                         
      [x] tValues:                    -1.961, -0.521, 5.382, -1.4...

---

    Code
      fitted
    Output
       [1] 1.56 1.81 1.85 1.94 1.95 1.95 2.21 2.52 2.70 2.82 2.84 2.95 3.02 3.20 3.21
      [16] 3.38 3.41 3.50 3.52 3.55 3.55 3.61 3.64 3.70 3.72 3.75 3.76 3.92 3.96 4.80
      [31] 5.26 5.41

# Random Forest Classifier works

    Code
      class(ml_random_forest_classifier(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_random_forest_classifier(sc))
    Output
      [1] "ml_random_forest_classifier" "ml_connect_estimator"       
      [3] "ml_estimator"                "ml_pipeline_stage"          

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                  "ml_model_random_forest_classifier"
      [3] "ml_model_classification"           "ml_model_prediction"              
      [5] "ml_model"                         

# Random Forest Regressor works

    Code
      class(ml_random_forest_regressor(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_random_forest_regressor(sc))
    Output
      [1] "ml_random_forest_regressor" "ml_connect_estimator"      
      [3] "ml_estimator"               "ml_pipeline_stage"         

---

    Code
      model
    Message
      
      -- MLib model: RandomForestRegressionModel --
      
    Output
      [x] bootstrap:                TRUE         [x] maxDepth:                 5         
      [x] cacheNodeIds:             FALSE        [x] maxMemoryInMB:            256       
      [x] checkpointInterval:       10           [x] minInfoGain:              0         
      [x] featureSubsetStrategy:    auto         [x] minInstancesPerNode:      1         
      [x] featuresCol:              features     [x] minWeightFractionPerNode: 0         
      [x] impurity:                 variance     [x] numTrees:                 20        
      [x] labelCol:                 label        [x] predictionCol:            prediction
      [x] leafCol:                               [x] seed:                     100       
      [x] maxBins:                  32           [x] subsamplingRate:          1         

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                 "ml_model_random_forest_regressor"
      [3] "ml_model_regression"              "ml_model_prediction"             
      [5] "ml_model"                        

# Decision Tree Classifier works

    Code
      class(ml_decision_tree_classifier(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_decision_tree_classifier(sc))
    Output
      [1] "ml_decision_tree_classifier" "ml_connect_estimator"       
      [3] "ml_estimator"                "ml_pipeline_stage"          

---

    Code
      model
    Message
      -- MLib model: DecisionTreeClassificationModel --
      
    Output
      [x] cacheNodeIds:             FALSE           [x] maxMemoryInMB:            256          
      [x] checkpointInterval:       10              [x] minInfoGain:              0            
      [x] featuresCol:              features        [x] minInstancesPerNode:      1            
      [x] impurity:                 gini            [x] minWeightFractionPerNode: 0            
      [x] labelCol:                 label           [x] predictionCol:            prediction   
      [x] leafCol:                                  [x] probabilityCol:           probability  
      [x] maxBins:                  32              [x] rawPredictionCol:         rawPrediction
      [x] maxDepth:                 5               [x] seed:                     100          

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                  "ml_model_decision_tree_classifier"
      [3] "ml_model_classification"           "ml_model_prediction"              
      [5] "ml_model"                         

---

    Code
      table(x$prediction)
    Output
      
       0  1  2 
      50 49 51 

# Decision Tree Regressor works

    Code
      class(ml_decision_tree_regressor(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_decision_tree_regressor(sc))
    Output
      [1] "ml_decision_tree_regressor" "ml_connect_estimator"      
      [3] "ml_estimator"               "ml_pipeline_stage"         

---

    Code
      model
    Message
      -- MLib model: DecisionTreeRegressionModel --
      
    Output
      [x] cacheNodeIds:             FALSE        [x] maxDepth:                 5         
      [x] checkpointInterval:       10           [x] maxMemoryInMB:            256       
      [x] featuresCol:              features     [x] minInfoGain:              0         
      [x] impurity:                 variance     [x] minInstancesPerNode:      1         
      [x] labelCol:                 label        [x] minWeightFractionPerNode: 0         
      [x] leafCol:                               [x] predictionCol:            prediction
      [x] maxBins:                  32           [x] seed:                     100       

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                 "ml_model_decision_tree_regressor"
      [3] "ml_model_regression"              "ml_model_prediction"             
      [5] "ml_model"                        

# Kmeans works

    Code
      class(ml_kmeans(sc))
    Output
      [1] "ml_kmeans"            "ml_connect_estimator" "ml_estimator"        
      [4] "ml_pipeline_stage"   

---

    Code
      class(ml_kmeans(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

# Bisecting Kmeans works

    Code
      class(ml_bisecting_kmeans(sc))
    Output
      [1] "ml_bisecting_kmeans"  "ml_connect_estimator" "ml_estimator"        
      [4] "ml_pipeline_stage"   

---

    Code
      class(ml_bisecting_kmeans(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      preds
    Output
      [1] 24 29 38 59

# AFT Survival works

    Code
      class(ml_aft_survival_regression(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_aft_survival_regression(sc))
    Output
      [1] "ml_aft_survival_regressor" "ml_connect_estimator"     
      [3] "ml_estimator"              "ml_pipeline_stage"        

---

    Code
      model
    Message
      -- MLib model: AFTSurvivalRegressionModel --
      
      -- Coefficients: 
    Output
      [x] Intercept:    10.632   [x] age:         -0.065 
      [x] ecog_ps:     -0.067    [x] resid_ds:    -0.521 
      [x] rx:           0.521    
      [x] aggregationDepth:      2                      [x] maxBlockSizeInMB:      0                   
      [x] censorCol:             fustat                 [x] maxIter:               100                 
      [x] featuresCol:           features               [x] predictionCol:         prediction          
      [x] fitIntercept:          TRUE                   [x] quantileProbabilities: c(0.01, 0.05, 0.1...
      [x] labelCol:              label                  [x] tol:                   1e-06               

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                "ml_model_aft_survival_regressor"
      [3] "ml_model_regression"             "ml_model_prediction"            
      [5] "ml_model"                       

# GBT classifiers works

    Code
      class(ml_gbt_classifier(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_gbt_classifier(sc))
    Output
      [1] "ml_gbt_classifier"    "ml_connect_estimator" "ml_estimator"        
      [4] "ml_pipeline_stage"   

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"        "ml_model_gbt_classifier"
      [3] "ml_model_classification" "ml_model_prediction"    
      [5] "ml_model"               

---

    Code
      table(x)
    Output
      x
       0  1 
      19 13 

# GBT Regressor works

    Code
      class(ml_gbt_regressor(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_gbt_regressor(sc))
    Output
      [1] "ml_gbt_regressor"     "ml_connect_estimator" "ml_estimator"        
      [4] "ml_pipeline_stage"   

---

    Code
      model
    Message
      
      -- MLib model: GBTRegressionModel --
      
    Output
      [x] cacheNodeIds:             FALSE        [x] maxIter:                  20        
      [x] checkpointInterval:       10           [x] maxMemoryInMB:            256       
      [x] featureSubsetStrategy:    auto         [x] minInfoGain:              0         
      [x] featuresCol:              features     [x] minInstancesPerNode:      1         
      [x] impurity:                 variance     [x] minWeightFractionPerNode: 0         
      [x] labelCol:                 label        [x] predictionCol:            prediction
      [x] leafCol:                               [x] seed:                     100       
      [x] lossType:                 squared      [x] stepSize:                 0.1       
      [x] maxBins:                  32           [x] subsamplingRate:          1         
      [x] maxDepth:                 5            [x] validationTol:            0.01      

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"       "ml_model_gbt_regressor" "ml_model_regression"   
      [4] "ml_model_prediction"    "ml_model"              

# Isotonic regression works

    Code
      class(ml_isotonic_regression(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_isotonic_regression(sc))
    Output
      [1] "ml_isotonic_regressor" "ml_connect_estimator"  "ml_estimator"         
      [4] "ml_pipeline_stage"    

---

    Code
      model
    Message
      -- MLib model: IsotonicRegressionModel --
      
    Output
      [x] featureIndex:  0            [x] labelCol:      label     
      [x] featuresCol:   features     [x] predictionCol: prediction
      [x] isotonic:      TRUE         

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"            "ml_model_isotonic_regressor"
      [3] "ml_model_regression"         "ml_model_prediction"        
      [5] "ml_model"                   

# Generalized linear regression works

    Code
      class(ml_generalized_linear_regression(ml_pipeline(sc)))
    Output
      [1] "ml_connect_pipeline"       "ml_pipeline"              
      [3] "ml_connect_estimator"      "ml_estimator"             
      [5] "ml_connect_pipeline_stage" "ml_pipeline_stage"        

---

    Code
      class(ml_generalized_linear_regression(sc))
    Output
      [1] "ml_generalized_linear_regressor" "ml_connect_estimator"           
      [3] "ml_estimator"                    "ml_pipeline_stage"              

---

    Code
      class(model)
    Output
      [1] "ml_connect_model"                     
      [2] "ml_model_generalized_linear_regressor"
      [3] "ml_model_regression"                  
      [4] "ml_model_prediction"                  
      [5] "ml_model"                             

# Functions work

    Code
      colnames(transform_impl(model, tbl_mtcars, prep = TRUE))
    Output
       [1] "mpg"           "cyl"           "disp"          "hp"           
       [5] "drat"          "wt"            "qsec"          "vs"           
       [9] "am"            "gear"          "carb"          "features"     
      [13] "rawPrediction" "probability"   "prediction"   

---

    Code
      colnames(transform_impl(model, tbl_mtcars, prep = TRUE, remove = TRUE))
    Output
       [1] "mpg"           "cyl"           "disp"          "hp"           
       [5] "drat"          "wt"            "qsec"          "vs"           
       [9] "am"            "gear"          "carb"          "rawPrediction"
      [13] "probability"   "prediction"   

# Snowpark Connect (Snowflake)

    Code
      sc_out
    Output
      $conn
      $conn[[1]]
      $conn[[1]]$user
      [1] "test@user.com"
      
      $conn[[1]]$password
      [1] "testtoken"
      
      $conn[[1]]$warehouse
      [1] "testwh"
      
      $conn[[1]]$database
      [1] "testdb"
      
      $conn[[1]]$schema
      [1] "testschema"
      
      $conn[[1]]$account
      [1] "testaccount"
      
      
      
      $master_label
      Snowpark Connect - testaccount
      
      $con_class
      [1] "connect_snowflake"
      
      $cluster_id
      NULL
      
      $method
      [1] "snowpark_connect"
      
      $config
      NULL
      
      $misc
      $misc$sql_catalogs
      [1] "show databases"
      
      $misc$sql_tables_schema
      [1] "show tables in {schema}"
      
      $misc$sql_tables_catalog_schema
      [1] "show tables in {catalog}.{schema}"
      
      $misc$sql_schemas_catalog
      [1] "show schemas in database {catalog}"
      
      $misc$sql_schemas
      [1] "show schemas"
      
      
      $quote
      [1] ""
      

# Schema outputs as expected

    Code
      sdf_schema(tbl_mtcars)
    Output
      $mpg
      $mpg$name
      [1] "mpg"
      
      $mpg$type
      [1] "DoubleType"
      
      
      $cyl
      $cyl$name
      [1] "cyl"
      
      $cyl$type
      [1] "DoubleType"
      
      
      $disp
      $disp$name
      [1] "disp"
      
      $disp$type
      [1] "DoubleType"
      
      
      $hp
      $hp$name
      [1] "hp"
      
      $hp$type
      [1] "DoubleType"
      
      
      $drat
      $drat$name
      [1] "drat"
      
      $drat$type
      [1] "DoubleType"
      
      
      $wt
      $wt$name
      [1] "wt"
      
      $wt$type
      [1] "DoubleType"
      
      
      $qsec
      $qsec$name
      [1] "qsec"
      
      $qsec$type
      [1] "DoubleType"
      
      
      $vs
      $vs$name
      [1] "vs"
      
      $vs$type
      [1] "DoubleType"
      
      
      $am
      $am$name
      [1] "am"
      
      $am$type
      [1] "DoubleType"
      
      
      $gear
      $gear$name
      [1] "gear"
      
      $gear$type
      [1] "DoubleType"
      
      
      $carb
      $carb$name
      [1] "carb"
      
      $carb$type
      [1] "DoubleType"
      
      

# Internal functions work

    Code
      .strsep("domino", 3)
    Output
      [[1]]
      [1] "dom"
      
      [[2]]
      [1] "ino"
      

---

    Code
      .str_split_n("hello", "l", n_max = 3)
    Output
      [[1]]
      [1] "he" ""   "o" 
      

---

    Code
      .list_indices(c("one", "two", "three"), 2)
    Output
      [1] "one, two, ..."

---

    Code
      .simplify_pieces(c("one", "two", "three"), 2, FALSE)
    Output
      $strings
      $strings[[1]]
      [1] NA NA NA
      
      $strings[[2]]
      [1] NA NA NA
      
      
      $too_big
      NULL
      
      $too_sml
      [1] 1 2 3
      

# Binary evaluation works

    Code
      class(ml_binary_classification_evaluator(sc))
    Output
      [1] "ml_connect_estimator" "ml_estimator"         "ml_pipeline_stage"   

---

    Code
      ml_binary_classification_evaluator(preds, label_col = "am")
    Output
      [1] 1

# Regression evaluation works

    Code
      class(ml_regression_evaluator(sc))
    Output
      [1] "ml_connect_estimator" "ml_estimator"         "ml_pipeline_stage"   

---

    Code
      ml_regression_evaluator(preds, label_col = "wt", metric_name = "r2")
    Output
      [1] 0.9442661

# Multiclass evaluation works

    Code
      class(ml_multiclass_classification_evaluator(sc))
    Output
      [1] "ml_connect_estimator" "ml_estimator"         "ml_pipeline_stage"   

---

    Code
      ml_multiclass_classification_evaluator(preds, label_col = "species_idx")
    Output
      [1] 0.9933327

# Clustering evaluation works

    Code
      class(ml_clustering_evaluator(sc))
    Output
      [1] "ml_connect_estimator" "ml_estimator"         "ml_pipeline_stage"   

---

    Code
      ml_regression_evaluator(compute(mutate(preds, prediction = as.numeric(
        prediction))), label_col = "species_idx")
    Output
      [1] 1.407125

# ml_formula() works

    Code
      ml_formula(am ~ mpg, mtcars)
    Output
      $label
      [1] "am"
      
      $features
      [1] "mpg"
      

# ml_installed() works on simulated interactive session

    Code
      ml_installed(envname = test_env)
    Message
      ! Required Python libraries to run ML functions are missing
        Could not find: torch, torcheval, and scikit-learn
        Do you wish to install? (This will be a one time operation)

# Pivot longer

    Code
      tbl_pivot %>% tidyr::pivot_longer(-id, names_to = c(".value", "n"), names_sep = "_") %>%
        collect()
    Output
      # A tibble: 4 x 5
        id    n         z     y     x
        <chr> <chr> <dbl> <dbl> <dbl>
      1 A     1         1     2     3
      2 A     2         4     5     6
      3 B     1         7     8     9
      4 B     2        10    11    12

# Functions work

    Code
      colnames(transform_impl(model, tbl_mtcars, prep = TRUE))
    Output
       [1] "mpg"         "cyl"         "disp"        "hp"          "drat"       
       [6] "wt"          "qsec"        "vs"          "am"          "gear"       
      [11] "carb"        "features"    "prediction"  "probability"

---

    Code
      colnames(transform_impl(model, tbl_mtcars, prep = TRUE, remove = TRUE))
    Output
       [1] "mpg"         "cyl"         "disp"        "hp"          "drat"       
       [6] "wt"          "qsec"        "vs"          "am"          "gear"       
      [11] "carb"        "prediction"  "probability"

# Cross validator works

    Code
      class(cv)
    Output
      [1] "ml_connect_cross_validator" "ml_cross_validator"        
      [3] "ml_connect_estimator"       "ml_estimator"              
      [5] "ml_pipeline_stage"         

---

    Code
      class(tuning_model)
    Output
      [1] "ml_connect_cross_validator_model" "ml_cross_validator_model"        

---

    Code
      use_test_pull(metrics)
    Output
      [1] 0.0 0.5 1.0

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ide-snippet.R
\name{connection_databricks_shinyapp}
\alias{connection_databricks_shinyapp}
\title{A Shiny app that can be used to construct a \code{spark_connect} statement}
\usage{
connection_databricks_shinyapp()
}
\value{
A Shiny app
}
\description{
A Shiny app that can be used to construct a \code{spark_connect} statement
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pysparklyr-package.R
\docType{package}
\name{pysparklyr-package}
\alias{pysparklyr}
\alias{pysparklyr-package}
\title{pysparklyr: Provides a 'PySpark' Back-End for the 'sparklyr' Package}
\description{
It enables 'sparklyr' to integrate with 'Spark Connect', and 'Databricks Connect' by providing a wrapper over the 'PySpark' 'python' library.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/mlverse/pysparklyr}
  \item Report bugs at \url{https://github.com/mlverse/pysparklyr/issues}
}

}
\author{
\strong{Maintainer}: Edgar Ruiz \email{edgar@posit.co}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/start-stop-service.R
\name{spark_connect_service_start}
\alias{spark_connect_service_start}
\alias{spark_connect_service_stop}
\title{Starts and stops Spark Connect locally}
\usage{
spark_connect_service_start(
  version = "4.0",
  scala_version = "2.13",
  python_version = NULL,
  python = NULL,
  include_args = TRUE,
  additional_args = NULL,
  ...
)

spark_connect_service_stop(version = "4.0", ...)
}
\arguments{
\item{version}{Spark version to use (3.4 or above)}

\item{scala_version}{Acceptable Scala version of packages to be loaded}

\item{python_version}{Python version to use if a temporary Python environment
will be created}

\item{python}{Path to the Python executable to use while running (optional)}

\item{include_args}{Flag that indicates whether to add the additional arguments
to the command that starts the service. At this time, only the 'packages'
argument is submitted.}

\item{additional_args}{Vector of additional arguments to use when starting
the service}

\item{...}{Optional arguments; currently unused}
}
\value{
It returns messages to the console with the status of starting, and
stopping the local Spark Connect service.
}
\description{
Starts and stops Spark Connect locally
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/python-install.R
\name{install_pyspark}
\alias{install_pyspark}
\alias{install_databricks}
\title{Installs PySpark and Python dependencies}
\usage{
install_pyspark(
  version = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = TRUE,
  method = c("auto", "virtualenv", "conda"),
  as_job = TRUE,
  install_ml = FALSE,
  ...
)

install_databricks(
  version = NULL,
  cluster_id = NULL,
  envname = NULL,
  python_version = NULL,
  new_env = TRUE,
  method = c("auto", "virtualenv", "conda"),
  as_job = TRUE,
  install_ml = FALSE,
  ...
)
}
\arguments{
\item{version}{Version of 'databricks.connect' to install. Defaults to \code{NULL}.
If \code{NULL}, it will check against PyPi to get the current library version.}

\item{envname}{The name of the Python Environment to use to install the
Python libraries. Defaults to \code{NULL.} If \code{NULL}, a name will automatically
be assigned based on the version that will be installed}

\item{python_version}{The minimum required version of Python to use to create
the Python environment. Defaults to \code{NULL}. If \code{NULL}, it will check against
PyPi to get the minimum required Python version.}

\item{new_env}{If \code{TRUE}, any existing Python virtual environment and/or
Conda environment specified by \code{envname} is deleted first.}

\item{method}{The installation method to use. If creating a new environment,
\code{"auto"} (the default) is equivalent to \code{"virtualenv"}. Otherwise \code{"auto"}
infers the installation method based on the type of Python environment
specified by \code{envname}.}

\item{as_job}{Runs the installation if using this function within the
RStudio IDE.}

\item{install_ml}{Installs ML related Python libraries. Defaults to TRUE. This
is mainly for machines with limited storage to avoid installing the rather
large 'torch' library if the ML features are not going to be used. This will
apply to any environment backed by 'Spark' version 3.5 or above.}

\item{...}{Passed on to \code{\link[reticulate:py_install]{reticulate::py_install()}}}

\item{cluster_id}{Target of the cluster ID that will be used with.
If provided, this value will be used to extract the cluster's
version}
}
\value{
It returns no value to the R session. This function purpose is to
create the 'Python' environment, and install the appropriate set of 'Python'
libraries inside the new environment. During runtime, this function will send
messages to the console describing the steps that the function is
taking. For example, it will let the user know if it is getting the latest
version of the Python library from 'PyPi.org', and the result of such
query.
}
\description{
Installs PySpark and Python dependencies

Installs Databricks Connect and Python dependencies
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deploy.R
\name{deploy_databricks}
\alias{deploy_databricks}
\title{Deploys Databricks backed content to publishing server}
\usage{
deploy_databricks(
  appDir = NULL,
  python = NULL,
  account = NULL,
  server = NULL,
  lint = FALSE,
  forceGeneratePythonEnvironment = TRUE,
  version = NULL,
  cluster_id = NULL,
  host = NULL,
  token = NULL,
  confirm = interactive(),
  ...
)
}
\arguments{
\item{appDir}{A directory containing an application (e.g. a Shiny app or plumber API)
Defaults to NULL. If left NULL, and if called within RStudio, it will attempt
to use the folder of the currently opened document within the IDE. If there are
no opened documents, or not working in the RStudio IDE, then it will use
\code{getwd()} as the default value.}

\item{python}{Full path to a python binary for use by \code{reticulate.} It defaults to NULL.
If left NULL, this function will attempt to find a viable local Python
environment to replicate using the following hierarchy:
\enumerate{
\item \code{version} - Cluster's DBR version
\item \code{cluster_id} - Query the cluster to obtain its DBR version
\item If one is loaded in the current R session, it will verify that the Python
environment is suited to be used as the one to use
}}

\item{account}{The name of the account to use to publish}

\item{server}{The name of the target server to publish}

\item{lint}{Lint the project before initiating the project? Default to FALSE.
It has been causing issues for this type of content.}

\item{forceGeneratePythonEnvironment}{If an existing requirements.txt file is found,
it will be overwritten when this argument is TRUE.}

\item{version}{The Databricks Runtime (DBR) version. Use if \code{python} is NULL.}

\item{cluster_id}{The Databricks cluster ID. Use if \code{python}, and \code{version} are
NULL}

\item{host}{The Databricks host URL. Defaults to NULL. If left NULL, it will
use the environment variable \code{DATABRICKS_HOST}}

\item{token}{The Databricks authentication token. Defaults to NULL. If left NULL, it will
use the environment variable \code{DATABRICKS_TOKEN}}

\item{confirm}{Should the user be prompted to confirm that the correct
information is being used for deployment? Defaults to \code{interactive()}}

\item{...}{Additional named arguments passed to \code{rsconnect::deployApp()} function}
}
\value{
No value is returned to R. Only output to the console.
}
\description{
This is a convenience function that is meant to make it easier for
you to publish your Databricks backed content to a publishing server. It is
meant to be primarily used with Posit Connect.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/python-install.R
\name{installed_components}
\alias{installed_components}
\title{Lists installed Python libraries}
\usage{
installed_components(list_all = FALSE)
}
\arguments{
\item{list_all}{Flag that indicates to display all of the installed packages
or only the top two, namely, \code{pyspark} and \code{databricks.connect}}
}
\value{
Returns no value, only sends information to the console. The
information includes the current versions of 'sparklyr', and 'pysparklyr',
as well as the 'Python' environment currently loaded.
}
\description{
Lists installed Python libraries
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/connect-utils.R
\name{pyspark_config}
\alias{pyspark_config}
\title{Read Spark configuration}
\usage{
pyspark_config()
}
\value{
A list object with the initial configuration that will be used for
the Connect session.
}
\description{
Read Spark configuration
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/python-requirements.R
\name{requirements_write}
\alias{requirements_write}
\title{Writes the 'requirements.txt' file, containing the needed Python libraries}
\usage{
requirements_write(
  envname = NULL,
  destfile = "requirements.txt",
  overwrite = FALSE,
  ...
)
}
\arguments{
\item{envname}{The name of, or path to, a Python virtual environment.}

\item{destfile}{Target path for the requirements file. Defaults to 'requirements.txt'.}

\item{overwrite}{Replace the contents of the file if it already exists?}

\item{...}{Additional arguments passed to \code{reticulate::py_list_packages()}}
}
\value{
No value is returned to R. The output is a text file with the
list of Python libraries.
}
\description{
This is a helper function that it is meant to be used for deployments
of the document or application. By default, \code{deploy_databricks()} will run this
function the first time you use that function to deploy content to Posit Connect.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ml-prepare-dataset.R
\name{ml_prepare_dataset}
\alias{ml_prepare_dataset}
\title{Creates the 'label' and 'features' columns}
\usage{
ml_prepare_dataset(
  x,
  formula = NULL,
  label = NULL,
  features = NULL,
  label_col = "label",
  features_col = "features",
  keep_original = TRUE,
  ...
)
}
\arguments{
\item{x}{A \code{tbl_pyspark} object}

\item{formula}{Used when \code{x} is a \code{tbl_spark}. R formula.}

\item{label}{The name of the label column.}

\item{features}{The name(s) of the feature columns as a character vector.}

\item{label_col}{Label column name, as a length-one character vector.}

\item{features_col}{Features column name, as a length-one character vector.}

\item{keep_original}{Boolean flag that indicates if the output will contain,
or not, the original columns from \code{x}. Defaults to \code{TRUE}.}

\item{...}{Added for backwards compatibility. Not in use today.}
}
\value{
A \code{tbl_pyspark}, with either the original columns from \code{x}, plus the
'label' and 'features' column, or, the 'label' and 'features' columns only.
}
\description{
Creates the 'label' and 'features' columns
}
\details{
At this time, 'Spark ML Connect', does not include a Vector Assembler
transformer. The main thing that this function does, is create a 'Pyspark'
array column. Pipelines require a 'label' and 'features' columns. Even though
it is is single column in the dataset, the 'features' column will contain all
of the predictors insde an array. This function also creates a new 'label'
column that copies the outcome variable. This makes it a lot easier to remove
the 'label', and 'outcome' columns.
}
